/*=======================================================================================================*/
/*      Distributed Speech Recognition                                                                   */
/*      ETSI ES 202 050 Advanced Front-End Feature Extraction Algorithm & Compression Algorithm &        */
/*      ETSI ES 202 212 Extended Advanced Front-End Feature Extraction Algorithm & Compression Algorithm */  
/*      Bit exact C-language software implementation using ETSI basic operations library                 */
/*      3GPP TS 26.243 v 6.1.0 November 2004                                                                  */
/*      ANSI C code for the Fixed-Point Distributed Speech Recognition Extended Advanced Front-end       */
/*=======================================================================================================*/

/*-----------------
 * File Inclusions
 *-----------------*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "ParmInterface_B.h"
#include "16kHzProcExports_B.h"
#include "MathFunc.h"

#include "typedef.h"
#include "basic_op.h"
#include "count.h"


#if defined(F_16KHZ)
/*-------------
 * definitions
 *-------------*/
#define NS_SPEC_ORDER_16K                (X_INT16)(65)
#define NS_HANGOVER_16K                  (X_INT16)(15)
#define NS_MIN_SPEECH_FRAME_HANGOVER_16K (X_INT16)(4)
#define NS_ANALYSIS_WINDOW_16K        (X_INT16)(80)

#define NS_EPS_16K                   (X_FLOAT32)(exp ((double) -10.0))   

#define PERC_CODED			0.7

#define LAMBDA_NSE16k			(X_INT16)(0.99*(1<<15))
/*------------
 * structures
 *------------*/
struct QMF_FIR
{
  X_INT32 lengthQMF;         // number of FIR coefficients
  X_INT16 *dp_l;             // pointer to array with FIR coeff.
  X_INT16 *dp_h;             // pointer to array with FIR coeff.
  X_INT16 *T;
  X_INT16 T_dec;
};
struct DataFor16kProc_B
{
  // info needed from FEParamsX
  // computation of numFramesInBuffer assumes the same 
  // FrameLength and FrameShift for both NSX and CCX
  X_INT32 FrameLength;
  X_INT32 FrameShift;
  X_INT32 numFramesInBuffer;
  X_INT32 SamplingFrequency;

  BOOLEAN Do16kHzProc;
  X_INT32 *hpBands_B;
  X_INT32 hpBandsSize;
  X_INT32 CodeForBands16k_B [HP16k_MEL_USED * NB_LP_BANDS_CODING];
  X_INT32 bufferCodeForBands16k_B [3 * HP16k_MEL_USED * NB_LP_BANDS_CODING];  /* 3 = delay between the 1st stage output and MFCC proc */
  X_INT16 codeWeights_B [NB_LP_BANDS_CODING];
  X_INT16 bufferCodeWeights_B [3 * NB_LP_BANDS_CODING];  /* 3 = delay between the 1st stage output and MFCC proc */
  
  QMF_FIR * pQMF_Fir;
  
  X_INT32 *bufferData16k_B;
  X_INT32 bufData16kSize;
  MelFB_Window *FirstWindow16k;
  X_INT32 noiseSE16k_B [HP16k_MEL_USED];
  X_INT16 noise_dec;

  X_INT32 BandsForCoding16k_B [9];

  X_INT32 vadCounter16k;
  X_INT32 vad16k;
  X_INT32 nbSpeechFrames16k;
  X_INT32 hangOver16k;
  X_INT32 meanEn16k;

  X_INT32 nb_frame_threshold_nse;

  X_INT16 lambda_nse;

  X_INT32 *dataHP_B;

  X_INT16 dec_16k[5];
  X_INT16 BFC_dec[1];
  X_INT16 fb16k_dec[3];
};


static QMF_FIR * QMF_FIR_Init_B (void) ;
static void hq_free_B (QMF_FIR *fir_ptr) ;


static X_INT32 Mult16x32(X_INT16 a16,X_INT32 b32) {
    X_INT16 low = extract_l(b32) ;
    X_INT16 high  = extract_h(b32) ;
    high = sub(high,shr(low,15)) ;
    return L_mac(L_mult(a16,high),mult_r(a16,shr_r(low,1)),1) ;
}




/*---------------------------------------------------------------------------
 * FUNCTION NAME: Pow2() 
 *
 * PURPOSE: 
 *     L_x = pow(2.0, exponent.fraction)         (exponent = integer part)  
 *         = pow(2.0, 0.fraction) << exponent                                
 *   The function Pow2(L_x) is approximated by a table and linear            
 *   interpolation.                                                          
 *                                                                           
 *   1- i = bit10-b15 of fraction,   0 <= i <= 31                            
 *   2- a = bit0-b9   of fraction                                            
 *   3- L_x = table[i]<<16 - (table[i] - table[i+1]) * a * 2                 
 *   4- L_x = L_x >> (30-exponent)     (with rounding)                       
 *
 * INPUT: x 
 *
 * OUTPUT: n/a
 *
 * RETURN VALUE : L_x 
 *
 *---------------------------------------------------------------------------*/
static const X_INT16 table_pow2[33] =
{
    16384, 16743, 17109, 17484, 17867, 18258, 18658, 19066, 19484, 19911,
    20347, 20792, 21247, 21713, 22188, 22674, 23170, 23678, 24196, 24726,
    25268, 25821, 26386, 26964, 27554, 28158, 28774, 29405, 30048, 30706,
    31379, 32066, 32767
};

/* Output 2**(30-exponent) */
X_INT32 Pow2(                               /* (o) Q0  : result       (range: 0<=val<=0x7fffffff) */
	     //X_INT16 exponent,                      /* (i) Q0  : Integer part.      (range: 0<=val<=30)   */
	     //X_INT16 fraction                       /* (i) Q15 : Fractionnal part.  (range: 0.0<=val<1.0) */
	     X_INT32 x
)
{
    X_INT16  i, a, tmp;
    X_INT32 L_x;
    X_INT16 exponent;
    X_INT16 fraction;

    exponent = extract_l(L_shr(x,10));				//extract_h(x);
    fraction = extract_l(L_msu(L_shl(x,5), exponent, 16384));	//(L_shr(x,1), exponent, 16384));

    L_x = L_mult(fraction, 32);            /* L_x = fraction<<6           */
    i = extract_h(L_x);                    /* Extract b10-b16 of fraction */
    L_x = L_shr(L_x, 1);
    a = extract_l(L_x);                    /* Extract b0-b9   of fraction */
    a = (X_INT16) (a & (X_INT16) 0x7fff);    logic16();

    L_x = L_deposit_h(table_pow2[i]);      /* table[i] << 16        */
    tmp = sub(table_pow2[i], table_pow2[i + 1]);        /* table[i] - table[i+1] */
    L_x = L_msu(L_x, tmp, a);              /* L_x -= tmp*a*2        */


    return (L_x);
}




/*---------------------------------------------------------------------------
 * FUNCTION NAME: MergeSSandCoded_B
 *
 * PURPOSE: Merge SS and coded Bands
 *
 * INPUT:
 * - *intBuffer, 8kHz filter for Merge
 * - *hp16kBandsSize_in, 16kHz Band Size
 * - *hp16kBands, 16k coded data
 * - numChannelsWI7, 1
 * - *pData16k, 16k data structure
 *   
 *
 * OUTPUT:
 * - *intBuffer, Merged output
 *
 *
 * RETURN VALUE : NULL
 *  
 *
 *---------------------------------------------------------------------------*/
void MergeSSandCoded_B (X_INT32 *intBuffer, X_INT32 *hp16kBands, X_INT32 numChannelsWI7, X_INT16 *hp16kBandsSize_in,struct  DataFor16kProc_B *pData16k)
{
  int i;
  X_INT32 bandAverage_B;
  X_INT16 hp16kBandsSize = *hp16kBandsSize_in; move16();

  for (i=0 ; i<hp16kBandsSize ; i++){
    intBuffer [numChannelsWI7 + i] = L_add(Mult16x32((X_INT16)(PERC_CODED*(1<<15)), hp16kBands[i]), 
    					   Mult16x32((X_INT16)((1-PERC_CODED)*(1<<15)), intBuffer[numChannelsWI7+i]) ); move32();
  }

  if (hp16kBandsSize > 0)
    { 
      // perform interpolation
      bandAverage_B = L_add(L_shr_r(intBuffer [numChannelsWI7 - 1],1), L_shr_r(intBuffer [numChannelsWI7],1));

      intBuffer [numChannelsWI7 - 1] = L_add (Mult16x32((X_INT16)(0.6*(1<<15)), intBuffer[numChannelsWI7 - 1]), 
					      Mult16x32( (X_INT16)(0.4*(1<<15)), bandAverage_B) ); move32();

      intBuffer [numChannelsWI7] = L_add (Mult16x32((X_INT16)(0.6*(1<<15)), intBuffer [numChannelsWI7]),
					  Mult16x32((X_INT16)(0.4*(1<<15)), bandAverage_B) ); move32();
    }

  *hp16kBandsSize_in = hp16kBandsSize; move16();
}




/*---------------------------------------------------------------------------
 * FUNCTION NAME: CorrectEnergy_B
 *
 * PURPOSE: Energy Correction after merging 8k and 16k data
 *
 * INPUT:
 *  - LogEnergy, energy calculated from LFB
 *  - *eng_dec, Multiplier to LogEnergy - also returns the modified multiplier
 *  - hp16kBandsSize, band size (length)
 *  - *intBuffer, 16khz Merged signal
 *  - preem, preemphasis value = 0.9
 *
 * OUTPUT
 *  - *end_dec, Multiplier of the corrected energy value
 *                
 *
 * RETURN VALUE
 *  - Log energy value after correction
 *
 *---------------------------------------------------------------------------*/

X_INT32 CorrectEnergy_B (X_INT32 LogEnergy, X_INT16 *eng_dec, X_INT32 hp16kBandsSize, X_INT32 * intBuffer, X_INT32 numChannelsWI7, float preem)
{
  int i;
  X_INT32 energyHP = 0;
  X_INT32 x;
  X_INT16 exponent ;
  X_INT16 prev_dec=30;
  const X_INT32 preemphs_correction = Log_2 ((1<<10) + (int)(1024*preem)) - (10*(1<<10)); /* wMOPS from Log_2 should not be counted there */
  move32(); move16();

  x = L_sub(intBuffer[numChannelsWI7 + 0], preemphs_correction);
  exponent = sub(30,extract_l(L_shr(x,10)));
  x = Pow2 (x); 
  prev_dec = sub(exponent, 1);
  energyHP = L_shr_r(x,1);
  
  for (i=1 ; i<hp16kBandsSize ; i++){
    x = L_sub(intBuffer[numChannelsWI7 + i], preemphs_correction);
    exponent = sub(30,extract_l(L_shr(x,10)));
    x = Pow2 (x); 

    test();
    if(sub(exponent,prev_dec)>=0)	//if(exponent >= prev_dec)
      {
	energyHP = L_add(L_shr_r(energyHP,1), L_shr_r(x, add(sub(exponent, prev_dec), 1)));
	prev_dec = sub(prev_dec,1);
      }
    else
      {
	energyHP = L_add( L_shr_r(energyHP, add( sub(prev_dec, exponent), 1)), L_shr_r(x, 1));
	prev_dec = sub(exponent, 1);
      }

  }

  test();
  if(sub(*eng_dec,prev_dec)>=0)		//if(*eng_dec >= prev_dec)
    {
      LogEnergy = L_shr_r(LogEnergy, add( sub(*eng_dec,prev_dec),1));
      energyHP = L_shr_r(energyHP,1);
      LogEnergy = L_add(LogEnergy,energyHP);
      *eng_dec = sub(prev_dec,1);
    }
  else
    {
      LogEnergy = L_shr_r(LogEnergy,1);
      energyHP = L_shr_r(energyHP, add( sub(prev_dec,*eng_dec),1));
      LogEnergy = L_add(LogEnergy,energyHP);
      *eng_dec = sub(*eng_dec,1);
    }

  return LogEnergy;
}




/*---------------------------------------------------------------------------
 * FUNCTION NAME: Do16kProcAlloc_B
 *
 * PURPOSE: Allocates memory for DataFor16kProc_B structure
 *
 * INPUT: NONE
 *
 * OUTPUT: NONE
 *
 * RETURN VALUE
 *  - Pointer to the allocated memory (structure)
 *
 *---------------------------------------------------------------------------*/
struct DataFor16kProc_B *Do16kProcAlloc_B ()
{
  struct DataFor16kProc_B *newData16k = calloc (1, sizeof (DataFor16kProc_B));
  if (newData16k == NULL)	
    return NULL;
  else
    return newData16k;
}



/*---------------------------------------------------------------------------
 * FUNCTION NAME: Do16kProcInit_B
 *
 * PURPOSE: Initializes DataFor16kProc_B structure
 *
 * INPUT:
 *  - * This, pointer to FEParamsX_F
 *   
 *
 * OUTPUT:
 *  - Initializes components for 16k processing in FEParamsX_F structure
 *                
 *
 * RETURN VALUE
 *  : NONE
 *
 *---------------------------------------------------------------------------*/
void Do16kProcInit_B (FEParamsX_F * This)
{
  DataFor16kProc_B *newData16k = (DataFor16kProc_B *)This->pData16k;
  int i;

  newData16k->Do16kHzProc       = This->Do16kHzProc;
  newData16k->FrameLength       = This->FrameLength;
  newData16k->FrameShift        = This->FrameShift;
  newData16k->SamplingFrequency = This->SamplingFrequency;

  newData16k->hpBandsSize = ((int) (newData16k->FrameLength / newData16k->FrameShift) + 1) * HP16k_MEL_USED;
  newData16k->hpBands_B     = (X_INT32 *) malloc (sizeof (X_INT32) * newData16k->hpBandsSize);
  for (i=0 ; i<newData16k->hpBandsSize ; i++) newData16k->hpBands_B [i] = 0;

  // HP-VAD related variables
  newData16k->vadCounter16k     = 0;
  newData16k->vad16k            = 0;
  newData16k->nbSpeechFrames16k = 0;
  newData16k->hangOver16k       = 0;
  newData16k->meanEn16k         = 0;

  test();
  if ((newData16k->FrameLength % newData16k->FrameShift) == 0)
    newData16k->numFramesInBuffer = newData16k->FrameLength / newData16k->FrameShift;
  else
    newData16k->numFramesInBuffer = (newData16k->FrameLength / newData16k->FrameShift) + 1;

  // assuming the same delay for both WF stages
  newData16k->bufData16kSize = 2 * newData16k->numFramesInBuffer * newData16k->FrameShift- NS_ANALYSIS_WINDOW_16K ; 
  newData16k->bufferData16k_B = (X_INT32 *) malloc (sizeof (X_INT32) * newData16k->bufData16kSize); 
  for (i=0 ; i<newData16k->bufData16kSize ; i++) newData16k->bufferData16k_B [i] = 0;
  for (i=0 ; i<HP16k_MEL_USED ; i++) newData16k->noiseSE16k_B [i] = 0;
  newData16k->noise_dec = 256; 

  newData16k->pQMF_Fir = QMF_FIR_Init_B ();

  newData16k->nb_frame_threshold_nse = NE16k_FRAMES_THRESH; 
  newData16k->lambda_nse             = LAMBDA_NSE16k; 
  newData16k->dataHP_B = newData16k->bufferData16k_B + (newData16k->bufData16kSize - newData16k->FrameShift) ;
  newData16k->dec_16k[4] = 16; 
}


/*---------------------------------------------------------------------------
 * FUNCTION NAME: Do16kProcDelete_B
 *
 * PURPOSE: Frees memory allocated for DataFor16kProc_B structure
 *
 * INPUT:
 *  - *pData16k, Pointer to the structure that needs to be freed
 *   
 *
 * OUTPUT : NONE
 *           
 * RETURN VALUE
 *  : NONE
 *
 *---------------------------------------------------------------------------*/
void Do16kProcDelete_B (DataFor16kProc_B *pData16k)
{
  if (pData16k->bufferData16k_B != NULL)
    free (pData16k->bufferData16k_B);

  if (pData16k->hpBands_B != NULL)
    free (pData16k->hpBands_B);

  if (pData16k->pQMF_Fir != NULL)
    hq_free_B (pData16k->pQMF_Fir);

  if (pData16k != NULL)
    free (pData16k);
}



/*---------------------------------------------------------------------------
 * FUNCTION NAME: GetBandsForDecoding16k_B
 *
 * PURPOSE: Initializes array used for decoding
 *
 * INPUT:
 *  - *intData, input from 8kHz SE
 *  - in_dec, Multiplier for intData
 *  - inDataSize, CEP_FFT_LENGTH/2+1
 *   
 *
 * OUTPUT
 *  - *bandsForCoding16k, computed bands for coding
 *  - *out_dec, multiplier for bandsForCoding16k
 *
 * RETURN VALUE
 *  NULL
 *
 *---------------------------------------------------------------------------*/

void GetBandsForDecoding16k_B (X_INT32 *intData, X_INT16 in_dec, X_INT32 *bandsForCoding16k, X_INT16 *out_dec, X_INT32 inDataSize)
{
  int i;

  X_INT32 aux[3];

  for (i=0 ; i<3 ; i++)
    {aux[i] = 0; move32();}

  for (i=66 ; i<77 ; i++)
    {aux[0] = L_add(aux[0], L_shr_r(intData[i],5)); move32();}

  for (i=78 ; i<97 ; i++)
    {aux[1] = L_add(aux[1], L_shr_r(intData[i],5)); move32();}

  for (i=98 ; i<129 ; i++)
    {aux[2] = L_add(aux[2], L_shr_r(intData[i],5)); move32();}

  *out_dec = sub(in_dec,4);

  for(i=0; i<3; i++)
    {bandsForCoding16k [i] = aux[i]; move32();}

}



/*---------------------------------------------------------------------------
 * FUNCTION NAME: DecodeBands16k_B
 *
 * PURPOSE: HP Decoding
 *
 * INPUT:
 *  - *lpBands_B, LFB value from 8k processing
 *  - *codeForBands16k_B, coding value
 *  - codeWeights, Weight for each coding value
 *
 * OUTPUT
 *  - *fb16k_B, Decoded value
 *                
 *
 * RETURN VALUE
 *  NULL
 *
 *---------------------------------------------------------------------------*/

void DecodeBands16k_B (X_INT32 *fb16k_B,X_INT32 *lpBands_B,X_INT32 *codeForBands16k_B, X_INT16 *codeWeights)
{
  int i, j;

  for (i=0 ; i<HP16k_MEL_USED ; i++)
    {
      fb16k_B[i] = 0; move32(); 
      for (j=0 ; j<3 ; j++)
	{fb16k_B [i] = L_add(fb16k_B[i], Mult16x32(codeWeights[j], L_sub(lpBands_B[j],codeForBands16k_B[3*i+j])) ); move32();}

    }
}

#define NS_NB_FRAME_THRESHOLD_NSE	100
static const X_INT16 LambdaNSEx2[NS_NB_FRAME_THRESHOLD_NSE] = {
  0,
  0,
  16384,
  21845,
  24576,
  26214,
  27306,
  28086,
  28672,
  29127,
  29491,
  29789,
  30037,
  30247,
  30427,
  30583,
  30720,
  30840,
  30947,
  31043,
  31129,
  31207,
  31278,
  31343,
  31402,
  31457,
  31507,
  31554,
  31597,
  31638,
  31675,
  31710,
  31744,
  31775,
  31804,
  31831,
  31857,
  31882,
  31905,
  31927,
  31948,
  31968,
  31987,
  32005,
  32023,
  32039,
  32055,
  32070,
  32085,
  32099,
  32112,
  32125,
  32137,
  32149,
  32161,
  32172,
  32182,
  32193,
  32203,
  32212,
  32221,
  32230,
  32239,
  32247,
  32256,
  32263,
  32271,
  32278,
  32286,
  32293,
  32299,
  32306,
  32312,
  32319,
  32325,
  32331,
  32336,
  32342,
  32347,
  32353,
  32358,
  32363,
  32368,
  32373,
  32377,
  32382,
  32386,
  32391,
  32395,
  32399,
  32403,
  32407,
  32411,
  32415,
  32419,
  32423,
  32426,
  32430,
  32433,
  32437
};


/*---------------------------------------------------------------------------
 * FUNCTION NAME: DoSpecSub16k_B
 *
 * PURPOSE: Performs Spectral Subtraction based on VADNest
 *
 * INPUT:
 *  - *FB16k32, QMF HP input
 *  - *decal1, Multiplier for FB16k32
 *  - *pData16k, Pointer to the data structure for 16k data
 *  - vadCounter16k, counter used for VADNest
 *  - inFB16kSize, 16k data length
 *
 * OUTPUT
 *  - *FB16k32, Output from Spectral Substraction
 *  - meanEn16k, mean energy of frame used by VADNest
 *  - noiseDec, Multiplier of Noise energy
 *
 * RETURN VALUE
 *  NULL
 *
 *---------------------------------------------------------------------------*/
void DoSpecSub16k_B (X_INT32 *FB16k32, X_INT16 * decal1, DataFor16kProc_B *pData16k, X_INT32 vadCounter16k, X_INT32 inFB16kSize)
{

  X_INT32 *nbSpeechFrames16k     = &(pData16k->nbSpeechFrames16k);
  X_INT32 *hangOver16k           = &(pData16k->hangOver16k);

  X_INT32 *meanEn16k           = &(pData16k->meanEn16k);
  const X_INT32 nb_frame_threshold_nse = pData16k->nb_frame_threshold_nse;

  const X_INT16 lambda_nse           = pData16k->lambda_nse;

  int i;
  X_INT32 flagVAD;
  X_INT32 nbFrame;

  X_INT32 frameEn32, meanEn32;

  X_INT16 lambdaNSE32;

  X_INT32 *noiseFB16k_B = pData16k->noiseSE16k_B;
  X_INT16 noise_dec = pData16k->noise_dec;
  X_INT16 decal = *decal1;
  X_INT32 diff32, floor32;

  move16(); move16();
  pData16k->vadCounter16k = vadCounter16k; move32();

  nbFrame = vadCounter16k; move32();
  
  meanEn32 = *meanEn16k; move32();

  test(); move16();
  if(L_sub(nbFrame,nb_frame_threshold_nse)<0)
      lambdaNSE32 = LambdaNSEx2[nbFrame];
  else
      lambdaNSE32 = lambda_nse;

  // frame energy calc.
  frameEn32 = 0; move32();
  for (i=0; i<HP16k_MEL_USED; i++)
    frameEn32 = L_add(frameEn32,L_shr(FB16k32[i],2));

  frameEn32 = Log_2(frameEn32) - shl(sub(decal, 2), 10); L_sub(0,0);

  test();
  if(L_sub(frameEn32,(-10205))<0)
    {frameEn32 = (-10205); move32();}

  frameEn32 <<= 5;                               shl(frameEn32,5);

  test(); test(); 
  if((L_sub(L_sub(frameEn32, meanEn32), (int)(1.2*47274.23+0.5))<0) || (sub(nbFrame,10)<0))
    {
      if (sub(nbFrame,10) < 0)
	{
	  sub(32767,lambdaNSE32);
	  meanEn32 = L_add(meanEn32, (Mult16x32((32767-lambdaNSE32), L_sub(frameEn32, meanEn32))));
 	}
      else
	{
	  test();
	  if (L_sub(frameEn32, meanEn32)<0){
	    L_add(0,16);
	    meanEn32 = L_add(meanEn32, L_shr_r(L_sub(Mult16x32(20972, frameEn32), Mult16x32(20972,meanEn32))+16, 5));
	  }
	  else{
	    L_add(0,32);
	    meanEn32 = L_add(meanEn32, L_shr_r(L_sub(Mult16x32(20972, frameEn32), Mult16x32(20972,meanEn32))+32, 7));
	  }
	}
    }

  test();
  if (L_sub(L_sub(frameEn32, meanEn32), (int)(2.2*47274.23 +0.5)) > 0)
    {
      flagVAD = 1; move16();
      (*nbSpeechFrames16k)++;              add(*nbSpeechFrames16k,1);
    }
  else
    {
      test();
      if (L_sub(*nbSpeechFrames16k,NS_MIN_SPEECH_FRAME_HANGOVER_16K)>0)
        { *hangOver16k = NS_HANGOVER_16K; move16(); }
      *nbSpeechFrames16k = 0; move16();
      test(); move16();
      if (*hangOver16k != 0)
        {
          (*hangOver16k)--;                sub(*hangOver16k,1);
          flagVAD = 1;
        }
      else
        flagVAD = 0;
    }

  //Adjust scale for Noise and input signal
  test();
  if(sub(decal,noise_dec)<=0){
    for(i=0; i<HP16k_MEL_USED; i++){
      noiseFB16k_B[i] = L_shr_r(noiseFB16k_B[i], sub(noise_dec, decal)); move32();
    }
    noise_dec = decal; move16();
  }
  else{
    for(i=0; i<HP16k_MEL_USED; i++){
      FB16k32[i] = L_shr_r(FB16k32[i], sub(decal, noise_dec)); move32();
    }
    decal = noise_dec; move16();
  }

  test();
  if (flagVAD == 0)
    {
      for (i=0 ; i<inFB16kSize ; i++)
        {
	  test();
          if (sub(nbFrame,10)<0){
	    noiseFB16k_B [i] = L_add ( Mult16x32(lambdaNSE32, noiseFB16k_B[i]),
					    Mult16x32(sub(32767,lambdaNSE32), FB16k32[i]) );
	    move32();
	  }
          else
            {
	      test();
	      if (L_sub(FB16k32 [i], noiseFB16k_B[i])<0){
		noiseFB16k_B [i] = L_add ( Mult16x32(lambdaNSE32, noiseFB16k_B[i]),
					    Mult16x32(sub(32767,lambdaNSE32), FB16k32[i]) );
		move32();
	      }

              else{
		noiseFB16k_B [i] = L_add ( Mult16x32((0.995*(1<<15)+0.5), noiseFB16k_B[i]),
					      Mult16x32(((1-0.995)*(1<<15)+0.5), FB16k32[i]) );
		move32();
	      }

            }
	  test();
          if (L_sub(noiseFB16k_B[i],1)<0)
	    {noiseFB16k_B[i] = 1; move32();}
        }
    }

  for (i=0 ; i<inFB16kSize ; i++)
    {
      
      floor32 = Mult16x32((0.1*(1<<15)+0.5), FB16k32[i]);
      diff32 = L_sub(FB16k32[i], L_add(noiseFB16k_B[i], L_shr_r(noiseFB16k_B[i],1)));

      test(); move32();
      if(L_sub(diff32,floor32)>0)	//if(diff32 > floor32)
	FB16k32 [i] = diff32;
      else
	FB16k32 [i] = floor32;
    }

  *meanEn16k = meanEn32; move32();

  pData16k->noise_dec = decal; move16();
  *decal1 = decal; move16();

  return;
}


/*---------------------------------------------------------------------------
 * FUNCTION NAME: GetBandsForCoding16k_B
 *
 * PURPOSE: Initializes values for Coding
 *
 * INPUT:
 *  - *inData, Input signal SE data
 *  - *nSigSE_dec, multiplier for *inData
 *
 * OUTPUT
 *  - *bandsForCoding16k, Initialized values for Coding
 *  - *decal, multiplier for bandsForcoding16k
 *
 * RETURN VALUE
 *  NULL
 *
 *---------------------------------------------------------------------------*/
void GetBandsForCoding16k_B (X_INT32 *inData, X_INT16 *nSigSE_dec , X_INT32 *bandsForCoding16k, X_INT16 *decal)
{
  int i;

  X_INT32 *codingBands32=bandsForCoding16k;
  X_INT16 dec_str;

  *decal = nSigSE_dec[33]; move16();

  for(i=33; i<65; i++){
    test();
    if(sub(*decal, nSigSE_dec[i])>0){
      *decal = nSigSE_dec[i]; move16();}
  }
  dec_str = shl(*decal,1);
  *decal = sub(dec_str,4);

  codingBands32[0] = codingBands32[1] = codingBands32[2] = 0; 
  move32(); move32(); move32();

  for (i=33; i<39; i++){
    dec_str = sub(shl(nSigSE_dec[i],1), *decal);
    codingBands32 [0] = L_add(codingBands32[0], L_shr_r(L_add(inData [i], L_shl(1, sub(dec_str,1))), dec_str)); move32();
  }
  for (i=39; i<49; i++){
    dec_str = sub(shl(nSigSE_dec[i],1), *decal);
    codingBands32 [1] = L_add(codingBands32[1], L_shr_r(L_add(inData [i], L_shl(1, sub(dec_str,1))), dec_str)); move32();
  }
  for (i=49; i<65; i++){
    dec_str = sub(shl(nSigSE_dec[i],1), *decal);
    codingBands32 [2] = L_add(codingBands32[2], L_shr_r(L_add(inData [i], L_shl(1, sub(dec_str,1))), dec_str)); move32();
  }

}



/*---------------------------------------------------------------------------
 * FUNCTION NAME: CodeBands16k_B
 *
 * PURPOSE: Performs coding on QMF HP
 *
 * INPUT:
 *  - fb16k, 16k data from QFM HP
 *  - lpBands, 8k data from QFM LP
 *
 * OUTPUT
 *  - *codeForBands16k_B, coding output
 *                
 *
 * RETURN VALUE
 *  NULL
 *
 *---------------------------------------------------------------------------*/

void CodeBands16k_B (X_INT32 *fb16k, X_INT32 *lpBands,X_INT32 *codeForBands16k_B)
{
  int i, j;

  for (i=0 ; i<HP16k_MEL_USED ; i++)
    for (j=0 ; j<3 ; j++){
      codeForBands16k_B [3 * i + j] = L_sub(lpBands [j], fb16k [i]); move32();
    }

}


/*---------------------------------------------------------------------------
 * FUNCTION NAME: DP_HP_filters_B
 *
 * PURPOSE: Initializes the QMF filter arrays
 *
 * INPUT:
 *  - *dp_h, Pointer to HP filter array
 *  - *dp_l, Pointer to LP filter array
 *  - *lengthQMF, Pointer to QMF filter length
 *
 * OUTPUT : Initializes all pointers with appropriate value
 *           
 *                
 *
 * RETURN VALUE
 *  NULL
 *
 *---------------------------------------------------------------------------*/
#define LENGTH_QMF 118
#define f24 0x01

#define SHFF_H	8

#define L_H		16

#define GT_HI(v) (X_INT16) (((X_INT32) (v))>> SHFF_H) 
#define GT_LO(v) (X_INT16) (((X_INT32) (v)) & ((1<<SHFF_H)-1)) 


static void DP_HP_filters_B (const X_INT16 *dp_h[],const X_INT16 *dp_l[], X_INT32 *lengthQMF)
{
static const X_INT16 dp02_h [LENGTH_QMF/2] = {
    (X_INT16) (1584. / f24), (X_INT16) ( 805. / f24), (X_INT16) ( -4192. / f24), (X_INT16) ( -8985. / f24),
	(X_INT16) (-5987. / f24), (X_INT16) ( 2583. / f24), (X_INT16) ( 4657. / f24), (X_INT16) ( -3035. / f24),
	(X_INT16) (-7004. / f24), (X_INT16) ( 1542. / f24), (X_INT16) ( 8969. / f24), (X_INT16) ( 567. / f24),
	(X_INT16) (-10924. / f24), (X_INT16) ( -3757. / f24), (X_INT16) ( 12320. / f24), (X_INT16) ( 7951. / f24),

	
	GT_HI(-12793. / f24), GT_HI( -13048. / f24), GT_HI( 11923. / f24), GT_HI( 18793. / f24),
	GT_HI(-9331. / f24), GT_HI( -24802. / f24), GT_HI( 4694. / f24), GT_HI( 30570. / f24),

	GT_HI(2233. / f24) , GT_HI( -35439. / f24) , GT_HI( -11526. / f24) , GT_HI( 38680. / f24) ,
	GT_HI(23114. / f24) , GT_HI( -39474. / f24) , GT_HI( -36701. / f24) , GT_HI( 36999. / f24) ,
	GT_HI(51797. / f24) , GT_HI( -30419. / f24) , GT_HI( -67658. / f24) , GT_HI( 18962. / f24) ,
	GT_HI(83318. / f24) , GT_HI( -1927. / f24) , GT_HI( -97566. / f24) , GT_HI( -21284. / f24) ,
	GT_HI(108971. / f24) , GT_HI( 51215. / f24) , GT_HI( -115837. / f24) , GT_HI( -88430. / f24) ,
	GT_HI(116130. / f24) , GT_HI( 133716. / f24) , GT_HI( -107253. / f24) , GT_HI( -188497. / f24) ,
	GT_HI(85497. / f24) , GT_HI( 255795. / f24) , GT_HI( -44643. / f24) , GT_HI( -342699. / f24) ,
	GT_HI(-28185. / f24) , GT_HI( 468096. / f24) , GT_HI( 167799. / f24) , GT_HI( -696809. / f24) ,
	GT_HI(-519818. / f24) , GT_HI( 1446093. / f24) , GT_HI( 3562497. / f24)

 };

static const X_INT16 dp02_l [LENGTH_QMF/2 - L_H] = {
	GT_LO(-12793. / f24), GT_LO( -13048. / f24), GT_LO( 11923. / f24), GT_LO( 18793. / f24),
	GT_LO(-9331. / f24), GT_LO( -24802. / f24), GT_LO( 4694. / f24), GT_LO( 30570. / f24),
	GT_LO(2233. / f24) , GT_LO( -35439. / f24) , GT_LO( -11526. / f24) , GT_LO( 38680. / f24) ,
	GT_LO(23114. / f24) , GT_LO( -39474. / f24) , GT_LO( -36701. / f24) , GT_LO( 36999. / f24) ,
	GT_LO(51797. / f24) , GT_LO( -30419. / f24) , GT_LO( -67658. / f24) , GT_LO( 18962. / f24) ,
	GT_LO(83318. / f24) , GT_LO( -1927. / f24) , GT_LO( -97566. / f24) , GT_LO( -21284. / f24) ,
	GT_LO(108971. / f24) , GT_LO( 51215. / f24) , GT_LO( -115837. / f24) , GT_LO( -88430. / f24) ,
	GT_LO(116130. / f24) , GT_LO( 133716. / f24) , GT_LO( -107253. / f24) , GT_LO( -188497. / f24) ,
	GT_LO(85497. / f24) , GT_LO( 255795. / f24) , GT_LO( -44643. / f24) , GT_LO( -342699. / f24) ,
	GT_LO(-28185. / f24) , GT_LO( 468096. / f24) , GT_LO( 167799. / f24) , GT_LO( -696809. / f24) ,
	GT_LO(-519818. / f24) , GT_LO( 1446093. / f24) , GT_LO( 3562497. / f24)

 };

  *lengthQMF = LENGTH_QMF; move16(); // store 'number of coefficients'
  *dp_l = dp02_l; move16();          // store pointer to dp02[]-array
  *dp_h = dp02_h; move16();
}



#undef f24
#undef LENGTH_QMF



/*---------------------------------------------------------------------------
 * FUNCTION NAME: fir_initialization_B
 *
 * PURPOSE: Initialize memory for QMF filter on 16k data
 *
 * INPUT:
 *  - lengthQMF, QMF filter length
 *  - dp_h, pointer to HP filter
 *  - dp_l, pointer to LP filter
 *
 * OUTPUT: Allocate memory for filter and assign to appropriate pointer
 *           
 *                
 *
 * RETURN VALUE
 *  Pointer to QMF_FIR (QMF filter)
 *
 *---------------------------------------------------------------------------*/
static QMF_FIR *fir_initialization_B (X_INT32 lengthQMF,const X_INT16 dp_h[],const  X_INT16 dp_l[])
{
  QMF_FIR *ptrFIR; // pointer to the new struct
  int k;

  // ......... ALLOCATION OF MEMORY .........
  // Allocate memory for a new struct
  ptrFIR      = (QMF_FIR *) malloc (sizeof (QMF_FIR));
  // Allocate memory for delay line

  // Allocate memory for impulse response
  ptrFIR->T   = (X_INT16 *)     malloc ((lengthQMF - 1+2*FRAME_SHIFT) * sizeof (X_INT16));
  ptrFIR->dp_h  = (X_INT16 *)   malloc ((lengthQMF/2) * sizeof (X_INT16));
  ptrFIR->dp_l  = (X_INT16 *)   malloc ((lengthQMF/2 - L_H) * sizeof (X_INT16));
  ptrFIR->T_dec = 16;                                 move16();

  // ......... STORE VARIABLES INTO STATE VARIABLE .........
  // Store number of FIR-coefficients
  ptrFIR->lengthQMF = lengthQMF;                      move32();
  // Fill FIR coefficients into struct
  for (k=0 ; k < ptrFIR->lengthQMF/2 ; k++)
    { ptrFIR->dp_h[k] = dp_h[k]; move16(); }
  // Fill FIR coefficients into struct
  for (k = 0; k < ptrFIR->lengthQMF/2 - L_H; k++)
    { ptrFIR->dp_l[k] = dp_l[k]; move16(); }
  // Clear Delay Line
  for (k = 0; k < ptrFIR->lengthQMF - 1; k++)
    { ptrFIR->T[k] = 0; move32(); }

  return (ptrFIR);
}



/*---------------------------------------------------------------------------
 * FUNCTION NAME: QMF_FIR_Init_B
 *
 * PURPOSE: Wrapper code for QMF filter initialization allocates 
 *	memory and initialize pointers
 *
 * INPUT: NONE
 *  
 *
 * OUTPUT: NONE
 *           
 *
 * RETURN VALUE
 *   - QMF_FIR with initialized pointers
 *
 *---------------------------------------------------------------------------*/
static QMF_FIR * QMF_FIR_Init_B (void)
{
  const X_INT16 *dp_h; const X_INT16 *dp_l;        // pointer to array with FIR coeff.
  X_INT32 lengthQMF;       // number of FIR coefficients

  // allocate array for FIR coeff. and fill with coefficients
  DP_HP_filters_B (&dp_h, &dp_l, &lengthQMF);
  
  // Returns: pointer to QMF_FIR-struct
  return fir_initialization_B (lengthQMF, dp_h, dp_l);
}



/*---------------------------------------------------------------------------
 * FUNCTION NAME: Do16kProcessing_B
 *
 * PURPOSE: Apply QMF filter to break 8k and 16k data (LP & HP)
 *
 * INPUT:
 *  - *inData_B, Input data (16k)
 *  - *pData16k, Pointer to data related to 16k processing
 *  - inDataLength, input data array length
 *
 * OUTPUT
 *  - *inData_B, Output data (lower 0-8k)
 *  - *pData16k->DataHP_B, output data (upper 8-16k)
 *
 * RETURN VALUE
 *  NULL
 *
 *---------------------------------------------------------------------------*/

void Do16kProcessing_B (X_INT16 *inData_B, DataFor16kProc_B * pData16k, X_INT32 inDataLength)
{
  QMF_FIR * pQMF_Fir = pData16k->pQMF_Fir;

  int i, j, k;
  X_INT32 hpSign;

  X_INT32 filterLength;

  X_INT32 outDataDP,outDataHP;
  X_INT16 *prevData;
  X_INT32 aux2;
  X_INT16 *filterDP_h;
  X_INT16 *filterDP_l;

  X_INT16 max = 1;

  X_INT16 scale = 0;
  X_INT16 prevscale = pQMF_Fir->T_dec;
  X_INT16 shift=0;

  move16(); move16(); move16(); move16();

    for(i=0; i<inDataLength; i++) {
      test(); abs_s(0); 
      if(sub(max,abs(inData_B[i]))<0) { //if(max<abs(inData_B[i]))
	max = abs((int)inData_B[i]);                    abs_s(0); 
      }
    }
    test(); logic16();
    while(!(max & 0xc000u))
      {
	scale = add(scale,1);
	max = shl(max,1);

	test(); logic16();
      }

    test(); move16();
    if(sub(prevscale,scale)>0)	//if(prevscale>scale)
      shift = scale;
    else   shift = prevscale;

    prevData = pQMF_Fir->T; move16();
    filterDP_h = pQMF_Fir->dp_h; move16();
    filterDP_l = pQMF_Fir->dp_l; move16();
    filterLength = pQMF_Fir->lengthQMF; move16();
    {
		for (i=0 ; i<inDataLength ; i++)
		{ 
			prevData [filterLength - 1 + i] = inData_B [i]; move16(); 
		}	//Q0
    }



  // FIR filtering
    hpSign = 1; move16();

  for (i=0, k=0 ; i<inDataLength ; i+=2, k++)
  {
      X_INT32 aux3e =0 ,  aux3o = 0 ;
      X_INT32 aux5e =0 ,  aux5o = 0 ;
      // computing only every second sample
      X_INT32 aux1e = 0 ;
      X_INT32 aux1o = 0 ;

      move32(); move32(); move32(); move32(); move32(); move32(); 

      aux2 = 0;                                         move32(); 
      for (j=0 ; j< L_H ; j+=2)
	  {
		  X_INT16 low1 = prevData[i+j] ;
		  X_INT16 high2 = filterDP_h[j] ;
		  move16(); move16();
		  aux1o = L_mac(aux1o,low1,high2);
		  low1 = prevData[i+filterLength-j-1] ; move16();
		  aux1e = L_mac(aux1e,low1,high2);
		  high2 = filterDP_h[j+1] ; move16();
		  low1 = prevData[i+j+1] ; move16();
		  aux1e = L_mac(aux1e,low1,high2);
		  low1 = prevData[i+filterLength-j-2] ; move16();
		  aux1o = L_mac(aux1o,low1,high2);
	  }
      for (j= L_H ; j<(filterLength/2 -1) ; j+=2 )
	  {
		  X_INT16 high2 = filterDP_h[j]  ;
		  X_INT16 low2 =  filterDP_l[j - L_H] ;
		  X_INT16 low1 = prevData[i+j] ;
		  move16(); move16(); move16();
		  aux3o = L_mac(aux3o,low1,high2);
		  aux5o = L_mac(aux5o,low1,low2);
		  low1 = prevData[i+filterLength-j-1] ; move16();
		  aux3e = L_mac(aux3e,low1,high2);
		  aux5e = L_mac(aux5e,low1,low2);
		  high2 = filterDP_h[j+1]; move16();
		  low2 =  filterDP_l[j+1 - L_H] ; move16();
		  low1 = prevData[i+j+1] ; move16();
		  aux3e = L_mac(aux3e,low1,high2);
		  aux5e = L_mac(aux5e,low1,low2);
		  low1 = prevData[i+filterLength-j-2] ; move16();
		  aux3o = L_mac(aux3o,low1,high2);
		  aux5o = L_mac(aux5o,low1,low2);
	  }
	  {
		  X_INT16 high2 = filterDP_h[j] ;
		  X_INT16 low2 =  filterDP_l[j-L_H] ;
		  X_INT16 low1 = prevData[i+j] ;
		  move16(); move16(); move16();
		  aux3o = L_mac(aux3o,low1,high2);
		  aux5o = L_mac(aux5o,low1,low2);
		  low1 = prevData[i+j+1] ; move16();
		  aux3e = L_mac(aux3e,low1,high2);
		  aux5e = L_mac(aux5e,low1,low2);
	  }
      /* 
	 only 14-shift and 14-SHFF_H-shift are used -
         not the critical path for the stack memory
         assessment, therefore, neglect the memory -
         take into account only the wMOPs
       */
      sub(14,shift); move16(); sub(14-SHFF_H,shift); move16(); 

	  outDataDP = L_shr_r(L_add(aux1e,aux1o),14-shift) ;
	  outDataDP = L_add(outDataDP,L_shr_r(L_add(aux3e,aux3o),14-SHFF_H-shift));
	  outDataDP = L_add(outDataDP, L_shr_r(L_add(aux5e,aux5o),14-shift));
	  
	  outDataHP = L_shr_r(L_sub(aux1e,aux1o),14-shift) ;
	  outDataHP = L_add(outDataHP,L_shr_r(L_sub(aux3e,aux3o),14-SHFF_H-shift));
	  outDataHP = L_add(outDataHP, L_shr_r(L_sub(aux5e,aux5o),14-shift));
	  test();
	  if(hpSign > 0) {
	      pData16k->dataHP_B [k] = outDataHP; move32(); //8
	  } else {
	      pData16k->dataHP_B [k] = L_negate(outDataHP); move32(); //8

	  }
	  
	  hpSign = negate(hpSign);

	  sub(6,shift);
      inData_B [k] = round(L_shl(outDataDP,6-shift)) ; move16();	//8

  }

  {
      for (i=0 ; i<filterLength-1 ; i++)
	  { 
		  prevData [i] = prevData [inDataLength + i]; move16(); 
		  
	  }
  }
  
  
  prevscale = scale; move16();
  pData16k->dec_16k[0] = pData16k->dec_16k[1]; move16();
  pData16k->dec_16k[1] = pData16k->dec_16k[2]; move16();
  pData16k->dec_16k[2] = pData16k->dec_16k[3]; move16();
  pData16k->dec_16k[3] = pData16k->dec_16k[4]; move16();
  pData16k->dec_16k[4] = shift; move16();
  pQMF_Fir->T_dec = scale; move16();
}



/*---------------------------------------------------------------------------
 * FUNCTION NAME: hq_free_B
 *
 * PURPOSE: Frees memory allocated for QMF filtering
 *
 * INPUT:
 *  - *fir_ptr, QMF_FIR pointer
 *   
 *
 * OUTPUT: Frees memory allcated in *fir_ptr
 *           
 *                
 *
 * RETURN VALUE
 *  NULL
 *
 *---------------------------------------------------------------------------*/
static void hq_free_B (QMF_FIR *fir_ptr)
{
  if (fir_ptr->T != NULL) free (fir_ptr->T);  // free state variables
  if (fir_ptr->dp_h)        free (fir_ptr->dp_h); // free state impulse response
  if (fir_ptr->dp_l)        free (fir_ptr->dp_l); // free state impulse response
  if (fir_ptr)            free (fir_ptr);     // free allocated struct
}




/*---------------
 * get functions 
 *---------------*/

/*---------------------------------------------------------------------------
 * PURPOSE: Following functions are used to access data 
 *	from DataFor16kProc_B structure
 *
 * INPUT: *pData16k
 *
 * OUTPUT: NONE
 *
 * RETURN VALUE:  value that was requested
 *---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------
 * FUNCTION NAME: Get16k_hpBandsSize_B
 *
 *---------------------------------------------------------------------------*/
X_INT32 Get16k_hpBandsSize_B (DataFor16kProc_B *pData16k)
{
  return pData16k->hpBandsSize;
}


/*---------------------------------------------------------------------------
 * FUNCTION NAME: Get16k_p_hpBands_B
 *
 *---------------------------------------------------------------------------*/
X_INT32 *Get16k_p_hpBands_B (DataFor16kProc_B *pData16k)
{
  return pData16k->hpBands_B;
}


/*---------------------------------------------------------------------------
 * FUNCTION NAME: Get16k_p_bufferCodeForBands16k_B
 *
 *---------------------------------------------------------------------------*/
X_INT32 *Get16k_p_bufferCodeForBands16k_B (DataFor16kProc_B *pData16k)
{
  return pData16k->bufferCodeForBands16k_B;
}


/*---------------------------------------------------------------------------
 * FUNCTION NAME: Get16k_p_CodeForBands16k_B
 *
 *---------------------------------------------------------------------------*/

X_INT32 *Get16k_p_CodeForBands16k_B (DataFor16kProc_B *pData16k)
{
  return pData16k->CodeForBands16k_B;
}


/*---------------------------------------------------------------------------
 * FUNCTION NAME: Get16k_p_bufferCodeWeights_B
 *
 *---------------------------------------------------------------------------*/

X_INT16 *Get16k_p_bufferCodeWeights_B (DataFor16kProc_B *pData16k)
{
  return pData16k->bufferCodeWeights_B;
}


/*---------------------------------------------------------------------------
 * FUNCTION NAME: Get16k_p_codeWeights_B
 *
 *---------------------------------------------------------------------------*/

X_INT16 *Get16k_p_codeWeights_B (DataFor16kProc_B *pData16k)
{
  return pData16k->codeWeights_B;
}



/*---------------------------------------------------------------------------
 * FUNCTION NAME: Get16k_p_bufferData16k_B
 *
 *---------------------------------------------------------------------------*/
X_INT32 *Get16k_p_bufferData16k_B (DataFor16kProc_B *pData16k)
{
  return pData16k->bufferData16k_B;
}


/*---------------------------------------------------------------------------
 * FUNCTION NAME: Get16k_bufData16kSize_B
 *
 *---------------------------------------------------------------------------*/
X_INT32 Get16k_bufData16kSize_B (DataFor16kProc_B *pData16k)
{
  return pData16k->bufData16kSize;
}


/*---------------------------------------------------------------------------
 * FUNCTION NAME: Get16k_p_BandsForCoding16k_B
 *
 *---------------------------------------------------------------------------*/
X_INT32 *Get16k_p_BandsForCoding16k_B (DataFor16kProc_B *pData16k)
{
  return pData16k->BandsForCoding16k_B;
}


/*---------------------------------------------------------------------------
 * FUNCTION NAME: Get16k_dataHP_B
 *
 *---------------------------------------------------------------------------*/
X_INT32 Get16k_dataHP_B (DataFor16kProc_B *pData16k, int i)
{
  return  pData16k->dataHP_B[i];
}


/*---------------------------------------------------------------------------
 * FUNCTION NAME: Get16k_dec1
 *
 * i= 1,2,3,4
 *---------------------------------------------------------------------------*/
X_INT16 Get16k_dec1 (DataFor16kProc_B *pData16k)
{
	return pData16k->dec_16k[0];
}
X_INT16 Get16k_dec2 (DataFor16kProc_B *pData16k)
{
	return pData16k->dec_16k[1];
}
X_INT16 Get16k_dec3 (DataFor16kProc_B *pData16k)
{
	return pData16k->dec_16k[2];
}
X_INT16 Get16k_dec4 (DataFor16kProc_B *pData16k)
{
	return pData16k->dec_16k[3];
}
X_INT16 Get16k_dec5 (DataFor16kProc_B *pData16k)
{
	return pData16k->dec_16k[4];
}


/*---------------------------------------------------------------------------
 * FUNCTION NAME: Get16k_BFC_dec_B
 *
 *---------------------------------------------------------------------------*/
X_INT16 * Get16k_BFC_dec_B (DataFor16kProc_B *pData16k)
{
  return pData16k->BFC_dec;
}


/*---------------------------------------------------------------------------
 * FUNCTION NAME: Get16k_hpBands_dec_B
 *
 *---------------------------------------------------------------------------*/
X_INT16  Get16k_hpBands_dec_B (DataFor16kProc_B *pData16k)
{
  return pData16k->fb16k_dec[2];
}


/*---------------------------------------------------------------------------
 * FUNCTION NAME: Set16k_hpBands_dec_B
 *
 * PURPOSE: Shift multiplier through 16k data multiplier array
 *
 * INPUT:
 * - *pData16k, pointer to 16k data structure
 * - decal, multiplier for the first array
 *
 * OUTPUT:
 * - value of multiplier array shift from [0]->[2] and new value is placed at [0]
 *
 * RETURN VALUE:
 * NULL
 *
 *---------------------------------------------------------------------------*/
void Set16k_hpBands_dec_B (DataFor16kProc_B *pData16k, X_INT16 decal)
{
  pData16k->fb16k_dec[2] = pData16k->fb16k_dec[1]; move16();
  pData16k->fb16k_dec[1] = pData16k->fb16k_dec[0]; move16();
  pData16k->fb16k_dec[0] = decal; move16();
}

#endif //defined(F_16KHZ)
