/*=======================================================================================================*/
/*      Distributed Speech Recognition                                                                   */
/*      ETSI ES 202 050 Advanced Front-End Feature Extraction Algorithm & Compression Algorithm &        */
/*      ETSI ES 202 212 Extended Advanced Front-End Feature Extraction Algorithm & Compression Algorithm */  
/*      Bit exact C-language software implementation using ETSI basic operations library                 */
/*      3GPP TS 26.243 v 6.1.0 November 2004                                                                  */
/*      ANSI C code for the Fixed-Point Distributed Speech Recognition Extended Advanced Front-end       */
/*=======================================================================================================*/

#include "ff4nrFix16_B.h"

#include "typedef.h"
#include "basic_op.h"
#include "count.h"


#define SIZE_SINE              64

const X_INT16   tabSin[SIZE_SINE] = {
         (X_INT16)        0         /* 32768 * sin(2*PI *    0 / 512)        */
        ,(X_INT16)    12540         /* 32768 * sin(2*PI *   32 / 512)        */
        ,(X_INT16)     6393         /* 32768 * sin(2*PI *   16 / 512)        */
        ,(X_INT16)    18205         /* 32768 * sin(2*PI *   48 / 512)        */
        ,(X_INT16)     3212         /* 32768 * sin(2*PI *    8 / 512)        */
        ,(X_INT16)    15447         /* 32768 * sin(2*PI *   40 / 512)        */
        ,(X_INT16)     9512         /* 32768 * sin(2*PI *   24 / 512)        */
        ,(X_INT16)    20788         /* 32768 * sin(2*PI *   56 / 512)        */
        ,(X_INT16)     1608         /* 32768 * sin(2*PI *    4 / 512)        */
        ,(X_INT16)    14010         /* 32768 * sin(2*PI *   36 / 512)        */
        ,(X_INT16)     7962         /* 32768 * sin(2*PI *   20 / 512)        */
        ,(X_INT16)    19520         /* 32768 * sin(2*PI *   52 / 512)        */
        ,(X_INT16)     4808         /* 32768 * sin(2*PI *   12 / 512)        */
        ,(X_INT16)    16846         /* 32768 * sin(2*PI *   44 / 512)        */
        ,(X_INT16)    11039         /* 32768 * sin(2*PI *   28 / 512)        */
        ,(X_INT16)    22006         /* 32768 * sin(2*PI *   60 / 512)        */
        ,(X_INT16)      804         /* 32768 * sin(2*PI *    2 / 512)        */
        ,(X_INT16)    13279         /* 32768 * sin(2*PI *   34 / 512)        */
        ,(X_INT16)     7180         /* 32768 * sin(2*PI *   18 / 512)        */
        ,(X_INT16)    18868         /* 32768 * sin(2*PI *   50 / 512)        */
        ,(X_INT16)     4011         /* 32768 * sin(2*PI *   10 / 512)        */
        ,(X_INT16)    16151         /* 32768 * sin(2*PI *   42 / 512)        */
        ,(X_INT16)    10279         /* 32768 * sin(2*PI *   26 / 512)        */
        ,(X_INT16)    21403         /* 32768 * sin(2*PI *   58 / 512)        */
        ,(X_INT16)     2411         /* 32768 * sin(2*PI *    6 / 512)        */
        ,(X_INT16)    14733         /* 32768 * sin(2*PI *   38 / 512)        */
        ,(X_INT16)     8740         /* 32768 * sin(2*PI *   22 / 512)        */
        ,(X_INT16)    20160         /* 32768 * sin(2*PI *   54 / 512)        */
        ,(X_INT16)     5602         /* 32768 * sin(2*PI *   14 / 512)        */
        ,(X_INT16)    17531         /* 32768 * sin(2*PI *   46 / 512)        */
        ,(X_INT16)    11793         /* 32768 * sin(2*PI *   30 / 512)        */
        ,(X_INT16)    22595         /* 32768 * sin(2*PI *   62 / 512)        */
        ,(X_INT16)      402         /* 32768 * sin(2*PI *    1 / 512)        */
        ,(X_INT16)    12910         /* 32768 * sin(2*PI *   33 / 512)        */
        ,(X_INT16)     6787         /* 32768 * sin(2*PI *   17 / 512)        */
        ,(X_INT16)    18538         /* 32768 * sin(2*PI *   49 / 512)        */
        ,(X_INT16)     3612         /* 32768 * sin(2*PI *    9 / 512)        */
        ,(X_INT16)    15800         /* 32768 * sin(2*PI *   41 / 512)        */
        ,(X_INT16)     9896         /* 32768 * sin(2*PI *   25 / 512)        */
        ,(X_INT16)    21097         /* 32768 * sin(2*PI *   57 / 512)        */
        ,(X_INT16)     2009         /* 32768 * sin(2*PI *    5 / 512)        */
        ,(X_INT16)    14373         /* 32768 * sin(2*PI *   37 / 512)        */
        ,(X_INT16)     8351         /* 32768 * sin(2*PI *   21 / 512)        */
        ,(X_INT16)    19841         /* 32768 * sin(2*PI *   53 / 512)        */
        ,(X_INT16)     5205         /* 32768 * sin(2*PI *   13 / 512)        */
        ,(X_INT16)    17190         /* 32768 * sin(2*PI *   45 / 512)        */
        ,(X_INT16)    11417         /* 32768 * sin(2*PI *   29 / 512)        */
        ,(X_INT16)    22302         /* 32768 * sin(2*PI *   61 / 512)        */
        ,(X_INT16)     1206         /* 32768 * sin(2*PI *    3 / 512)        */
        ,(X_INT16)    13646         /* 32768 * sin(2*PI *   35 / 512)        */
        ,(X_INT16)     7571         /* 32768 * sin(2*PI *   19 / 512)        */
        ,(X_INT16)    19195         /* 32768 * sin(2*PI *   51 / 512)        */
        ,(X_INT16)     4410         /* 32768 * sin(2*PI *   11 / 512)        */
        ,(X_INT16)    16500         /* 32768 * sin(2*PI *   43 / 512)        */
        ,(X_INT16)    10660         /* 32768 * sin(2*PI *   27 / 512)        */
        ,(X_INT16)    21706         /* 32768 * sin(2*PI *   59 / 512)        */
        ,(X_INT16)     2811         /* 32768 * sin(2*PI *    7 / 512)        */
        ,(X_INT16)    15091         /* 32768 * sin(2*PI *   39 / 512)        */
        ,(X_INT16)     9127         /* 32768 * sin(2*PI *   23 / 512)        */
        ,(X_INT16)    20475         /* 32768 * sin(2*PI *   55 / 512)        */
        ,(X_INT16)     5998         /* 32768 * sin(2*PI *   15 / 512)        */
        ,(X_INT16)    17869         /* 32768 * sin(2*PI *   47 / 512)        */
        ,(X_INT16)    12167         /* 32768 * sin(2*PI *   31 / 512)        */
        ,(X_INT16)    22884         /* 32768 * sin(2*PI *   63 / 512)        */
        
} ;
const X_INT16   tabCos[SIZE_SINE] = {
         (X_INT16)    32767         /* 32768 * cos(2*PI *    0 / 512)        */
        ,(X_INT16)    30274         /* 32768 * cos(2*PI *   32 / 512)        */
        ,(X_INT16)    32138         /* 32768 * cos(2*PI *   16 / 512)        */
        ,(X_INT16)    27246         /* 32768 * cos(2*PI *   48 / 512)        */
        ,(X_INT16)    32610         /* 32768 * cos(2*PI *    8 / 512)        */
        ,(X_INT16)    28899         /* 32768 * cos(2*PI *   40 / 512)        */
        ,(X_INT16)    31357         /* 32768 * cos(2*PI *   24 / 512)        */
        ,(X_INT16)    25330         /* 32768 * cos(2*PI *   56 / 512)        */
        ,(X_INT16)    32729         /* 32768 * cos(2*PI *    4 / 512)        */
        ,(X_INT16)    29622         /* 32768 * cos(2*PI *   36 / 512)        */
        ,(X_INT16)    31786         /* 32768 * cos(2*PI *   20 / 512)        */
        ,(X_INT16)    26320         /* 32768 * cos(2*PI *   52 / 512)        */
        ,(X_INT16)    32413         /* 32768 * cos(2*PI *   12 / 512)        */
        ,(X_INT16)    28106         /* 32768 * cos(2*PI *   44 / 512)        */
        ,(X_INT16)    30853         /* 32768 * cos(2*PI *   28 / 512)        */
        ,(X_INT16)    24279         /* 32768 * cos(2*PI *   60 / 512)        */
        ,(X_INT16)    32758         /* 32768 * cos(2*PI *    2 / 512)        */
        ,(X_INT16)    29957         /* 32768 * cos(2*PI *   34 / 512)        */
        ,(X_INT16)    31972         /* 32768 * cos(2*PI *   18 / 512)        */
        ,(X_INT16)    26791         /* 32768 * cos(2*PI *   50 / 512)        */
        ,(X_INT16)    32522         /* 32768 * cos(2*PI *   10 / 512)        */
        ,(X_INT16)    28511         /* 32768 * cos(2*PI *   42 / 512)        */
        ,(X_INT16)    31114         /* 32768 * cos(2*PI *   26 / 512)        */
        ,(X_INT16)    24812         /* 32768 * cos(2*PI *   58 / 512)        */
        ,(X_INT16)    32679         /* 32768 * cos(2*PI *    6 / 512)        */
        ,(X_INT16)    29269         /* 32768 * cos(2*PI *   38 / 512)        */
        ,(X_INT16)    31581         /* 32768 * cos(2*PI *   22 / 512)        */
        ,(X_INT16)    25833         /* 32768 * cos(2*PI *   54 / 512)        */
        ,(X_INT16)    32286         /* 32768 * cos(2*PI *   14 / 512)        */
        ,(X_INT16)    27684         /* 32768 * cos(2*PI *   46 / 512)        */
        ,(X_INT16)    30572         /* 32768 * cos(2*PI *   30 / 512)        */
        ,(X_INT16)    23732         /* 32768 * cos(2*PI *   62 / 512)        */
        ,(X_INT16)    32766         /* 32768 * cos(2*PI *    1 / 512)        */
        ,(X_INT16)    30118         /* 32768 * cos(2*PI *   33 / 512)        */
        ,(X_INT16)    32058         /* 32768 * cos(2*PI *   17 / 512)        */
        ,(X_INT16)    27020         /* 32768 * cos(2*PI *   49 / 512)        */
        ,(X_INT16)    32568         /* 32768 * cos(2*PI *    9 / 512)        */
        ,(X_INT16)    28707         /* 32768 * cos(2*PI *   41 / 512)        */
        ,(X_INT16)    31238         /* 32768 * cos(2*PI *   25 / 512)        */
        ,(X_INT16)    25073         /* 32768 * cos(2*PI *   57 / 512)        */
        ,(X_INT16)    32706         /* 32768 * cos(2*PI *    5 / 512)        */
        ,(X_INT16)    29448         /* 32768 * cos(2*PI *   37 / 512)        */
        ,(X_INT16)    31686         /* 32768 * cos(2*PI *   21 / 512)        */
        ,(X_INT16)    26078         /* 32768 * cos(2*PI *   53 / 512)        */
        ,(X_INT16)    32352         /* 32768 * cos(2*PI *   13 / 512)        */
        ,(X_INT16)    27897         /* 32768 * cos(2*PI *   45 / 512)        */
        ,(X_INT16)    30715         /* 32768 * cos(2*PI *   29 / 512)        */
        ,(X_INT16)    24008         /* 32768 * cos(2*PI *   61 / 512)        */
        ,(X_INT16)    32746         /* 32768 * cos(2*PI *    3 / 512)        */
        ,(X_INT16)    29792         /* 32768 * cos(2*PI *   35 / 512)        */
        ,(X_INT16)    31881         /* 32768 * cos(2*PI *   19 / 512)        */
        ,(X_INT16)    26557         /* 32768 * cos(2*PI *   51 / 512)        */
        ,(X_INT16)    32470         /* 32768 * cos(2*PI *   11 / 512)        */
        ,(X_INT16)    28311         /* 32768 * cos(2*PI *   43 / 512)        */
        ,(X_INT16)    30986         /* 32768 * cos(2*PI *   27 / 512)        */
        ,(X_INT16)    24548         /* 32768 * cos(2*PI *   59 / 512)        */
        ,(X_INT16)    32647         /* 32768 * cos(2*PI *    7 / 512)        */
        ,(X_INT16)    29086         /* 32768 * cos(2*PI *   39 / 512)        */
        ,(X_INT16)    31471         /* 32768 * cos(2*PI *   23 / 512)        */
        ,(X_INT16)    25583         /* 32768 * cos(2*PI *   55 / 512)        */
        ,(X_INT16)    32214         /* 32768 * cos(2*PI *   15 / 512)        */
        ,(X_INT16)    27467         /* 32768 * cos(2*PI *   47 / 512)        */
        ,(X_INT16)    30425         /* 32768 * cos(2*PI *   31 / 512)        */
        ,(X_INT16)    23453         /* 32768 * cos(2*PI *   63 / 512)        */
        
} ;

/* prototypes for local functions */
static void Permut(X_INT16 *b0,X_INT16 M, X_INT16 N) ;


/********************************************************************************/

/*----------------------------------------------------------------------------
 * FUNCTION NAME: ff4NR16_B
 *
 * PURPOSE: Inplace FFT for an INT16 real signal with size N = 2**M
 *              
 *
 * INPUT:
 *      - b   input real signal
 *      - M   Logarithm in base 2 of FFT length
 *
 * OUTPUT:
 *      - b  : FFT ouput  
 *   b[0] = Re[0]
 *   b[1] = Re[N/2]
 *   b[2] = Re[1]
 *   b[3] = -Im[1]
 *   ...
 *   b[2*k] = Re[k]
 *   b[2*k+1] = -Im[k]
 * Warnings :
 *  - The routines returns the opposite of the imaginary parts.
 *
 * RETURN VALUE: none
 *   
 *
 *---------------------------------------------------------------------------*/

void ff4NR16_B(
    int         M,  /* (i)   Logarithm in base 2 of FFT length */
    X_INT16     *b  /* (i/o) int16 signal array                */
)

{
    const static X_INT16 P7 = (X_INT16) (0.70710681 * 32768) ; /*  1 / SQRT(2)  */
    
    
    X_INT16   N ,ji,jl,jr,jthet ;
    X_INT16   intc ;
    X_INT16   *b0;
    
    
    b0 = b ;
    
    /*  Compute N = 2**M */
    N = (X_INT16) shl(1,M) ;
    
    /* RADIX 2 if necessary */
    test(); logic16();
    if(M & 1) {
        intc = (X_INT16) shr(N,1) ;
        {
            X_INT16 k ;
            X_INT16 * b1 = b0 + intc ;
            move16();
            for(k=0;k<intc;k++) {
                X_INT16 T0 = b0[k] ; move16();
                b0[k] = mac_r(L_mult(T0,0x4000),b1[k],0x4000) ; move16();
                b1[k] = msu_r(L_mult(T0,0x4000),b1[k],0x4000) ; move16();
            }
        }
    } else {
        intc = N ; move16() ;
    }
    /* loop for  RADIX 4 */
    test(); sub(intc, 1);
    while (intc > 1) {
        X_INT16 * b1, *b2 , *b3 ;
        X_INT16 k ;
        intc = shr(intc,2) ;
        b1 = b0 + intc ;move16();
        b2 = b1 + intc ;move16();
        b3 = b2 + intc ;move16();
        /* Simple radix  BFYR0 */
        for (k = 0 ; k < intc ; k++){
            X_INT32 T0 ;
            X_INT16 bt3 ;
            T0 = L_mac(L_mult(b0[k],0x4000),b2[k],0x4000) ;
            b2[k] = msu_r(L_mult(b0[k],0x4000),b2[k],0x4000) ; move16();
            b0[k] = round(L_mac(L_mac(T0,b1[k],0x4000),b3[k],0x4000)) ; move16();
            bt3 = msu_r(L_mult(b1[k],0x4000),b3[k],0x4000) ; 
            b1[k] = round(L_msu(L_msu(T0,b1[k],0x4000),b3[k],0x4000)) ; move16();
            b3[k] = bt3 ; move16();

        }
        /* BFYI0 if necessary */
        test();
        if ( sub(N,shl(intc,2)) > 0) {
            /* k<add(intc,shl(intc,2)) can be used instead */
            for(k = shl(intc,2); sub(k,add(intc,shl(intc,2))) < 0 ; k++) {
                X_INT32 PR,PII ;
                PR = msu_r(L_mult(P7,b1[k]),P7,b3[k]);
                PII =mac_r(L_mult(P7,b1[k]),P7,b3[k]);
                b3[k] = mac_r(L_mult(b2[k],0x4000),PII,0x4000) ; move16();
                b1[k] = msu_r(L_mult(PII,0x4000),b2[k],0x4000) ; move16();
                b2[k] = msu_r(L_mult(b0[k],0x4000),PR,0x4000) ; move16();
                b0[k] = mac_r(L_mult(b0[k],0x4000),PR,0x4000) ; move16();
            }
        }
        jl = shl(intc,3); 
        ji = add(jl,shl(intc,2)); 
        jr = jl ; move16();
        jthet = 1 ; move16();

        /* loop for Radix4 BFY */
        test(); sub(jr, N);
        while (jr < N ) {
            X_INT16 j ;
            X_INT16   c1,c2,c3 ;
            X_INT16   s1,s2,s3 ;
            s1 = tabSin[jthet] ; move16();
            c1 = tabCos[jthet] ; move16();
            c2 = round(L_msu(L_mult(c1,c1),s1,s1)) ; 
            s2 = round(L_mac(L_mult(s1,c1),c1,s1)) ;
            c3 = round(L_msu(L_mult(c1,c2),s1,s2)) ;
            s3 = round(L_mac(L_mult(s1,c2),c1,s2)) ;

            k = sub(ji,jr) ; 
            for( j = jr ; j < add(jr,intc) ; j++) {
                
                X_INT32   T0,T1,T2,T3,T4,T5,T6,T7 ;
                X_INT32   TR1,TR2,TR3 ;
                X_INT32   TI1,TI2,TI3 ;

                TR1 = L_shr( L_msu( L_mult(b1[j],c1),b1[j+k],s1),1);
                TI1 = L_shr( L_mac( L_mult(b1[j],s1),b1[j+k],c1),1);
                TR2 = L_shr( L_msu( L_mult(b2[j],c2),b2[j+k],s2),1);
                TI2 = L_shr( L_mac( L_mult(b2[j],s2),b2[j+k],c2),1);
                TR3 = L_shr( L_msu( L_mult(b3[j],c3),b3[j+k],s3),1);
                TI3 = L_shr( L_mac( L_mult(b3[j],s3),b3[j+k],c3),1);

                T0 = L_mac(TR2,b0[j],0x4000) ;
                T4 = L_mac(TI2,b0[j+k],0x4000) ;
                T2 = L_msu(TR2,b0[j],0x4000) ; 
                T6 = L_msu(TI2,b0[j+k],0x4000) ; 
                T1 = L_add(TR1,TR3);
                T5 = L_add(TI1,TI3);
                T3 = L_sub(TR1,TR3);
                T7 = L_sub(TI1,TI3);


                b0[j] =   round(L_add(T0,T1)); move16();
                b3[j+k] = round(L_add(T4,T5)); move16();
                b2[j+k] = round(L_sub(T0,T1)); move16();
                b1[j] =   round(L_sub(T5,T4)); move16();
                b2[j] =   negate(round(L_add(T2,T7))); move16();
                b1[j+k] = round(L_sub(T3,T6)); move16();
                b0[j+k] = round(L_sub(T7,T2)); move16();
                b3[j] =   round(L_add(T3,T6)); move16();

            }

            jthet = add(jthet,1) ;
            jr = add(jr,shl(intc,3)) ;
            ji = sub(ji,shl(intc,3)) ;
            test();
            if ( sub(ji,jl)<0 ) {
                ji = shl(sub(jr,shl(intc,1)),1); 
                jl = jr ; move16();
            }
            test(); sub(jr, N); /* while */
        }
        test(); sub(intc, 1);/* while */
    }
    Permut(b0,(X_INT16)M,N);
    return ;

 }
 
 
 
 /*********************** local functions ******************************************************/

#define Perm_D(i)    tmp = b0[i] ; move16()
#define Perm(i,j)    b0[i] = b0[j] ; move16() 
#define Perm_F(i)    b0[i] = tmp ; move16()


/*----------------------------------------------------------------------------
 * FUNCTION NAME: Permut
 *
 * PURPOSE: Permutation of imaginary parts, and implements bit reverse
 *              
 *
 * INPUT:
 *      - b0 : input butterfly result
 *      - N : FFT length
 *      - M : Logarithm in base 2 of FFT length  (N = 2**M)
 *
 * OUTPUT:
 *      - b0  : FFT output  
 *
 * RETURN VALUE: none
 *   
 *
 *---------------------------------------------------------------------------*/
static void Permut(X_INT16 *b0,X_INT16 M, X_INT16 N) {
    X_INT16   jr,k ,ij,ji;
    test();
    if(sub(M,8)==0) {
    /* Tabulated permutations for N = 2**8 */
        X_INT16 tmp ;
        Perm_D(2) ;Perm(2,128) ; Perm_F(128) ;
        Perm_D(3) ;Perm(3,255) ; Perm(255,129) ; Perm_F(129) ;
        Perm_D(4) ;Perm(4,64) ; Perm_F(64) ;
        Perm_D(5) ;Perm(5,127) ; Perm(127,131) ; Perm(131,253) ; Perm(253,65) ; Perm(65,7) ;
        Perm(7,191) ; Perm(191,133) ; Perm(133,125) ; Perm(125,67) ; Perm(67,251) ; Perm(251,193) ; Perm_F(193) ;
        Perm_D(6) ;Perm(6,192) ; Perm_F(192) ;
        Perm_D(8) ;Perm(8,32) ; Perm_F(32) ;
        Perm_D(9) ;Perm(9,63) ; Perm(63,135) ; Perm(135,189) ; Perm(189,69) ; Perm(69,123) ;
        Perm(123,195) ; Perm(195,249) ; Perm(249,33) ; Perm(33,15) ; Perm(15,159) ; 
        Perm(159,141) ; Perm(141,93) ; Perm(93,75) ; Perm(75,219) ; Perm(219,201) ;
        Perm(201,57) ; Perm(57,39) ; Perm(39,183) ; Perm(183,165) ; Perm(165,117) ;
        Perm(117,99) ; Perm(99,243) ; Perm(243,225) ; Perm_F(225) ;
        Perm_D(10) ;Perm(10,160) ; Perm_F(160) ;
        Perm_D(11) ;Perm(11,223) ; Perm(223,137) ; Perm(137,61) ; Perm(61,71) ; Perm(71,187) ;
        Perm(187,197) ; Perm(197,121) ; Perm(121,35) ; Perm(35,247) ; Perm(247,161) ;
        Perm(161,13) ; Perm(13,95) ; Perm(95,139) ; Perm(139,221) ; Perm(221,73) ; 
        Perm(73,59) ; Perm(59,199) ; Perm(199,185) ; Perm(185,37) ; Perm(37,119) ; 
        Perm(119,163) ; Perm(163,245) ; Perm(245,97) ; Perm_F(97) ;
        Perm_D(12) ;Perm(12,96) ; Perm_F(96) ;
        Perm_D(14) ;Perm(14,224) ; Perm_F(224) ;
        Perm_D(17) ;Perm(17,31) ; Perm(31,143) ; Perm(143,157) ; Perm(157,77) ; Perm(77,91) ;
        Perm(91,203) ; Perm(203,217) ; Perm(217,41) ; Perm(41,55) ; Perm(55,167) ; 
        Perm(167,181) ; Perm(181,101) ; Perm(101,115) ; Perm(115,227) ; Perm(227,241) ;
        Perm_F(241) ;
        Perm_D(18) ;Perm(18,144) ; Perm_F(144) ;
        Perm_D(19) ;Perm(19,239) ; Perm(239,145) ; Perm(145,29) ; Perm(29,79) ; Perm(79,155) ;
        Perm(155,205) ; Perm(205,89) ; Perm(89,43) ; Perm(43,215) ; Perm(215,169) ; 
        Perm(169,53) ; Perm(53,103) ; Perm(103,179) ; Perm(179,229) ; Perm(229,113) ;
        Perm_F(113) ;
        Perm_D(20) ;Perm(20,80) ; Perm_F(80) ;
        Perm_D(21) ;Perm(21,111) ; Perm(111,147) ; Perm(147,237) ; Perm(237,81) ; Perm(81,27) ; 
        Perm(27,207) ; Perm(207,153) ; Perm(153,45) ; Perm(45,87) ; Perm(87,171) ; 
        Perm(171,213) ; Perm(213,105) ; Perm(105,51) ; Perm(51,231) ; Perm(231,177) ;
        Perm_F(177) ;
        Perm_D(22) ;Perm(22,208) ; Perm_F(208) ;
        Perm_D(23) ;Perm(23,175) ; Perm(175,149) ; Perm(149,109) ; Perm(109,83) ; 
        Perm(83,235) ; Perm(235,209) ; Perm(209,25) ; Perm(25,47) ; Perm(47,151) ;
        Perm(151,173) ; Perm(173,85) ; Perm(85,107) ; Perm(107,211) ; Perm(211,233) ;
        Perm(233,49) ; Perm_F(49) ;
        Perm_D(24) ;Perm(24,48) ; Perm_F(48) ;
        Perm_D(26) ;Perm(26,176) ; Perm_F(176) ;
        Perm_D(28) ;Perm(28,112) ; Perm_F(112) ;
        Perm_D(30) ;Perm(30,240) ; Perm_F(240) ;
        Perm_D(34) ;Perm(34,136) ; Perm_F(136) ;
        Perm_D(36) ;Perm(36,72) ; Perm_F(72) ;
        Perm_D(38) ;Perm(38,200) ; Perm_F(200) ;
        Perm_D(42) ;Perm(42,168) ; Perm_F(168) ;
        Perm_D(44) ;Perm(44,104) ; Perm_F(104) ;
        Perm_D(46) ;Perm(46,232) ; Perm_F(232) ;
        Perm_D(50) ;Perm(50,152) ; Perm_F(152) ;
        Perm_D(52) ;Perm(52,88) ; Perm_F(88) ;
        Perm_D(54) ;Perm(54,216) ; Perm_F(216) ;
        Perm_D(58) ;Perm(58,184) ; Perm_F(184) ;
        Perm_D(60) ;Perm(60,120) ; Perm_F(120) ;
        Perm_D(62) ;Perm(62,248) ; Perm_F(248) ;
        Perm_D(66) ;Perm(66,132) ; Perm_F(132) ;
        Perm_D(70) ;Perm(70,196) ; Perm_F(196) ;
        Perm_D(74) ;Perm(74,164) ; Perm_F(164) ;
        Perm_D(76) ;Perm(76,100) ; Perm_F(100) ;
        Perm_D(78) ;Perm(78,228) ; Perm_F(228) ;
        Perm_D(82) ;Perm(82,148) ; Perm_F(148) ;
        Perm_D(86) ;Perm(86,212) ; Perm_F(212) ;
        Perm_D(90) ;Perm(90,180) ; Perm_F(180) ;
        Perm_D(92) ;Perm(92,116) ; Perm_F(116) ;
        Perm_D(94) ;Perm(94,244) ; Perm_F(244) ;
        Perm_D(98) ;Perm(98,140) ; Perm_F(140) ;
        Perm_D(102) ;Perm(102,204) ; Perm_F(204) ;
        Perm_D(106) ;Perm(106,172) ; Perm_F(172) ;
        Perm_D(110) ;Perm(110,236) ; Perm_F(236) ;
        Perm_D(114) ;Perm(114,156) ; Perm_F(156) ;
        Perm_D(118) ;Perm(118,220) ; Perm_F(220) ;
        Perm_D(122) ;Perm(122,188) ; Perm_F(188) ;
        Perm_D(126) ;Perm(126,252) ; Perm_F(252) ;
        Perm_D(134) ;Perm(134,194) ; Perm_F(194) ;
        Perm_D(138) ;Perm(138,162) ; Perm_F(162) ;
        Perm_D(142) ;Perm(142,226) ; Perm_F(226) ;
        Perm_D(150) ;Perm(150,210) ; Perm_F(210) ;
        Perm_D(154) ;Perm(154,178) ; Perm_F(178) ;
        Perm_D(158) ;Perm(158,242) ; Perm_F(242) ;
        Perm_D(166) ;Perm(166,202) ; Perm_F(202) ;
        Perm_D(174) ;Perm(174,234) ; Perm_F(234) ;
        Perm_D(182) ;Perm(182,218) ; Perm_F(218) ;
        Perm_D(190) ;Perm(190,250) ; Perm_F(250) ;
        Perm_D(206) ;Perm(206,230) ; Perm_F(230) ;
        Perm_D(222) ;Perm(222,246) ; Perm_F(246) ;
    } else {
        /* Imaginary parts */
        jr = 2  ; move16();
        for(k = 2 ; k < M ; k++) {
            jr =shl(jr,1);
            ij = add(jr,1);
            ji = sub(shl(jr,1),1) ;
            test();
            while ( sub(ji,ij) > 0 ) {
                X_INT32 T = b0[ij] ; move16();
                b0[ij] = b0[ji] ; move16() ;
                b0[ji] = T ; move16(); 
                ij = add(ij,2);
                ji = sub(ji,2);
                test() ;
            }
        }
        /*   Bit reverse */
        ji = 0 ; move16();
        ij = 0 ; move16();
        test();
        while ( sub(ij,N) < 0 ) {
            test();
            if ( sub(ij,ji) < 0 ) {
                X_INT32 T = b0[ij] ; move16() ;
                b0[ij] = b0[ji] ;  move16() ;
                b0[ji] = T ; move16();
                T = b0[ij+1] ; move16() ;
                b0[ij+1] = b0[ji+1] ; move16() ;
                b0[ji+1] = T ; move16() ;
            }
            jr = shr(N,1);
            test();
            while ( sub(jr ,ji) <= 0 ) {
                ji = sub(ji,jr); 
                jr = shr(jr,1); 
                test();
            }
            ji = add(ji,jr); 
            ij = add(ij,2) ; 
            test();
       }
    } /* M != 8 */
    return ;
 }

 
 
