/*=======================================================================================================*/
/*      Distributed Speech Recognition                                                                   */
/*      ETSI ES 202 050 Advanced Front-End Feature Extraction Algorithm & Compression Algorithm &        */
/*      ETSI ES 202 212 Extended Advanced Front-End Feature Extraction Algorithm & Compression Algorithm */  
/*      Bit exact C-language software implementation using ETSI basic operations library                 */
/*      3GPP TS 26.243 v 6.1.0 November 2004                                                                  */
/*      ANSI C code for the Fixed-Point Distributed Speech Recognition Extended Advanced Front-end       */
/*=======================================================================================================*/

/*=======================================================================================================
*   PURPOSE:
*
*      Defines a number of general types, macros, and constants
*      used in pitch extraction
*
*=======================================================================================================
*/

#ifndef RVC_PITCH_B_H
#define RVC_PITCH_B_H

#include "ParmType.h"


/* =========================================================================
*
*                GENERAL TYPES, MACROS, AND CONSTANTS
*
* =========================================================================*/

/* 
 *Define boolean as an integer. 
 */

typedef int BOOL;


/*
 * Macro for memory allocation.
 *
 * The following macro will allocate memory and reset it with 0 values.
 *
 * Parameters:
 *
 * pVar		A pointer on which memory will be allocated
 * VarType	A string defining the pointer type, such as short, float, etc.
 * Number	An integer holding the number of fields of type VarType
 *		to be allocated.
 * iReturnCode	An enum integer of type RvcResult that on exit will hold
 *		RVC_OK if memory was successfully allocated and
 *		RVC_ERR_NOT_ENOUGH_MEMORY if memory allocation failed.
 */

#define MALLOCATE(pVar, VarType, Number, iReturnCode)  						\
	if ((VarType *)NULL==(pVar=(VarType *)malloc((Number)*sizeof(VarType)))) {	\
		iReturnCode = RVC_ERR_NOT_ENOUGH_MEMORY;				\
	} else {													\
		memset(pVar,0,(Number)*(sizeof(VarType)));					\
        iReturnCode = RVC_OK;                                      \
	}

/*
 * Macro for freeing the allocated memory
 * 
 * Parameters:
 *
 * var - The pointer to be freed
 *
 */

#define MDEALLOCATE(var) \
 if (NULL!=(var)) {\
  free(var);\
  var=NULL;\
 }



/*
 * Mathematical Macros
 */

#ifndef MIN
// Minimum
#define MIN(x,y)  (((x)>(y)) ? (y) :(x))
#endif

#ifndef MAX
// Maximum
#define MAX(x,y)  (((x)>(y)) ? (x) :(y))
#endif

#ifndef ABS
//Absolute
#define ABS(x)    (((x)<0) ? -(x) : (x))
#endif



/*
 * Complex structures
 */

#ifndef CMPLX_BE_DEFINED
#define CMPLX_BE_DEFINED

typedef struct CMPLX_be_t {
  Longword lwReal;
  Longword lwImag;
} CMPLX_be;

typedef struct CMPLXsw_be_t {
  Shortword swReal;
  Shortword swImag;
} CMPLXsw_be;

#endif



/* =========================================================================
*
*                      PITCH  RELATED CONSTANTS
*
* =========================================================================*/

/*
 * Dirichlet Interpolation Constants
 */
#define DIRICHLET_KERNEL_SPAN  8 // Length of Direchlet interpolation 


/*
 * Sampling rate of output pitch and some bandwidths
 */
#define REF_SAMPLE_RATE          8000.f 
#define REF_BANDWIDTH           (REF_SAMPLE_RATE/2.f) 
#define	lwTHIRD_REF_BANDWIDTH		 87381333
#define	lwTWO_THIRDS_REF_BANDWIDTH	174762667


/*
 * Constants for square DFT amp weighting
 * (under the constraint CENTER_WEIGHT+2*SIDE_WEIGHT=1)
 */
#define swCENTER_WEIGHT 0x5000
#define swSIDE_WEIGHT   0x1800


/*
 * Constants used to optionaly scale down the spectral peaks
 * amplitude in the second and third thirds of the bandwidth
 */
#define swAMP_SCALE_DOWN1 0x5333 
#define swAMP_SCALE_DOWN2 0x399A
#define swAMP_SCALE_DOWN2b 0x7333


/*
 * Constants related to the shape of the utility function
 *
 *   1                +------+------+
 *                    |             |
 *  USTEP     +-------+             +-------+
 *            |                             |
 *   +--------+                             +---------+  
 * -0.5   -UDIST2  -UDIST1   0   +UDIST1  +UDIST2   +0.5
 *    
 */
#define swUDIST1 (4160) 
#define swUDIST2 (6400)   
#define swUSTEP  (16384)



/*
 * The following constants are used to select the best pitch
 * from a list of local maxima in the utility function and
 * the pitch from the previous frame
 */
#define swFREQ_MARGIN1 0x4AE1
#define swAMP_MARGIN1  0x07AE
#define swAMP_MARGIN2  0x07AE


/*
 * Constants related to continuous pitch track
 */
#define MIN_STABLE_FRAMES 6  
#define MAX_TRACK_GAP_FRAMES 2

//Upper margin scaled to Q14
#define swSTABLE_FREQ_UPPER_MARGIN  0x4E14
#define swSTABLE_FREQ_LOWER_MARGIN  0x68EB


/*
 * Pitch frequency of an unvoiced frame
 */
#define UNVOICED 0


/*
 * Pitch Frequency Range (in Hz)
 */
#define lwMAX_PITCH_FREQ 0x01A40000L   // Maximum pitch frequency in Hz
#define lwMIN_PITCH_FREQ 0x00340000L   // Minimum pitch frequency in Hz

#define MAX_PITCH_FREQ 420.f   // Maximum pitch frequency in Hz
#define MIN_PITCH_FREQ  52.f   // Minimum pitch frequency in Hz


/*
 * High-pass cutoff frequency
 */
#define HIGHPASS_CUTOFF_FREQ   300.f 


/*
 * Number of fractions in the fractions table
 */
#define NO_OF_FRACS  (1+(int)(REF_BANDWIDTH/MIN_PITCH_FREQ)) 


/*
 *  The utility function is calculated 
 *  in the range [Fmin,F2] by peaks from the double window DFt,
 *  in the range [F1,  F4] by peaks from the single (regular) window DFT, and
 *  in the range [F3,Fmax] by peaks from the short window DFT.
 *  This means that there could be pitch candidates in the range 
 *  [F1,F2] or [F3,F4] originating from two differnt utility functions.
 *
 *                   +------------+   Short  window
 *          +--------------+          Single window
 *   +----------+                     Double window
 *  Fmin   F1   F2   F3   F4     Fmax
 *(=52 Hz)                      (=420 Hz)
 *
 */
#define lwSHORT_WIN_START_FREQ  0x00C80000L // = F3
#define lwSHORT_WIN_END_FREQ   (lwMAX_PITCH_FREQ) // = Fmax

#define lwSINGLE_WIN_START_FREQ 0x00640000L // = F1
#define lwSINGLE_WIN_END_FREQ   0x00D20000L // = F4

#define lwDOUBLE_WIN_START_FREQ (lwMIN_PITCH_FREQ) // = Fmin
#define lwDOUBLE_WIN_END_FREQ   0x00780000L //114.f // = F2


/*
 * The following constant limits computational complexity
 */
#define MAX_LOCAL_MAXIMA_ON_SPECTRUM 70

#define MAX_PEAKS_FOR_SORT  30
#define MAX_PEAKS_PRELIM  7
#define MIN_PEAKS  MAX_PEAKS_PRELIM   
#define MAX_PEAKS_FINAL  20

#define MAX_PRELIM_CANDS  4

#define CREATE_PIECEWISE_FUNC_LOOP_LIM_SH   20  
#define CREATE_PIECEWISE_FUNC_LOOP_LIM_SNG  30 
#define CREATE_PIECEWISE_FUNC_LOOP_LIM_DBL  60 

#define swSUM_FRACTION  0x799A
#define swAMP_FRACTION  0x33F8 

#define MAX_BEST_CANDS  2
#define N_OF_BEST_CANDS_SHORT  2
#define N_OF_BEST_CANDS_SINGLE 2
#define N_OF_BEST_CANDS_DOUBLE 2
#define N_OF_BEST_CANDS \
 (N_OF_BEST_CANDS_SHORT+N_OF_BEST_CANDS_SINGLE+N_OF_BEST_CANDS_DOUBLE)


/*
 * The following constants define scratch memory requirement
 */

/*
 * The acutal size required here is 1090 INT16 words. But because the C
 * compiler assumes 32-bit pointer variables and likes structures to be
 * on even-word boundaries, the size is defined as 1632 INT16 words.
 */

#define		SIZE_SCRATCH_DOPITCH		1632

/*
 * The actual size required here is 825 INT16 words. But because the C
 * compiler likes structures to be on even-word boundaries, the size
 * defined here is 1100 INT16 words.
 */

#define		SIZE_SCRATCH_ADVPROCESS		1100


 
/* ========================================================================
 *
 *                   PITCH RELATED TYPES
 *
 * ========================================================================*/

/*
 * To hold table of fractions defining the
 * location of each utility function component
 */ 
typedef struct {
 Shortword swFarLow; 
 Shortword swLow;    
 Shortword swHigh;   
 Shortword swFarHigh;
} FRACTIONS_be;

/*
 * To hold spectral peaks, utlity function edge-points
 * and pitch candidates with their utility function values
 */ 
typedef struct {
  Longword lwFreq;
  Shortword swAmp;
} POINT_be;

typedef struct {
  X_INT16   ind;
  Shortword swAmp;
} IPOINT_be;

typedef struct xpoint_be {
  Longword lwFreq;
  Shortword swAmp;
  struct xpoint_be *next;
} XPOINT_be;

typedef struct {
  XPOINT_be *addr;
  X_INT16 length;
} ARRAY_OF_XPOINTS_be;

typedef struct {
  Longword lwFreq;    		/* Q16 */
  Shortword swAmp;   		/* Q15 */
  Shortword swCorr;		/* Q15 */
} PITCH_be;


/*
 * PITCH_ROM and PITCH_METER structures used for pitch extraction
 */

#define RVC_PITCH_ROM_SIG  11031

typedef struct rvc_pitch_rom_tag_be
{
	X_INT16 signature;

	X_INT16 iSampleRate;
	X_INT16 iFrameSize;
	X_INT16 iWinSize;    
	X_INT16 iDftSize;    	 
	
	Shortword swPreEmph;

	X_INT16 i4KhzDftBin; // 4 Khz bin in input DFT
	X_INT16 iInterpDftLen;
	
	// To hold the inverse preemphasis filter table
	Shortword *pswInvPreEmpFilter; 
	
	FRACTIONS_be *astFrac;

	// To hold window shift table of complex exponents
	const CMPLXsw_be     *pstWindowShiftTable; 

	const Shortword *aswDirichletImag;

	X_INT16 iHighPassCutOffDftPoint;
	
} RVC_PITCH_ROM_be;


#define RVC_PITCH_METER_SIG  21053

typedef struct rvc_pitch_meter_tag_be {
  
	X_INT16 signature;
  
	void *pvScratchMemory; 

	// RAM variables
                           
	// Count the number of frames with continuous pitch track
	X_INT16   iStablePitchCount;
          
	// To hold pitch value of previous frame (or UNVOICED)
	PITCH_be stPrevPitch;
	PITCH_be stStableTrackPitch;
	X_INT16   iDistFromStableTrack;

	// To hold pitch candidates for a window
	PITCH_be *pstPitchCand;
	X_INT16   iNoOfPitchCand;

	// To hold data unique to each window type
	CMPLXsw_be *pstSingleInterpDft;
	X_INT16  iSingleDftShift;
	CMPLXsw_be *pstDoubleInterpDft;
	X_INT16  iDoubleDftShift;

	X_INT16    iInterpDftLen; // Copy of the value contained in Rom

	// To hold reduced list of spectral peaks
	POINT_be *pstSpectralPeaks; 
	X_INT16    iNoOfSpectralPeaks;

	// To hold resulting utility function
	POINT_be *pstPoints;
	X_INT16   iNoOfPoints;

	// Used for limitting computational complexity
	X_INT16    iCreatePieceWiseFunc_LoopCount;
	X_INT16    iCreatePieceWiseFunc_LoopLimit;

	// Miscellaneous
	X_INT16    iFrameNo;
	X_INT16    iHighPassCutOffDftPoint;
  
} RVC_PITCH_METER_be;


#endif

