/*=======================================================================================================*/
/*      Distributed Speech Recognition                                                                   */
/*      ETSI ES 202 050 Advanced Front-End Feature Extraction Algorithm & Compression Algorithm &        */
/*      ETSI ES 202 212 Extended Advanced Front-End Feature Extraction Algorithm & Compression Algorithm */  
/*      Bit exact C-language software implementation using ETSI basic operations library                 */
/*      3GPP TS 26.243 v 6.1.0 November 2004                                                                  */
/*      ANSI C code for the Fixed-Point Distributed Speech Recognition Extended Advanced Front-end       */
/*=======================================================================================================*/
/*-------------------------------------------------------------------------------
 *
 * FILE NAME: 
 *
 *    coder.c
 *
 * PURPOSE:   
 *
 *    This module is the main program for the encoding of (Extended) DSR Advanced 
 *    Front-End parameters. It performs compression using different quantisation 
 *    techniques and generates the bitstream format for the (Extended) Advanced DSR 
 *    standard proposal.
 *
 * INPUT: 
 *          
 *    The code can be compiled in two different modes: With speech reconstruction
 *    extensions enabled (XAFE defined), or without speech reconstruction 
 *    extensions (XAFE not defined). 
 *
 *    In both modes, the executable requires HTK files with MFCC data generated
 *    by ETSI ES 202 050 DSR Advanced Frontend. The input MFCC feature vectors 
 *    must contain both c0 and logE values. If -VAD option is passed, an 
 *    additional ASCII file containing the VAD flag for each frame must also
 *    be provided.
 * 
 *    When reconstruction extensions are enabled, in addition to HTK MFCC data,
 *    a  pitch period file and a classification file generated by ETSI 202 212 
 *    Extended Advanced DSR Frontend Feature Algorithm have to be supplied.
 *   
 *
 * OUTPUT: 
 *
 *    The executable generates a bitstream file compatible with ETSI DSR 
 *    specifications. By default the bitstream contains split vector quantiser 
 *    output for MFCC coefficients and VAD flags for each frame. If reconstruction 
 *    extensions are enabled, quantised pitch and classification information is also 
 *    included in bitstream. To help debugging, a text file showing output 
 *    bitstream data is also produced.
 *
 *-------------------------------------------------------------------------------*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "typedef.h"
#include "basic_op.h"
#include "count.h"
#include "common.h"
#include "coder.h"
#include "golay.h"

/* quantiser tables */
#include "quantiser_16kHz_BEC.tab"
#include "quantiser_8kHz_BEC.tab"

/* global variables */
FILE *HTK_File, *VAD_File, *BitStream_File, *TXT_File;
#ifdef XAFE
FILE *Pitch_Period_File, *Quantised_Pitch_Period_File, *Classification_File;
#endif
Word16 weight_c0_norm = 0;
Word16 weight_c0_shift = 0;
Word16 weight_logE = 0;
int in_pos =0, out_pos=0; /*Global variables to keep VAD data in sync with HTK data*/
unsigned char sampRate; /*Can be 00 (8), 01 (11) or 11 (16). Set by a command line flag.*/
float FrameRateFactor = VQFRAMERATEFACTOR; /* each frame pair is 20 ms */
Flag swapHTK = FALSE;



/*----------------------------------------------------------------------------
 *
 * FUNCTION NAME:    
 *
 *    main
 *
 * PURPOSE:          
 *
 *    Main control routine for split vector quantisation,
 *    pitch quantisation, and multiframe generation.
 *
 * INPUT:            
 *
 *    argc -  Number of command-line arguments.
 *
 *    argv[] - Individiual command-line argument strings.
 *
 * OUTPUT:
 *
 *    None.
 *
 *
 * RETURN VALUE:     
 *
 *    0 upon successful completion.
 *    1 upon any error.
 *
 *
 *---------------------------------------------------------------------------*/

int main(int argc, char *argv[])
{

   HTK_Header input_header;
   HTK_DataFrame data_chunk[24];
   HTK_DataFrame previous_data;
   unsigned int multi_frames=0;
   unsigned int data_frames;
   unsigned int i;
   int status;
   int done=FALSE;
   int sendnullmultiframe=FALSE;
   int *VAD_data;
   unsigned char zero = 0;
   unsigned char sync[2];
   unsigned short int mframeCounter = 1;  /* counts the number of frames */
#ifdef XAFE
   float pitch_chunk[24];
   float previous_pitch;
   UWord8 class_chunk[24];
   UWord8 previous_class;
#endif
#ifdef WMOPS
   int VQId; /* WMOPS counter */
#endif

   /* print welcome message */
   fprintf( stderr, "DSR Advanced Front-End Split Vector Quantiser.\n" );
   fprintf( stderr, "Motorola, France Telecom, Alcatel.\n\n" );
  
   /* parse command-line arguments */
   process_params( argc, argv );

   /* Verify that input file is a valid MFCC file in HTK format */
   if( htk_header_error( &input_header ) )
   {
      fprintf( stderr, "Exiting due to error in input HTK file header...\n" );
      exit(1);
   }

   /* print progress message */
   fprintf( stderr, "No. of 10ms frames: %d\n", input_header.nSamples );

#ifdef WMOPS
   /* initialise WMOPS counter */
   VQId=getCounterId("VQ");
   setCounter(VQId);
   Init_WMOPS_counter();
#endif

   if(VAD_File != NULL)
   {
      /*Read VAD info into an array*/
      if( ( VAD_data = calloc( 2 * input_header.nSamples, sizeof(int) ) ) == NULL )
      {
         fprintf( stderr, "Memory allocation error for VAD array.\n" );
         exit(1);
      }
      for ( i=0; i < input_header.nSamples; i++ )
      {
         fscanf( VAD_File, "%d ", &VAD_data[i] );
         if( ( VAD_data[i] != 0 ) && ( VAD_data[i] != 1 ) )
         {
            fprintf( stderr, "Incorrect VAD file format, must be integer 0 or 1 separated by space.\n" );
            exit(1);
         }
      }
        
   }
   else
   {
      VAD_data=NULL;
   }

   /* generate header for text vq index output file */
   print_txt_header( );


   /*Need to keep previous frame, since EOF is not detected, until you read past end.*/
   status = read_htk_coefs( VAD_data, &previous_data );

   if(status == error)
   {
      fprintf( stderr, "Error occured while reading data.\n");
      fprintf( stderr, "Exiting...\n");
      done = TRUE;
   }

   if(status == last_frame)
   {
      done = TRUE; /*Must have been an empty HTK file with good header.*/
   }

#ifdef XAFE
  /* Read pitch period and class */
  if (status == standard)
  {

    fread(&previous_pitch, sizeof(previous_pitch), 1, Pitch_Period_File);

    if((ferror(Pitch_Period_File) != 0) || (feof(Pitch_Period_File) != 0))
    {
      fprintf( stderr, "Error occured while reading pitch period file.\n" );
      fprintf( stderr, "Exiting...\n" );
      done = TRUE;
    }

    fscanf(Classification_File,"%c",&previous_class);

    if((ferror(Classification_File) != 0) || (feof(Classification_File) != 0))
    {
      fprintf( stderr, "Error occured while reading classification file.\n");
      fprintf( stderr, "Exiting...\n");
      done = TRUE;
    }

  }
#endif

   for(multi_frames=0; !done; multi_frames++)
   {  

      memcpy(&data_chunk[0], &previous_data, sizeof(HTK_DataFrame));
  
      data_frames=0;

#ifdef XAFE
    pitch_chunk[0] = previous_pitch;
    class_chunk[0] = previous_class;
#endif
      
      do{ /*Try and load a block of 24 MFCC data frames to fill one multiframe. (one loaded already)*/
    
         if((status=read_htk_coefs(VAD_data, &data_chunk[data_frames+1])) == standard)
         {
#ifdef XAFE
            /* Also, load 24 pitch period and class values (one loaded already). */
            fread(&pitch_chunk[data_frames+1], sizeof(pitch_chunk[0]), 1, Pitch_Period_File);
            fscanf(Classification_File,"%c",&class_chunk[data_frames+1]);
#endif
            data_frames++;
         }

      }while(data_frames<23 && (status==standard));

      /* see if possible to read the next speech frame from HTK file to determine if this will be last multiframe */ 
      /*EOF status only returned when you try to read past end of file */
      if(status==standard)
      {
         status = read_htk_coefs(VAD_data, &previous_data);
#ifdef XAFE
         /* Also, read pitch and class */
         fread(&previous_pitch, sizeof(previous_pitch), 1, Pitch_Period_File);
         fscanf(Classification_File,"%c",&previous_class);
#endif
      }

      print_bit_header(mframeCounter); /*output header for each multiframe*/
      mframeCounter++;
      mframeCounter = mframeCounter % 16;

      switch(status)
      {
    
      case standard:/*0*/ /* full multiframe of 24 speech frames */
         
         for( i=0; i<23; i+=2 )  
         {
#ifdef XAFE
            quantise_and_print( &data_chunk[i], &data_chunk[i+1], &pitch_chunk[i], &pitch_chunk[i+1],
                                &class_chunk[i], &class_chunk[i+1], TWO_FRAMES );
#else
            quantise_and_print( &data_chunk[i], &data_chunk[i+1], TWO_FRAMES );
#endif
         }

         break;

      case last_frame:/*1*/ /* last multiframe with (data_frames +1) speech frames */

         for( i=0; i<data_frames; i+=2 )
         {
#ifdef XAFE
            quantise_and_print( &data_chunk[i], &data_chunk[i+1], &pitch_chunk[i], &pitch_chunk[i+1],
                                &class_chunk[i], &class_chunk[i+1], TWO_FRAMES );

#else
            quantise_and_print( &data_chunk[i], &data_chunk[i+1], TWO_FRAMES );
#endif
         }
            
         if( (data_frames%2) == 0 )/*if odd number of frames do last one*/
         {
#ifdef XAFE
            quantise_and_print( &data_chunk[data_frames], NULL, &pitch_chunk[data_frames], NULL,
                                &class_chunk[data_frames], NULL, ONE_FRAME );
            
#else
            quantise_and_print( &data_chunk[data_frames], NULL, ONE_FRAME );                      
#endif
            data_frames+=2; /*cause just printed a zero one in bitstream*/
         }
       
         /* if there are less than 21 speech frames then need to send an extra multiframe with null data */
         if( data_frames >= 20 ) 
         {
            sendnullmultiframe = TRUE;
         }
       
         /*pad end with zeros for empty frames and CRC*/
       
         while( data_frames < 23 ) /*since print dataframe, and dataframe + 1*/
         {  
            switch( data_frames % 4 )
            {

            case 0:  
#ifdef XAFE
               for( i = 0; i< 27; i++ )
               {
                  fwrite(&zero, sizeof(unsigned char), 1, BitStream_File);
               }
#else
               for( i = 0; i < 23; i++ )
               {
                  fwrite(&zero, sizeof(unsigned char), 1, BitStream_File);
               }
#endif
               fflush(BitStream_File);
               data_frames+=4;
               break;
             
            case 1:  
               data_frames++;
               break;
             
            case 2:  
#ifdef XAFE
               for( i = 0; i< 13; i++ )
               {
                  fwrite(&zero, sizeof(unsigned char), 1, BitStream_File);
               }
#else
               for( i = 0; i < 11; i++ )
               {
                  fwrite(&zero, sizeof(unsigned char), 1, BitStream_File);
               }
#endif
               fflush(BitStream_File);
               data_frames+=2;
               break;
             
            case 3:  
               data_frames++;
               break;             
             
            default: 
               fprintf( stderr, "ERROR: x%%4 has to be in [0,1,2,3]\n" );
               exit(1);
             
            }
         }

         /* send a null multiframe at end */
         if(sendnullmultiframe == TRUE)
         {
            print_bit_header( mframeCounter ); /*output header for null multiframe*/
#ifdef XAFE
            for( i=0; i< (6*27); i++)
            {
               fwrite(&zero, sizeof(unsigned char), 1, BitStream_File);
            }
#else
            for( i=0; i< (6*23); i++)
            {
               fwrite(&zero, sizeof(unsigned char), 1, BitStream_File);
            }
#endif
         }     

         /* Finally write out an inverse sync after all the data. */
         sync[0] = 0xff & INV_SYNC_SEQ;
         sync[1] = 0xff & ((0xff00 & INV_SYNC_SEQ) >> 8);

         fwrite(sync, sizeof(unsigned char), 2, BitStream_File);

         fflush(BitStream_File);

         done = TRUE;
         break;

      case error:/*-1*/
      default:  
         fprintf( stderr, "\n\nError reading input HTK file.\n" );
         fprintf( stderr, "Exiting...\n" );
         done = TRUE;
         break;
      } /* switch (status) */
   }


   /* close output files and deallocate memory */
   fclose(HTK_File);
   fclose(BitStream_File);
   fclose(TXT_File);  
#ifdef XAFE
   fclose(Pitch_Period_File);
   fclose(Classification_File);
   fclose(Quantised_Pitch_Period_File);
#endif
   if (VAD_File != NULL)
   {
      free(VAD_data);
      fclose(VAD_File);
   }

#ifdef WMOPS        
   /* Report WMOPS counts */
   printf("\n");
   WMOPS_output(0);
#endif

   return 0;

}



/*----------------------------------------------------------------------------
 *
 * FUNCTION NAME:    
 *
 *    htk_header_error
 *
 * PURPOSE:          
 *
 *    Verify consistency of input HTK file header.       
 *
 * INPUT:            
 *
 *    *given - Pointer to structure containing HTK header data.
 *
 * OUTPUT:
 * 
 *    None. 
 *
 * RETURN VALUE:     
 *
 *    0 upon success.
 *    Otherwise, number of encountered errors.
 *
 *---------------------------------------------------------------------------*/


int htk_header_error( HTK_Header *given )
{
   long int htk_file_length;
#ifdef XAFE
   long int pitch_file_length;
   long int class_file_length;
#endif
   int errors = 0;

   fseek(HTK_File, 0L, SEEK_END);
   /*Get the file length in bytes*/
   htk_file_length = ftell(HTK_File); 
   htk_file_length -= 12; /*The HTK header is 12 bytes long*/
   rewind(HTK_File);
  
   read_htk_header(given);

   if ( given -> sampSize >= 256 )
   {
      swapHTK = 1;
      Swap16( &given -> sampSize );
      Swap16( &given -> parmKind );
      Swap32( &given -> sampPeriod );
      Swap32( &given -> nSamples );
   }

   if( given->sampPeriod != EXPECTED_SAMP_PERIOD )
   {
      fprintf( stderr, "Error: Incorrect sample period in input HTK file.\n");
      errors++;
   }

   if(given->sampSize != EXPECTED_SAMP_SIZE)
   {
      fprintf( stderr, "Error: Incorrect sample size in input HTK file.\n");
      errors++;
   }

   if(given->parmKind != EXPECTED_PARM_KIND)
   {
      fprintf( stderr, "Error: Input HTK file must contain MFCC data.\n");
      errors++;
   }


   if((given->nSamples * EXPECTED_SAMP_SIZE) != htk_file_length)
   {
      fprintf( stderr, "Error: Incorrect HTK file length.\n");
      errors++;
   }
    

#ifdef XAFE 
  /*Get the pitch file length in bytes*/
  fseek(Pitch_Period_File, 0L, SEEK_END);
  pitch_file_length = ftell(Pitch_Period_File);
  rewind(Pitch_Period_File);

  if((pitch_file_length * EXPECTED_SAMP_SIZE) < (htk_file_length * sizeof(float)))
  {
    fprintf(stderr, 
            "Error: The sizes of the input HTK file and the input Pitch\n"
            "       Period file do not correspond to each other %ld %ld.\n", (pitch_file_length * EXPECTED_SAMP_SIZE), (htk_file_length * sizeof(float)));

    errors++;
  }


  /*Get the classification file length in bytes*/
  fseek(Classification_File, 0L, SEEK_END);
  class_file_length = ftell(Classification_File);
  rewind(Classification_File);

  if((class_file_length * EXPECTED_SAMP_SIZE) < htk_file_length)
  {
    fprintf(stderr,
            "Error: The sizes of the input HTK file and the input Classi-\n"
            "       fication file do not correspond to each other.\n");

    errors++;
  }
#endif
   return errors;
}



/*----------------------------------------------------------------------------
 * FUNCTION NAME:    
 *
 *    quantise_and_print
 *
 * PURPOSE:          
 *
 *    Quantise an input cepstral vector & pitch pair. 
 *    Write output to bit file.               
 *
 *
 * INPUT:            
 *
 *    *first, *second -  First and second MFCC vectors to be quantised.
 *    *pitch1, *pitch2 - First and second pitch periods to be quantised.
 *    *class1, *class2 - First and second pitch class bits to be quantised.
 *    numFrames - Number of frames to be quantised. Either a frame pair
 *                or a single frame.
 *
 *
 * OUTPUT:    
 *   
 *    None.
 *
 * RETURN VALUE:     
 * 
 *    None.
 *
 *---------------------------------------------------------------------------*/


#ifdef XAFE
void quantise_and_print( HTK_DataFrame *first, HTK_DataFrame *second, 
                         float *pitch1, float *pitch2, unsigned char *class1, unsigned char *class2, 
                         int numFrames )

#else
void quantise_and_print( HTK_DataFrame *first, HTK_DataFrame *second, int numFrames )
#endif
{
   long int file_pos;
   static Flag print_two_pairs=FALSE;

   UWord16 index;

#ifdef XAFE
   static UWord8 four_frames[27];
#else
   static UWord8 four_frames[23];
#endif

   UWord8 two_frames_copy[11] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
   UWord8 index_array[14] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
   UWord8 crc;

#ifdef XAFE
  Word32 lwPitch1, lwQuantPitch1;
  Word32 lwPitch2, lwQuantPitch2;
  UWord8 pc_data[14];
  UWord8 pc_crc[2];
  float quant_pitch1, quant_pitch2;
  UWord8 pitch1_index = 0, pitch2_index = 0;
  UWord8 class1_index = 0, class2_index = 0;
  static Word32 plwQPHistory[3] = {0, 0, 0};
  static Word16 iReliableFlag = FALSE;
#endif
                
   Reset_WMOPS_counter();


   /* print progress message */
   {
      static int frame_counter = 1;
      fprintf( stderr, "\rProcessing frame pair %d", frame_counter++ );
   }

                
   /* Quantise MFCC data.*/
   get_best_dataframe(&index_array[0], first);  


#ifdef XAFE
  /*
   * Quantize the pitch and class information.
   * Write out the quantized pitch value to the
   * Quantized_Pitch_Period_File. Update pitch
   * period history.
   */
   
  lwPitch1 = ((*pitch1)*65536.0+0.5);
  quant_pitch_abs(lwPitch1,&lwQuantPitch1,&pitch1_index);
  quant_pitch1 = (float)lwQuantPitch1/65536.0;
  get_class_bit(*class1,&class1_index);
  fwrite(&quant_pitch1,4,1,Quantised_Pitch_Period_File);
  plwQPHistory[2] = plwQPHistory[1];	move32();
  plwQPHistory[1] = plwQPHistory[0];	move32();
  plwQPHistory[0] = lwQuantPitch1;	move32();
#endif

   /* write indices to output text file */
#ifdef XAFE
   print_txt_coefs( index_array, pitch1_index, class1_index );
#else
   print_txt_coefs( index_array );
#endif

   test();
   if( sub(numFrames,TWO_FRAMES) == 0 )
   {
      get_best_dataframe(&index_array[7], second);

#ifdef XAFE
    lwPitch2 = ((*pitch2)*65536.0+0.5);
    quant_pitch_diff(lwPitch2, plwQPHistory, &iReliableFlag, &lwQuantPitch2,&pitch2_index);
    quant_pitch2 = (float)lwQuantPitch2/65536.0;
    get_class_bit(*class2,&class2_index);
    fwrite(&quant_pitch2,4,1,Quantised_Pitch_Period_File);
    plwQPHistory[2] = plwQPHistory[1];	move32();
    plwQPHistory[1] = plwQPHistory[0];	move32();
    plwQPHistory[0] = lwQuantPitch2;	move32();
#endif

      /* write indices to output text file */
#ifdef XAFE
    print_txt_coefs( index_array+7, pitch2_index, class2_index );
#else
    print_txt_coefs( index_array+7 );
#endif
   }


   /* Format the frame packet stream into binary bitstream for output */
   /* note that even frame packet pairs are aligned differently to odd frame packet pairs in the multiframe */

   if( print_two_pairs==FALSE )
   {

      /* byte align the even frame pair packet for binary bitstream output format (octet aligned)*/
      four_frames[0] = (0x3f & index_array[0]); 
      four_frames[0] |= (0xc0 & shl(index_array[1], 6)); 
      four_frames[1] = (0x0f & shr(index_array[1], 2)); 
      four_frames[1] |= (0xf0 & shl(index_array[2], 4)); 
      four_frames[2] = (0x03 & shr(index_array[2], 4)); 
      four_frames[2] |= (0xfc & shl(index_array[3], 2)); 
      four_frames[3] = (0x3f & index_array[4]); 
      four_frames[3] |= (0xc0 & shl(index_array[5], 6)); 
      four_frames[4] =  (0x0f & shr(index_array[5], 2)); 
      four_frames[4] |= (0xf0 & shl(index_array[6], 4)); 
      four_frames[5] = (0x0f & shr(index_array[6], 4)); 
      four_frames[5] |= (0xf0 & shl(index_array[7], 4)); 
      four_frames[6] = (0x03 & shr(index_array[7], 4)); 
      four_frames[6] |= (0xfc & shl(index_array[8], 2)); 
      four_frames[7] = (0x3f & index_array[9]); 
      four_frames[7] |= (0xc0 & shl(index_array[10], 6)); 
      four_frames[8] = (0x0f & shr(index_array[10], 2)); 
      four_frames[8] |= (0xf0 & shl(index_array[11], 4)); 
      four_frames[9] = (0x03 & shr(index_array[11], 4)); 
      four_frames[9] |= (0xfc & shl(index_array[12], 2)); 
      four_frames[10] = (0xff & index_array[13]); 
     


      for(index=0; index<11; index++) 
      {
         two_frames_copy[index] = four_frames[index]; 
      }

      crc = mfcc_crc_encode(two_frames_copy);   /* compute the crc for the frame pair packet */

      four_frames[11] = (0x0f & crc); logic16(); move16();

#ifdef XAFE
    /*
     * Insert pitch & class bits into the binary stream
     */

    for (index = 0; index < 7; index++)
    {
      pc_data[index] = (0x01 & (pitch1_index >> index));
    }
    for (index = 7; index < 12; index++)
    {
      pc_data[index] = (0x01 & (pitch2_index >> (index-7)));
    }
    pc_data[12] = (0x01 & class1_index);
    pc_data[13] = (0x01 & class2_index);

    pc_crc_encode(pc_data,pc_crc);

    four_frames[11] |= (0xf0 & shl(pitch1_index,4));
    four_frames[12] = (0x07 & shr(pitch1_index,4));
    four_frames[12] |= (0xf8 & shl(pitch2_index,3));
    four_frames[13] = (0x01 & class1_index);
    four_frames[13] |= (0x02 & shl(class2_index,1));
    four_frames[13] |= (0x04 & shl(pc_crc[0],2));
    four_frames[13] |= (0x08 & shl(pc_crc[1],3));
#endif

#ifdef XAFE
      fwrite(four_frames, sizeof(unsigned char), 14, BitStream_File);
#else
      fwrite(four_frames, sizeof(unsigned char), 12, BitStream_File);
#endif
      fflush(BitStream_File);

      print_two_pairs=TRUE; 
   }
   else
   {
      /*the data in two_frames_copy[] is for calculating the crc only.*/
  
      two_frames_copy[0] = (0x3f & index_array[0]); 
      two_frames_copy[0] |= (0xc0 & shl(index_array[1], 6)); 
      two_frames_copy[1] = (0x0f & shr(index_array[1], 2)); 
      two_frames_copy[1] |= (0xf0 & shl(index_array[2], 4)); 
      two_frames_copy[2] = (0x03 & shr(index_array[2], 4)); 
      two_frames_copy[2] |= (0xfc & shl(index_array[3], 2)); 
      two_frames_copy[3] = (0x3f & index_array[4]); 
      two_frames_copy[3] |= (0xc0 & shl(index_array[5], 6)); 
      two_frames_copy[4] =  (0x0f & shr(index_array[5], 2)); 
      two_frames_copy[4] |= (0xf0 & shl(index_array[6], 4)); 
      two_frames_copy[5] = (0x0f & shr(index_array[6], 4)); 
      two_frames_copy[5] |= (0xf0 & shl(index_array[7], 4)); 
      two_frames_copy[6] = (0x03 & shr(index_array[7], 4)); 
      two_frames_copy[6] |= (0xfc & shl(index_array[8], 2)); 
      two_frames_copy[7] = (0x3f & index_array[9]); 
      two_frames_copy[7] |= (0xc0 & shl(index_array[10], 6)); 
      two_frames_copy[8] = (0x0f & shr(index_array[10], 2)); 
      two_frames_copy[8] |= (0xf0 & shl(index_array[11], 4)); 
      two_frames_copy[9] = (0x03 & shr(index_array[11], 4)); 
      two_frames_copy[9] |= (0xfc & shl(index_array[12], 2)); 
      two_frames_copy[10] = (0xff & index_array[13]); 
     
      crc = mfcc_crc_encode(two_frames_copy); /* compute crc for frame pair packet */
     
     
      /* byte align the odd numbered frame pair packet for binary bitstream output format (octet aligned) */
#ifdef XAFE
      four_frames[13] |= (0xf0 & shl(index_array[0],4));
      four_frames[14] = (0x03 & shr(index_array[0],4));
      four_frames[14] |= (0xfc & shl(index_array[1],2));
      four_frames[15] = (0x3f & index_array[2]);
      four_frames[15] |= (0xc0 & shl(index_array[3],6));
      four_frames[16] = (0x0f & shr(index_array[3],2));
      four_frames[16] |= (0xf0 & shl(index_array[4],4));
      four_frames[17] = (0x03 & shr(index_array[4],4));
      four_frames[17] |= (0xfc & shl(index_array[5],2));
      four_frames[18] = (0xff & index_array[6]);
      four_frames[19] = (0x3f & index_array[7]);
      four_frames[19] |= (0xc0 & shl(index_array[8],6));
      four_frames[20] = (0x0f & shr(index_array[8],2));
      four_frames[20] |= (0xf0 & shl(index_array[9],4));
      four_frames[21] = (0x03 & shr(index_array[9],4));
      four_frames[21] |= (0xfc & shl(index_array[10],2));
      four_frames[22] = (0x3f & index_array[11]);
      four_frames[22] |= (0xc0 & shl(index_array[12],6));
      four_frames[23] =  (0x0f & shr(index_array[12],2));
      four_frames[23] |= (0xf0 & shl(index_array[13],4));
      four_frames[24] = (0x0f & shr(index_array[13],4));
      four_frames[24] |= (0xf0 & shl(crc,4));
#else
      four_frames[11] |= (0xf0 & shl(index_array[0], 4)); 
      four_frames[12] = (0x03 & shr(index_array[0], 4)); 
      four_frames[12] |= (0xfc & shl(index_array[1], 2)); 
      four_frames[13] = (0x3f & index_array[2]); 
      four_frames[13] |= (0xc0 & shl(index_array[3], 6)); 
      four_frames[14] = (0x0f & shr(index_array[3], 2)); 
      four_frames[14] |= (0xf0 & shl(index_array[4], 4)); 
      four_frames[15] = (0x03 & shr(index_array[4], 4)); 
      four_frames[15] |= (0xfc & shl(index_array[5], 2)); 
      four_frames[16] = (0xff & index_array[6]); 
      four_frames[17] = (0x3f & index_array[7]); 
      four_frames[17] |= (0xc0 & shl(index_array[8], 6)); 
      four_frames[18] = (0x0f & shr(index_array[8], 2)); 
      four_frames[18] |= (0xf0 & shl(index_array[9], 4)); 
      four_frames[19] = (0x03 & shr(index_array[9], 4)); 
      four_frames[19] |= (0xfc & shl(index_array[10], 2)); 
      four_frames[20] = (0x3f & index_array[11]); 
      four_frames[20] |= (0xc0 & shl(index_array[12], 6)); 
      four_frames[21] =  (0x0f & shr(index_array[12], 2)); 
      four_frames[21] |= (0xf0 & shl(index_array[13], 4)); 
      four_frames[22] = (0x0f & shr(index_array[13], 4)); 
      four_frames[22] |= (0xf0 & shl(crc, 4)); 
#endif


#ifdef XAFE
    /*
     * Insert pitch & class bits into the binary stream
     */

    for (index = 0; index < 7; index++)
    {
        pc_data[index] = (0x01 & shr(pitch1_index, index)); 
    }
    for (index = 7; index < 12; index++)
    {
        pc_data[index] = (0x01 & shr(pitch2_index, (index-7))); 
    }
    pc_data[12] = (0x01 & class1_index); 
    pc_data[13] = (0x01 & class2_index); 

    pc_crc_encode(pc_data,pc_crc);

    four_frames[25] = (0x7f & pitch1_index);
    four_frames[25] |= (0x80 & shl(pitch2_index,7));
    four_frames[26] = (0x0f & shr(pitch2_index,1));
    four_frames[26] |= (0x10 & shl(class1_index,4));
    four_frames[26] |= (0x20 & shl(class2_index,5));
    four_frames[26] |= (0x40 & shl(pc_crc[0],6));
    four_frames[26] |= (0x80 & shl(pc_crc[1],7));
#endif

      /*Go back one byte before writing to achieve contiguous bitstream on octet bounaries */

      file_pos = ftell(BitStream_File);
      file_pos--;
      fseek(BitStream_File, file_pos, SEEK_SET);
#ifdef XAFE
      fwrite(&four_frames[13], sizeof(unsigned char), 14, BitStream_File);
#else
      fwrite(&four_frames[11], sizeof(unsigned char), 12, BitStream_File);
#endif
      fflush(BitStream_File);
        
      print_two_pairs=FALSE; move16();
    
   }


   fflush(BitStream_File);

#ifdef WMOPS
   fwc();
#endif

   return;

}


/*----------------------------------------------------------------------------
 * FUNCTION NAME:   
 *
 *    best_centroid
 *  
 * PURPOSE:         
 *
 *    Find best centroid for a given mel-cepstral coefficient pair.
 * 
 * INPUT:           
 *
 *    a, b - Mel-cepstral coefficients to be quantised.
 *    table[][2] - Vector quantisation (VQ) table.  
 *    NumCentroids - No. of entries in VQ table.
 *    weighting - Whether to assign equal weights to a and b.
 *
 *
 * OUTPUT:          
 *
 *    None.
 *
 * RETURN VALUE:    
 *
 *    VQ table index corresponding to (a, b).
 *
 *
 *---------------------------------------------------------------------------*/


Word16 best_centroid(Word16 a, Word16 b, const Word16 table[][2], int NumCentroids, Flag weighting)
{
   /*Find the best centroid from a given quantiser table, using Euclidian distance*/
   /*or for logE and c0 weighted distance from the given vectors a and b.*/

   /*Apply index addressing for table, since the first dimension (NumCentroids) is*/
   /*either 256 or 64. Address is table[i][j] = *(table + i*2 + j)*/
   Word16 i;
   Word16 BestIndex;
   Word16 temp;
   Word32 BestDistance;
   Word32 Distance;

   BestIndex = 0; move16();
   BestDistance = 0x7fffffff; move32();

   for( i=0; i < NumCentroids; i++ )
   {
      test();
      if(weighting == FALSE)
      {                
         temp = sub(a, table[i][0]);
         Distance = L_mult(temp, temp);
         temp = sub(b, table[i][1]);
         Distance = L_mac(Distance, temp, temp);
      }
      else /* (c0, logE) quantisation */
      {
         temp = sub(a, table[i][0]);
         /* to prevent underflows, multiply with scaled version of weight_c0 */
         Distance = L_mult(mult(temp, weight_c0_norm), temp);
         Distance = L_shr(Distance, weight_c0_shift);
         temp = sub(b, table[i][1]);
         /* weight_logE is 0x7ffff, therefore  mult(temp, weight_logE) = temp */
         Distance = L_add(L_mult(temp, temp), Distance);
      }
     
      test();
      if ( L_sub(Distance, BestDistance) < 0 )
      {
         BestDistance = Distance; move32();
         BestIndex = i; move16();
      }     
   }       

   return BestIndex;
}


/*----------------------------------------------------------------------------
 * FUNCTION NAME  
 *
 *    read_htk_header
 * 
 * PURPOSE:        
 *
 *    Parse input HTK file header.
 *
 * INPUT:           
 *
 *    None.
 *
 *
 * OUTPUT:          
 *
 *    *header: Pointer to structure containing HTK header data.
 *                                        
 *
 * RETURN VALUE:    
 *
 *    None.
 *
 *
 *---------------------------------------------------------------------------*/


void read_htk_header( HTK_Header *header )
{

   /*This function reads the header into the struct*/
   fread(&header->nSamples, 4, 1, HTK_File);
   fread(&header->sampPeriod, 4, 1, HTK_File);
   fread(&header->sampSize, 2, 1, HTK_File);
   fread(&header->parmKind, 2, 1, HTK_File);

   if(feof(HTK_File)!=0 || ferror(HTK_File)!=0)
   {
      /*Should never find EOF here since you would be expecting the data in the MultiFrame too.*/
      fprintf( stderr, "Error or end of file while reading input HTK file header.\n");
      exit(1);
   }

   return;
}


/*----------------------------------------------------------------------------
 * FUNCTION NAME:  
 *
 *    read_htk_coefs
 *
 * PURPOSE:
 *
 *    Read MFCC data from HTK_File. 
 *                 
 * INPUT:
 *
 *    *VAD_data: VAD array.
 *
 *
 * OUTPUT:         
 *
 *    *given: Pointer to MFCC vector structure  read from file. Structure also
 *            includes VAD information for the frame.
 *
 *
 * RETURN VALUE:   
 *  
 *    standard -  this frame was not the last frame.
 *    last_frame -  this frame was the last frame.
 *    error - an error reading HTK_File occurred.
 *
 *
 *---------------------------------------------------------------------------*/


int read_htk_coefs( int *VAD_data, HTK_DataFrame *given )
{

   fread(&given->c1, 4, 1, HTK_File);
   fread(&given->c2, 4, 1, HTK_File);
   fread(&given->c3, 4, 1, HTK_File);
   fread(&given->c4, 4, 1, HTK_File);
   fread(&given->c5, 4, 1, HTK_File);
   fread(&given->c6, 4, 1, HTK_File);
   fread(&given->c7, 4, 1, HTK_File);
   fread(&given->c8, 4, 1, HTK_File);
   fread(&given->c9, 4, 1, HTK_File);
   fread(&given->c10, 4, 1, HTK_File);
   fread(&given->c11, 4, 1, HTK_File);
   fread(&given->c12, 4, 1, HTK_File);
   fread(&given->c0, 4, 1, HTK_File);
   fread(&given->logE, 4, 1, HTK_File);
  
   if( swapHTK ) 
   {
	  Swap32( (Word32 *) &given -> c1);
	  Swap32( (Word32 *) &given -> c2);
	  Swap32( (Word32 *) &given -> c3);
	  Swap32( (Word32 *) &given -> c4);
	  Swap32( (Word32 *) &given -> c5);
	  Swap32( (Word32 *) &given -> c6);
	  Swap32( (Word32 *) &given -> c7);
	  Swap32( (Word32 *) &given -> c8);
	  Swap32( (Word32 *) &given -> c9);
	  Swap32( (Word32 *) &given -> c10);
	  Swap32( (Word32 *) &given -> c11);
	  Swap32( (Word32 *) &given -> c12);
	  Swap32( (Word32 *) &given -> c0);
	  Swap32( (Word32 *) &given -> logE);
  }
   
   
   if (VAD_data == NULL) /*if VAD not being used, just set speech to true*/
   {
      given->VAD = 1;
   }     
   else
   {
      given->VAD = VAD_data[in_pos]; /*Read VAD data from input array*/
      in_pos++;
   }
        
   if(ferror(HTK_File)!=0)
   {
      return error;
   }

   if(feof(HTK_File)!=0)
   {
      return last_frame;
   }

   return standard;

}



/*----------------------------------------------------------------------------
 * FUNCTION NAME:     
 *
 *    print_txt_coefs
 *
 * PURPOSE:           
 *
 *    Write vector quantiser output to a file in text format.
 *
 * INPUT:             
 *
 *    *index - MFCC Vector quantiser indices.
 *    pitch_index - quantised pitch.
 *    class_index - class bit.
 *
 * OUTPUT:            
 *
 *    None.
 *
 * RETURN VALUE:      
 *
 *    None.
 *
 *
 *---------------------------------------------------------------------------*/

#ifdef XAFE
void print_txt_coefs( UWord8 *index, UWord8 pitch_index, UWord8 class_index )
#else
void print_txt_coefs( UWord8 *index )
#endif
{

   static int frame = 0;
  
   fprintf( TXT_File, "%5i      ", frame );
   fprintf( TXT_File, "%3i         ", index[0] );
   fprintf( TXT_File, "%3i         ", index[1] );
   fprintf( TXT_File, "%3i         ", index[2] );
   fprintf( TXT_File, "%3i         ", index[3] );
   fprintf( TXT_File, "%3i         ", index[4] );
   fprintf( TXT_File, "%3i         ", index[5] >> 1 );
   fprintf( TXT_File, "%4i       ", index[6] );
   if ( VAD_File != NULL )
   {
      fprintf( TXT_File, "%3i", index[5] & 0x01 );
   }
   else
   {
      fprintf( TXT_File, "  - " );
   }
#ifdef XAFE
   fprintf( TXT_File, "   %3i      ", pitch_index );
   fprintf( TXT_File, "%3i", class_index );
#endif
   fprintf( TXT_File, "\n" );

   if( ferror( TXT_File ) != 0 )
   {
      fprintf( stderr, "An error occured while writing vq output to text file.\n");
      exit(1);
   }
  
  
   fflush(TXT_File);
  
   frame++;
  
   return;

}





/*----------------------------------------------------------------------------
 * FUNCTION NAME:  
 *
 *    get_best_dataframe
 *
 * PURPOSE:        
 *
 *   Split vector quantisation of input mel-cepstral coefficient vector.
 *                
 *
 * INPUT:          
 *
 *   *given: Floating-point MFCC vector to be quantised.
 *
 *
 * OUTPUT:         
 *
 *   *index_array: Output indices after split VQ.
 *
 *
 * RETURN VALUE:   
 *
 *   None.
 *
 *
 *---------------------------------------------------------------------------*/


void get_best_dataframe( UWord8 *index_array, HTK_DataFrame *given )
{

   Word16 c1, c2, c3, c4, c5, c6, c7, c8, c9, c10, c11, c12, c0, logE;


   /* Convert floating-point MFCC coefficients to bit-exact */
   c1   = (Word16) ( given->c1   *  C_FACTOR    );
   c2   = (Word16) ( given->c2   *  C_FACTOR    );
   c3   = (Word16) ( given->c3   *  C_FACTOR    );
   c4   = (Word16) ( given->c4   *  C_FACTOR    );
   c5   = (Word16) ( given->c5   *  C_FACTOR    );
   c6   = (Word16) ( given->c6   *  C_FACTOR    );
   c7   = (Word16) ( given->c7   *  C_FACTOR    );
   c8   = (Word16) ( given->c8   *  C_FACTOR    );
   c9   = (Word16) ( given->c9   *  C_FACTOR    );
   c10  = (Word16) ( given->c10  *  C_FACTOR    );
   c11  = (Word16) ( given->c11  *  C_FACTOR    ); 
   c12  = (Word16) ( given->c12  *  C_FACTOR    ); 
   c0   = (Word16) ( given->c0   *  C0_FACTOR   ); 
   logE = (Word16) ( given->logE *  LOGE_FACTOR );

   move16(); move16(); move16(); move16(); move16(); move16(); move16(); 
   switch(sampRate)
   {
   case eight_kHz:
      index_array[0] = best_centroid(  c1,   c2,   quantiser8kHz_0_1,    64, FALSE );
      index_array[1] = best_centroid(  c3,   c4,   quantiser8kHz_2_3,    64, FALSE );
      index_array[2] = best_centroid(  c5,   c6,   quantiser8kHz_4_5,    64, FALSE );
      index_array[3] = best_centroid(  c7,   c8,   quantiser8kHz_6_7,    64, FALSE );
      index_array[4] = best_centroid(  c9,  c10,   quantiser8kHz_8_9,    64, FALSE );
      /*c11 and c12 are only quantised using 5 bits*/
      index_array[5] = best_centroid( c11,  c12,   quantiser8kHz_10_11,  32, FALSE );
      /*VAD flag is packed into 6th bit*/
      index_array[5] = (shl(index_array[5],1))| given->VAD; logic16(); move16();
      index_array[6] = best_centroid(  c0, logE,   quantiser8kHz_12_13, 256,  TRUE  );
      break;
       
   case sixteen_kHz:
      index_array[0] = best_centroid(  c1,   c2,     quantiser16kHz_0_1,  64, FALSE );
      index_array[1] = best_centroid(  c3,   c4,     quantiser16kHz_2_3,  64, FALSE );
      index_array[2] = best_centroid(  c5,   c6,     quantiser16kHz_4_5,  64, FALSE );
      index_array[3] = best_centroid(  c7,   c8,     quantiser16kHz_6_7,  64, FALSE );
      index_array[4] = best_centroid(  c9,  c10,     quantiser16kHz_8_9,  64, FALSE );
      /*c11 and c12 are only quantised using 5 bits*/
      index_array[5] = best_centroid( c11,  c12,   quantiser16kHz_10_11,  32, FALSE );
      /*VAD flag is packed into 6th bit*/
      index_array[5] = (shl(index_array[5],1))| given->VAD; logic16(); move16();
      index_array[6] = best_centroid(  c0, logE,   quantiser16kHz_12_13, 256,  TRUE );
      break;


   case eleven_kHz: 
   default: 
      fprintf( stderr, "Error: Incorrect sampling rate.\n");
      exit(1);

   }
  

   return;
}




/*----------------------------------------------------------------------------
 * FUNCTION NAME:  
 *
 *    print_txt_header
 *
 * PURPOSE:        
 *
 *    Generate header for human readable vector quantiser output file.
 *
 * INPUT:          
 *
 *    None.
 *
 * OUTPUT:         
 *
 *    None.
 *
 * RETURN VALUE:   
 *
 *    None.
 *
 *
 *---------------------------------------------------------------------------*/


void print_txt_header( void )
{

   fprintf( TXT_File, "frame       (c1, c2)    (c3, c4)    (c5, c6)    (c7, c8)    (c9, c10)   (c11, c12)  (c0, logE)  VAD" ); 
#ifdef XAFE
   fprintf( TXT_File, "   pitch    class" );
#endif
   fprintf( TXT_File, "\n" );
   fprintf( TXT_File, "-----       --------    --------    --------    --------    ---------   ----------  ----------  ---" );
#ifdef XAFE
   fprintf( TXT_File, "   -----    -----" );
#endif
   fprintf( TXT_File, "\n\n" );
   
   fflush( TXT_File );

   return;

}


/*----------------------------------------------------------------------------
 * FUNCTION NAME:   
 *
 *    print_bit_header
 *
 * PURPOSE:         
 *
 *    Construct multiframe header and write to bitstream file.
 *
 * INPUT:           
 *
 *    mframeCounter: Multiframe counter.
 *
 *
 * OUTPUT:      
 *
 *    None.
 *
 * RETURN VALUE:    
 *
 *    None.
 *
 *
 *---------------------------------------------------------------------------*/


void print_bit_header( UWord16 mframeCounter )
{

   unsigned char sync[2];
   unsigned short int golay_crc, golay_data;
   unsigned char header[4] = {0, 0, 0, 0}; 
  
   /* Put out info in BitStream - SYNC first. */
   /* sync is sent lsb first */
   sync[0] = 0xff & SYNC_SEQ;
   sync[1] = 0xff & ((0xff00 & SYNC_SEQ) >> 8);
   fwrite(sync, sizeof(unsigned char), 2, BitStream_File);
   fflush(BitStream_File);


   /* Construct header info from the flags #define'd at the top. */
   /* And put on the golay code. */
   golay_data  =  0x0003 & sampRate;  /*takes up the bottom two bits.*/
   golay_data |=  0x0004 & (feType<<2);
   golay_data |=  0x0078 & (mframeCounter<<3);
   /* Insert additional header information to the expansion bits below */
#ifdef XAFE
  golay_data |= 0x0080;  /* Bit set to indicate Extended DSR function */
#endif
   golay_crc = golay_encode(golay_data);


   /* A 4 byte header */
   header[0] = 0xff & golay_data;
   header[1] = 0xff & (golay_data >> 8);
   header[2] = 0xff & golay_crc ;
   header[3] = 0xff & (golay_crc >> 8);
  
   /* Print out to bitstream */
   fwrite(header, sizeof(unsigned char), 4, BitStream_File);
   fflush(BitStream_File);

   return;

}


/*----------------------------------------------------------------------------
 * FUNCTION NAME:   
 *
 *    process_params
 *
 * PURPOSE:         
 *
 *    Parse command-line arguments.
 *
 * INPUT:           
 *
 *    argc - Number of command-line arguments.
 *    argv[] - Individiual command-line argument strings.
 *
 *
 * OUTPUT:  
 *
 *    None.
 *
 * RETURN VALUE:   
 *
 *   None.
 *
 *
 *---------------------------------------------------------------------------*/

 

void process_params( int argc, char *argv[] )
{
   int errors=0;
   int argCounter=0;
   int htk_file_arg;
   
   char *ptP;
   char VADFilename[199];


#ifdef XAFE
   if (argc!=10)
   {
      print_usage();
      exit(1);
   }
#else     
   if(argc!=7)
   {
      print_usage();
      exit(1);
   }
#endif


   if((HTK_File = fopen(argv[++argCounter], "rb"))==NULL)
   {
      fprintf(stderr, "Error: Cannot open %s.\n", argv[argCounter]);
      errors++;  
   }
   htk_file_arg = argCounter;

#ifdef XAFE
   if((Pitch_Period_File = fopen(argv[++argCounter], "rb"))==NULL)
   {
      fprintf(stderr, "Error: Cannot open %s.\n", argv[argCounter]);
      errors++;  
   }
#endif


#ifdef XAFE  
   if((Classification_File = fopen(argv[++argCounter], "r"))==NULL)
   {
      fprintf(stderr, "Error: Cannot open %s.\n", argv[argCounter]);
      errors++;  
   }
#endif


   if((BitStream_File = fopen(argv[++argCounter], "wb"))==NULL)
   {
      fprintf(stderr, "Error: Cannot create %s.\n", argv[argCounter]);
      errors++;  
   }


#ifdef XAFE
   if((Quantised_Pitch_Period_File = fopen(argv[++argCounter], "wb"))==NULL)
   {
      fprintf(stderr, "Error: Cannot create %s.\n", argv[argCounter]);
      errors++;  
   }
#endif



   if((TXT_File = fopen(argv[++argCounter], "w"))==NULL)
   {
      fprintf(stderr, "Error: Cannot create %s.\n", argv[argCounter]);
      errors++;  
   }

 
   if ( strcmp("-freq", argv[++argCounter]) != 0 )
   {
      print_usage();
      exit(1);
   }

   
   sampRate = (unsigned char) atoi(argv[++argCounter]);
   fprintf(stderr, "Sampling Rate: %i kHz\n", sampRate);
   switch(sampRate)
   {  
   
   case 8:   
      sampRate = eight_kHz;
      weight_c0_norm  = weight8kHz_c0_norm;
      weight_c0_shift = weight8kHz_c0_shift; 
      weight_logE = weight8kHz_logE; 
      break;    
    
   case 16:  
      sampRate = sixteen_kHz;              
      weight_c0_norm  = weight16kHz_c0_norm;
      weight_c0_shift = weight16kHz_c0_shift;
      weight_logE = weight16kHz_logE;
      break;    

   case 11:  
   default:  
      fprintf( stderr, "Error: Incorrect sampling rate.\n" );
      errors++;
   }

   

   if (strcmp("-VAD", argv[++argCounter])==0)
   {
      fprintf(stderr, "Running with VAD.\n");
      strcpy(VADFilename, argv[htk_file_arg]);
      if( (ptP= strrchr(VADFilename,'.')) != NULL) 
      {
          *ptP = '\0' ;
      } 
      strcat(VADFilename,".vad");
     
      if((VAD_File = fopen(VADFilename, "r"))==NULL)
      {
         fprintf(stderr, "Error: Cannot open %s\n", VADFilename);
         errors++;  
      }
   }     
   else
   {
      if (strcmp("-No_VAD", argv[argCounter])!=0)
      {
         errors++;
      }
      printf("Running without VAD.\n");
      VAD_File = NULL;
    
   }

  
   if(errors!=0)
   {
      fprintf(stderr, "%d error(s) parsing command-line arguments.\n", errors);
      fprintf(stderr, "Launch executable without any arguments for usage information.\n");
      exit(1);
   }


   return;
}



/*----------------------------------------------------------------------------
 * FUNCTION NAME:
 *  
 *    print_usage
 *
 * PURPOSE:       
 *
 *    Print usage information to stderr.
 *
 * INPUT:         
 *
 *    None.
 *
 * OUTPUT:        
 *
 *    None.
 *
 * RETURN VALUE:  
 *
 *    None.
 *
 *
 *---------------------------------------------------------------------------*/


void print_usage( void )
{

#ifdef XAFE
   fprintf( stderr,
            "Usage: extcoder htk_file_in pitch_file_in class_file_in bitstream_file_out pitch_file_out\n"
            "                txt_file_out -freq x -VAD/No_VAD\n\n"
            "                htk_file_in:     Input mel-frequency cepstral coefficient file. Must\n"
            "                                 be in HTK MFCC format.\n"
            "                pitch_file_in:   Input pitch period file.\n"
            "                class_file_in:   Input classification file.\n"
            "                bit_file_out:    Output binary bitstream.\n"
            "                pitch_file_out:  Output quantised pitch period file.\n"
            "                txt_file_out:    Vector quantiser output in text format.\n"
            "                -freq x:         Sampling frequency in kHz (8, 11, or 16).\n"
            "                -VAD:            Use voice activity detector data. Voice activity input\n"
            "                                 file must have same name as htk_file, but extension .vad\n"
            "                -No_VAD:         Do not incorporate voice activity detector information\n"
            "                                 in output bitstream.\n");
#else
   fprintf( stderr,                
            "Usage: coder htk_file_in bitstream_file_out txt_file_out -freq x -VAD/No_VAD\n\n"
            "             htk_file_in:    Input mel-frequency cepstral coefficient file. Must\n"
            "                             be in HTK MFCC format.\n"
            "             bit_file_out:   Binary output bitstream.\n"
            "             txt_file_out:   Vector quantiser output in text format.\n"
            "             -freq x:        Sampling frequency in kHz (8, 11, or 16).\n"
            "             -VAD:           Use voice activity detector data. Voice activity input\n"
            "                             file must have same name as htk_file, but extension .vad\n"
            "             -No_VAD:        Do not incorporate voice activity detector information\n"
            "                             in output bitstream.\n");
#endif

}



/*----------------------------------------------------------------------------
 *
 *   FUNCTION NAME:
 *
 *      quant_pitch_abs
 *
 *   PURPOSE:
 *
 *      To quantize the pitch period absolutely
 *
 *   INPUTS:
 *
 *      lwPeriod - Input pitch period
 *
 *   OUTPUTS:
 *
 *      *plwQuantPeriod - Quantized pitch period
 *
 *      *pcQPIndex - Index of the quantized pitch period 
 *
 *   RETURN VALUE:
 *
 *      None
 *
 *---------------------------------------------------------------------------*/



#ifdef XAFE
void quant_pitch_abs(Word32 lwPeriod, Word32 *plwQuantPeriod,
                     UWord8 *pcQPIndex)
{


  /*
  ********************************************************************
  *                                                                  *
  *                         Static variables                         *
  *                                                                  *
  ********************************************************************
  */

/*
 * Quantization levels and sections
 */

static const Word32 plwQuantLevels[127] = {
1255013,
1274906,
1295114,
1315641,
1336495,
1357679,
1379198,
1401059,
1423266,
1445825,
1468742,
1492022,
1515671,
1539695,
1564100,
1588891,
1614076,
1639659,
1665648,
1692050,
1718869,
1746114,
1773790,
1801905,
1830466,
1859479,
1888953,
1918893,
1949308,
1980206,
2011592,
2043477,
2075867,
2108770,
2142194,
2176149,
2210641,
2245681,
2281276,
2317435,
2354167,
2391481,
2429387,
2467893,
2507010,
2546747,
2587114,
2628120,
2669777,
2712094,
2755081,
2798750,
2843111,
2888175,
2933954,
2980458,
3027699,
3075689,
3124440,
3173963,
3224271,
3275377,
3327293,
3380032,
3433606,
3488030,
3543316,
3599479,
3656532,
3714489,
3773365,
3833174,
3893931,
3955651,
4018349,
4082041,
4146743,
4212470,
4279239,
4347066,
4415969,
4485963,
4557067,
4629298,
4702674,
4777213,
4852933,
4929854,
5007993,
5087372,
5168008,
5249923,
5333136,
5417668,
5503539,
5590772,
5679388,
5769408,
5860855,
5953751,
6048120,
6143984,
6241369,
6340296,
6440792,
6542880,
6646587,
6751938,
6858958,
6967675,
7078115,
7190305,
7304274,
7420049,
7537659,
7657133,
7778501,
7901793,
8027039,
8154270,
8283517,
8414814,
8548191,
8683683,
8821322,
8961143,
9103179 };


static const Word32 pplwQuantSections[8][3] = {
{1588891,0,15},
{2043477,15,31},
{2628120,31,47},
{3380032,47,63},
{4347066,63,79},
{5590772,79,95},
{7190305,95,111},
{9103179,111,126}};



  /*
  ********************************************************************
  *                                                                  *
  *                        Automatic variables                       *
  *                                                                  *
  ********************************************************************
  */

  Word16 i;
  Word16 j;
  Word16 iLowIndex = 0;
  Word16 iHighIndex = 0;


  /*
  ********************************************************************
  *                                                                  *
  *                         Executable Code                          *
  *                                                                  *
  ********************************************************************
  */

  move16();move16();
  test();
  test();
  test();
  if (lwPeriod <= 0)
  {
    *plwQuantPeriod = 0; move32();
    *pcQPIndex = 0; move16();
  }
  else if (L_sub(lwPeriod,plwQuantLevels[0]) <= 0)
  {
    *plwQuantPeriod = plwQuantLevels[0]; move32();
    *pcQPIndex = 1; move16();
  }
  else if (L_sub(lwPeriod,plwQuantLevels[126]) >= 0)
  {
    *plwQuantPeriod = plwQuantLevels[126]; move32();
    *pcQPIndex = 127; move16();
  }
  else
  {

    for (i = 0; i < 8; i++)
    {
      test();
      if (L_sub(lwPeriod,pplwQuantSections[i][0]) <= 0)
      {
        iLowIndex = extract_l(pplwQuantSections[i][1]);
        iHighIndex = extract_l(pplwQuantSections[i][2]);
        break;
      }
    }

    for (i = iLowIndex; i < iHighIndex; i++)
    {
      j = add(i,1);
      test();test();
      if ((L_sub(lwPeriod,plwQuantLevels[i]) >= 0) &&
          (L_sub(lwPeriod,plwQuantLevels[j]) <= 0))
      {
        test();
        if (L_sub(L_sub(lwPeriod,plwQuantLevels[i]),
                  L_sub(plwQuantLevels[j],lwPeriod)) < 0)
        {
          *plwQuantPeriod = plwQuantLevels[i]; move32();
          *pcQPIndex = add(i,1);
        }
        else
        {
          *plwQuantPeriod = plwQuantLevels[j]; move32();
          *pcQPIndex = add(j,1);
        }
      }
    }

  }

  return;

}
#endif /* void quant_pitch_abs(...) */





/*----------------------------------------------------------------------------
 *
 *   FUNCTION NAME:
 *
 *      quant_pitch_diff
 *
 *   PURPOSE:
 *
 *      To quantize the pitch period differentially
 *
 *   INPUTS:
 *
 *      lwPeriod - Input pitch period
 *
 *      plwQPHistory[0:2] - History of three previous quantized
 *           pitch periods: plwQPHistory[0] corresponds to the
 *           previous frame, plwQPHistory[1] corresponds to one
 *           before the previous frame, and plwQPHistory[2]
 *           corresponds to two frames before the previous frame.
 *
 *   INPUT/OUTPUT:
 *
 *      *piReliableFlag - This flag indicates if the pitch
 *           period of the second frame of a frame-pair has
 *           been quantized differentially using a reference
 *           value from the previous frame pair.  In this
 *           case, the quantized pitch period is considered
 *           unreliable and the flag is set to FALSE.  If the 
 *           pitch period has been absolutely quantized or it
 *           has been quantized differentially using the pitch
 *           period value of the first frame of the same
 *           frame-pair, then the flag is set to TRUE to
 *           indicate that this value is reliable.  This
 *           flag helps to avoid propagating errors through
 *           differential quantization from frame-pair to
 *           frame-pair.  On input, this flag indicates if
 *           plwQPHistory[1] is reliable.  On output, this
 *           flag indicates if the output *pfQuantPeriod is
 *           reliable.  plwQPHistory[0] and plwQPHistory[2]
 *           are always considered reliable since they are
 *           absolutely quantized.
 *
 *   OUTPUTS:
 *
 *      *plwQuantPeriod - Quantized pitch period
 *
 *      *pcQPIndex - Index of the quantized pitch period
 *
 *   RETURN VALUE:
 *
 *      None
 *
 *---------------------------------------------------------------------------*/



#ifdef XAFE
void quant_pitch_diff(Word32 lwPeriod, Word32 *plwQPHistory,
                      Word16 *piReliableFlag, Word32 *plwQuantPeriod,
                      UWord8 *pcQPIndex)
{


  /*
  ********************************************************************
  *                                                                  *
  *                         Static variables                         *
  *                                                                  *
  ********************************************************************
  */

  /*
   * Quantization levels and sections (for absolute quantization)
   */

    static const Word32 plwQuantLevels[31] = {
    1285948,
    1371524,
    1462794,
    1560138,
    1663959,
    1774690,
    1892790,
    2018748,
    2153089,
    2296370,
    2449185,
    2612170,
    2786001,
    2971400,
    3169136,
    3380032,
    3604961,
    3844859,
    4100721,
    4373610,
    4664658,
    4975075,
    5306150,
    5659256,
    6035860,
    6437525,
    6865921,
    7322824,
    7810133,
    8329870,
    8884194 };


    static const Word32 pplwQuantSections[4][3] = {
    {2018748,0,7},
    {3380032,7,15},
    {5659256,15,23},
    {8884194,23,30}};


  /*
   * Ratio threshold table and multi-level threshold index
   * table for Case 2
   */

    static const Word16 pswRatioThld_1[4][6] = {
    {3350, 5013, 8192, 12288, 16384, 20480},
    {3350, 5013, 6144,  8192, 10240, 12288},
    {2048, 2731, 3350,  5013,  6144,  8192},
    {1024, 1365, 2048,  2731,  3350,  5013}
    };
    
  
    static const Word16 piMultiLevelIndex_1[4] = {
    0,
    0,
    2,
    4
  };
    

  static const Word16 swAlpha1 = 4160;


  /*
   * Ratio threshold table and multi-level threshold index
   * table for Cases 3 and 4
   */

  static const Word16 pswRatioThld_2[4][8] = {

    {3182, 5269, 6144, 8192, 10240, 12288, 16384, 20480},
    {2731, 3182, 5269, 6144,  8192, 10240, 12288, 16384},
    {1365, 2048, 2731, 3182,  5269,  6144,  7168,  8192},
    { 819, 1024, 1365, 2048,  2731,  3182,  5269,  6144}

  };


  static const Word16 piMultiLevelIndex_2[4] = {

    0,
    1,
    3,
    5

  };
    

  static const Word16 swAlpha2 = 4183;


  /*
  ********************************************************************
  *                                                                  *
  *                        Automatic variables                       *
  *                                                                  *
  ********************************************************************
  */

  Word16 i;
  Word16 j;
  Word16 iLowIndex = 0;
  Word16 iHighIndex = 0;

  Word16 iNumInterval;
  Word16 iCodeWord;
  Word16 iMultiLevelIndex;
  Word16 iRowIndex;
  Word16 iColIndex;
  Word16 iEndColIndex;

  Word16 swRefHigh;
  Word16 swRefLow;
  Word16 swTemp;
  Word16 swHigh;
  Word16 swLow;

  Word32 lwTemp;
  Word32 lwTempHigh;
  Word32 lwTempLow = 0;
  Word32 lwLowPeriod;
  Word32 lwHighPeriod;
  Word32 lwRefPeriod;


  /*
  ********************************************************************
  *                                                                  *
  *                         Executable Code                          *
  *                                                                  *
  ********************************************************************
  */


  /*
   * First, handle the unvoiced condition
   */
  move16();move16();
  test();
  test();test();test();test();
  test();
  if (lwPeriod <= 0)
  {
    *plwQuantPeriod = 0; move32();
    *pcQPIndex = UNVOICED_CODE; move16();
    *piReliableFlag = FALSE; move16();
    return;
  }


  /*
   * Next, handle case 1 - Absolute quantization
   */

  else if ((plwQPHistory[0] <= 0) &&
           ((plwQPHistory[1] <= 0) || (*piReliableFlag == FALSE)) &&
           (plwQPHistory[2] <= 0))
  {

    test();
    test();
    if (L_sub(lwPeriod,plwQuantLevels[0]) <= 0)
    {
      *plwQuantPeriod = plwQuantLevels[0]; move32();
      *pcQPIndex = 1; move16();
    }
    else if (L_sub(lwPeriod,plwQuantLevels[30]) >= 0)
    {
      *plwQuantPeriod = plwQuantLevels[30]; move32();
      *pcQPIndex = 31; move16();
    }
    else
    {

      for (i = 0; i < 4; i++)
      {
        test();
        if (L_sub(lwPeriod,pplwQuantSections[i][0]) <= 0)
        {
          iLowIndex = extract_l(pplwQuantSections[i][1]);
          iHighIndex = extract_l(pplwQuantSections[i][2]);
          break;
        }
      }
 
      for (i = iLowIndex; i < iHighIndex; i++)
      {
        j = add(i,1);
        test();test();
        if ((L_sub(lwPeriod,plwQuantLevels[i]) >= 0) &&
            (L_sub(lwPeriod,plwQuantLevels[j]) <= 0))
        {
          test();
          if (L_sub(L_sub(lwPeriod,plwQuantLevels[i]),
                    L_sub(plwQuantLevels[j],lwPeriod)) < 0)
          {
            *plwQuantPeriod = plwQuantLevels[i]; move32();
            *pcQPIndex = add(i,1);
          }
          else
          {
            *plwQuantPeriod = plwQuantLevels[j]; move32();
            *pcQPIndex = add(j,1);
          }
        }
      }

    }

    *piReliableFlag = TRUE; move16();

    return;

  }


  /*
   * Next, handle case 2 - Differential quantization with
   * plwQPHistory[0] as reference
   */

  else if (plwQPHistory[0] > 0)
  {

    lwRefPeriod = plwQPHistory[0]; move32();
    lwTemp = L_shl(lwRefPeriod,7);
    swRefHigh = extract_h(lwTemp);
    swRefLow = extract_l(L_shr(lwTemp,1) & 0x7FFF); logic16();

    /*
     * Establish the appropriate row of ratio thresholds to use
     * based on the reference pitch period
     */

    test();test();test();
    if (L_sub(lwRefPeriod,1966080) <= 0)
    {
      iRowIndex = 0; move16();
    }
    else if (L_sub(lwRefPeriod,3932160) <= 0)
    {
      iRowIndex = 1; move16();
    }
    else if (L_sub(lwRefPeriod,6225920) <= 0)
    {
      iRowIndex = 2; move16();
    }
    else
    {
      iRowIndex = 3; move16();
    }

    iColIndex = 0; move16();
    iEndColIndex = 5; move16();
    iMultiLevelIndex = piMultiLevelIndex_1[iRowIndex]; move16();
    iCodeWord = 1; move16();


    /*
     * Quantize the pitch period
     */

    /*
     * First, check the thresholds below the multi-level threshold
     * to quantize the pitch period
     */

    test();
    if (sub(iColIndex,iMultiLevelIndex) < 0)
    {

      swTemp = pswRatioThld_1[iRowIndex][iColIndex]; move16();
      lwLowPeriod = L_shr(L_add(L_shl(L_mult(swRefHigh,swTemp),3),
                                L_shr(L_mult(swRefLow,swTemp),13)),7);

      test();
      if (L_sub(lwPeriod,lwLowPeriod) < 0)
      {
        *plwQuantPeriod = lwLowPeriod; move32();
        *pcQPIndex = iCodeWord; move16();
        *piReliableFlag = TRUE; move16();
        return;
      }

      test();
      while (sub(iColIndex,iMultiLevelIndex) < 0)
      {

        swTemp = pswRatioThld_1[iRowIndex][iColIndex+1]; move16();
        lwHighPeriod = L_shr(L_add(L_shl(L_mult(swRefHigh,swTemp),3),
                                   L_shr(L_mult(swRefLow,swTemp),13)),7);

        test();
        if (L_sub(lwPeriod,lwHighPeriod) < 0)
        {

          test();
          if (L_sub(L_sub(lwPeriod,lwLowPeriod),
                    L_sub(lwHighPeriod,lwPeriod)) < 0)
          {
            *plwQuantPeriod = lwLowPeriod; move32();
            *pcQPIndex = iCodeWord; move16();
          }
          else
          {
            *plwQuantPeriod = lwHighPeriod; move32();
            *pcQPIndex = add(iCodeWord,1);
          }

          *piReliableFlag = TRUE; move16();

          return;

        }

        iCodeWord = add(iCodeWord,1);
        iColIndex = add(iColIndex,1);
        lwLowPeriod = lwHighPeriod; move32();

        test();
      }

    }


    /*
     * Next, check the multi-level threshold to quantize
     * the pitch period
     */

    swTemp = pswRatioThld_1[iRowIndex][iColIndex]; move16();
    lwLowPeriod = L_shr(L_add(L_shl(L_mult(swRefHigh,swTemp),3),
                              L_shr(L_mult(swRefLow,swTemp),13)),7);

    test();
    if (L_sub(lwLowPeriod,MIN_PERIOD) < 0)
    {
      lwLowPeriod = MIN_PERIOD; move32();
    }
    test();
    if (L_sub(lwPeriod,lwLowPeriod) < 0)
    {
      *plwQuantPeriod = lwLowPeriod; move32();
      *pcQPIndex = iCodeWord; move16();
      *piReliableFlag = TRUE; move16();
      return;
    }

    swTemp = pswRatioThld_1[iRowIndex][iColIndex+1]; move16();
    lwHighPeriod = L_shr(L_add(L_shl(L_mult(swRefHigh,swTemp),3),
                               L_shr(L_mult(swRefLow,swTemp),13)),7);

    test();
    if (L_sub(lwHighPeriod,MAX_PERIOD) > 0)
    {
      lwHighPeriod = MAX_PERIOD; move32();
    }
    test();test();
    if ((L_sub(lwHighPeriod,MAX_PERIOD) == 0) && 
        (L_sub(lwPeriod,lwHighPeriod) >= 0))
    {
      *plwQuantPeriod = lwHighPeriod; move32();
      *pcQPIndex = add(iCodeWord,NUM_MULTI_LEVELS_1);
      *piReliableFlag = TRUE; move16();
      return;
    }

    test();
    if (L_sub(lwPeriod,lwHighPeriod) < 0)
    {
      lwTempHigh = lwLowPeriod; move32();
      iNumInterval = 0; move16();
      test();test();
      while ((L_sub(lwTempHigh,lwPeriod) <= 0) &&
             (sub(iNumInterval,NUM_MULTI_LEVELS_1) < 0))
      {
        lwTempLow = lwTempHigh; move32();
        lwTempHigh = L_shl(lwTempHigh,7);
        swHigh = extract_h(lwTempHigh);
        swLow = extract_l(L_shr(lwTempHigh,1) & 0x7FFF); logic16();
        lwTempHigh = L_shr(L_add(L_shl(L_mult(swHigh,swAlpha1),3),
                                 L_shr(L_mult(swLow,swAlpha1),13)),7);
        iNumInterval = add(iNumInterval,1);
        test();test();
      }

      test();
      if (sub(iNumInterval,NUM_MULTI_LEVELS_1) == 0)
      {
        lwTempHigh = lwHighPeriod; move32();
      }
      test();
      if (L_sub(L_sub(lwTempHigh,lwPeriod),L_sub(lwPeriod,lwTempLow)) <= 0)
      {
        *plwQuantPeriod = lwTempHigh; move32();
        iCodeWord = add(iCodeWord,iNumInterval);
      }
      else
      {
        *plwQuantPeriod = lwTempLow; move32();
        iCodeWord = sub(add(iCodeWord,iNumInterval),1);
      }

      *pcQPIndex = iCodeWord; move16();
      *piReliableFlag = TRUE; move16();

      return;

    }
    else
    {
      iCodeWord = add(iCodeWord,NUM_MULTI_LEVELS_1);
      iColIndex = add(iColIndex,1);
      lwLowPeriod = lwHighPeriod; move32();
    }


    /*
     * Finally, check the thresholds above the multi-level
     * threshold to quantize the pitch period
     */

    test();
    while (sub(iColIndex,iEndColIndex) < 0)
    {

      swTemp = pswRatioThld_1[iRowIndex][iColIndex+1]; move16();
      lwHighPeriod = L_shr(L_add(L_shl(L_mult(swRefHigh,swTemp),3),
                                 L_shr(L_mult(swRefLow,swTemp),13)),7);

      test();
      if (L_sub(lwHighPeriod,MAX_PERIOD) > 0)
      {
        lwHighPeriod = MAX_PERIOD; move32();
      }
      test();test();
      if ((L_sub(lwHighPeriod,MAX_PERIOD) == 0) &&
          (L_sub(lwPeriod,lwHighPeriod) >= 0))
      {
        *plwQuantPeriod = lwHighPeriod; move32();
        *pcQPIndex = add(iCodeWord,1);
        *piReliableFlag = TRUE; move16();
        return;
      }

      test();
      if (L_sub(lwPeriod,lwHighPeriod) < 0)
      {
        test();
        if (L_sub(L_sub(lwPeriod,lwLowPeriod),L_sub(lwHighPeriod,lwPeriod)) < 0)
        {
          *plwQuantPeriod = lwLowPeriod; move32();
          *pcQPIndex = iCodeWord; move16();
        }
        else
        {
          *plwQuantPeriod = lwHighPeriod; move32();
          *pcQPIndex = add(iCodeWord,1);
        }

        *piReliableFlag = TRUE; move16();

        return;
      }

      iCodeWord = add(iCodeWord,1);
      iColIndex = add(iColIndex,1);
      lwLowPeriod = lwHighPeriod; move32();

      test();
    }

    *plwQuantPeriod = lwLowPeriod; move32();
    *pcQPIndex = iCodeWord; move16();
    *piReliableFlag = TRUE; move16();

    return;

  }


  /*
   * Next, handle cases 3 and 4 - Differential quantization
   * with plwQPHistory[1] and plwQPHistory[2] as references
   */

  else
  {

    test();test();
    if ((plwQPHistory[1] > 0) && (*piReliableFlag == TRUE))
    {
      lwRefPeriod = plwQPHistory[1]; move32();
    }
    else
    {
      lwRefPeriod = plwQPHistory[2]; move32();
    }
    lwTemp = L_shl(lwRefPeriod,7);
    swRefHigh = extract_h(lwTemp);
    swRefLow = extract_l(L_shr(lwTemp,1) & 0x7FFF); logic16();


    /*
     * Establish the appropriate row of ratio thresholds to use
     * based on the reference pitch period
     */

    test();test();test();
    if (L_sub(lwRefPeriod,1966080) <= 0)
    {
      iRowIndex = 0; move16();
    }
    else if (L_sub(lwRefPeriod,3932160) <= 0)
    {
      iRowIndex = 1; move16();
    }
    else if (L_sub(lwRefPeriod,6225920) <= 0)
    {
      iRowIndex = 2; move16();
    }
    else
    {
      iRowIndex = 3; move16();
    }

    iColIndex = 0; move16();
    iEndColIndex = 7; move16();
    iMultiLevelIndex = piMultiLevelIndex_2[iRowIndex]; move16();
    iCodeWord = 1; move16();


    /*
     * Quantize the pitch period
     */

    /*
     * First, check the thresholds below the multi-level threshold
     * to quantize the pitch period
     */

    test();
    if (sub(iColIndex,iMultiLevelIndex) < 0)
    {

      swTemp = pswRatioThld_2[iRowIndex][iColIndex]; move16();
      lwLowPeriod = L_shr(L_add(L_shl(L_mult(swRefHigh,swTemp),3),
                                L_shr(L_mult(swRefLow,swTemp),13)),7);

      test();
      if (L_sub(lwPeriod,lwLowPeriod) < 0)
      {
        *plwQuantPeriod = lwLowPeriod; move32();
        *pcQPIndex = iCodeWord; move16();
        *piReliableFlag = FALSE; move16();
        return;
      }

      test();
      while (sub(iColIndex,iMultiLevelIndex) < 0)
      {

        swTemp = pswRatioThld_2[iRowIndex][iColIndex+1]; move16();
        lwHighPeriod = L_shr(L_add(L_shl(L_mult(swRefHigh,swTemp),3),
                                   L_shr(L_mult(swRefLow,swTemp),13)),7);

        test();
        if (L_sub(lwPeriod,lwHighPeriod) < 0)
        {

          test();
          if (L_sub(L_sub(lwPeriod,lwLowPeriod),
                    L_sub(lwHighPeriod,lwPeriod)) < 0)
          {
            *plwQuantPeriod = lwLowPeriod; move32();
            *pcQPIndex = iCodeWord; move16();
          }
          else
          {
            *plwQuantPeriod = lwHighPeriod; move32();
            *pcQPIndex = add(iCodeWord,1);
          }

          *piReliableFlag = FALSE; move16();

          return;

        }

        iCodeWord = add(iCodeWord,1);
        iColIndex = add(iColIndex,1);
        lwLowPeriod = lwHighPeriod; move32();

        test();
      }

    }


    /*
     * Next, check the multi-level threshold to quantize
     * the pitch period
     */

    swTemp = pswRatioThld_2[iRowIndex][iColIndex]; move16();
    lwLowPeriod = L_shr(L_add(L_shl(L_mult(swRefHigh,swTemp),3),
                              L_shr(L_mult(swRefLow,swTemp),13)),7);

    test();
    if (L_sub(lwLowPeriod,MIN_PERIOD) < 0)
    {
      lwLowPeriod = MIN_PERIOD; move32();
    }
    test();
    if (L_sub(lwPeriod,lwLowPeriod) < 0)
    {
      *plwQuantPeriod = lwLowPeriod; move32();
      *pcQPIndex = iCodeWord; move16();
      *piReliableFlag = FALSE; move16();
      return;
    }

    swTemp = pswRatioThld_2[iRowIndex][iColIndex+1]; move16();
    lwHighPeriod = L_shr(L_add(L_shl(L_mult(swRefHigh,swTemp),3),
                               L_shr(L_mult(swRefLow,swTemp),13)),7);

    test();
    if (L_sub(lwHighPeriod,MAX_PERIOD) > 0)
    {
      lwHighPeriod = MAX_PERIOD; move32();
    }
    test();test();
    if ((L_sub(lwHighPeriod,MAX_PERIOD) == 0) &&
        (L_sub(lwPeriod,lwHighPeriod) >= 0))
    {
      *plwQuantPeriod = lwHighPeriod; move32();
      *pcQPIndex = add(iCodeWord,NUM_MULTI_LEVELS_2);
      *piReliableFlag = FALSE; move16();
      return;
    }

    test();
    if (L_sub(lwPeriod,lwHighPeriod) < 0)
    {
      lwTempHigh = lwLowPeriod; move32();
      iNumInterval = 0; move16();
      test();test();
      while ((L_sub(lwTempHigh,lwPeriod) <= 0) &&
             (sub(iNumInterval,NUM_MULTI_LEVELS_2) < 0))
      {
        lwTempLow = lwTempHigh; move32();
        lwTempHigh = L_shl(lwTempHigh,7);
        swHigh = extract_h(lwTempHigh);
        swLow = extract_l(L_shr(lwTempHigh,1) & 0x7FFF); logic16();
        lwTempHigh = L_shr(L_add(L_shl(L_mult(swHigh,swAlpha2),3),
                                 L_shr(L_mult(swLow,swAlpha2),13)),7);
        iNumInterval = add(iNumInterval,1);
        test();test();
      }
      test();
      if (sub(iNumInterval,NUM_MULTI_LEVELS_2) == 0)
      {
        lwTempHigh = lwHighPeriod; move32();
      }
      test();
      if (L_sub(L_sub(lwTempHigh,lwPeriod),L_sub(lwPeriod,lwTempLow)) <= 0)
      {
        *plwQuantPeriod = lwTempHigh; move32();
        iCodeWord = add(iCodeWord,iNumInterval);
      }
      else
      {
        *plwQuantPeriod = lwTempLow; move32();
        iCodeWord = sub(add(iCodeWord,iNumInterval),1);
      }

      *pcQPIndex = iCodeWord; move16();
      *piReliableFlag = FALSE; move16();

      return;

    }
    else
    {
      iCodeWord = add(iCodeWord,NUM_MULTI_LEVELS_2);
      iColIndex = add(iColIndex,1);
      lwLowPeriod = lwHighPeriod; move32();
    }


    /*
     * Finally, check the thresholds above the multi-level
     * threshold to quantize the pitch period
     */

    test();
    while (sub(iColIndex,iEndColIndex) < 0)
    {

      swTemp = pswRatioThld_2[iRowIndex][iColIndex+1]; move16();
      lwHighPeriod = L_shr(L_add(L_shl(L_mult(swRefHigh,swTemp),3),
                                 L_shr(L_mult(swRefLow,swTemp),13)),7);

      test();
      if (L_sub(lwHighPeriod,MAX_PERIOD) > 0)
      {
        lwHighPeriod = MAX_PERIOD; move32();
      }
      test();test();
      if ((L_sub(lwHighPeriod,MAX_PERIOD) == 0) &&
          (L_sub(lwPeriod,lwHighPeriod) >= 0))
      {
        *plwQuantPeriod = lwHighPeriod; move32();
        *pcQPIndex = add(iCodeWord,1);
        *piReliableFlag = FALSE; move16();
        return;
      }

      test();
      if (L_sub(lwPeriod,lwHighPeriod) < 0)
      {
        test();
        if (L_sub(L_sub(lwPeriod,lwLowPeriod),L_sub(lwHighPeriod,lwPeriod)) < 0)
        {
          *plwQuantPeriod = lwLowPeriod; move32();
          *pcQPIndex = iCodeWord; move16();
        }
        else
        {
          *plwQuantPeriod = lwHighPeriod; move32();
          *pcQPIndex = add(iCodeWord,1);
        }

        *piReliableFlag = FALSE; move16();

        return;
      }

      iCodeWord = add(iCodeWord,1);
      iColIndex = add(iColIndex,1);
      lwLowPeriod = lwHighPeriod; move32();

      test();
    }

    *plwQuantPeriod = lwLowPeriod; move32();
    *pcQPIndex = iCodeWord; move16();
    *piReliableFlag = FALSE; move16();

    return;

  }

}
#endif /* void quant_pitch_diff(...) */


/*----------------------------------------------------------------------------
 *
 *   FUNCTION NAME:
 *
 *      get_class_bit  
 *
 *   PURPOSE:
 *
 *      To quantize the class information
 *
 *   INPUTS:
 *
 *      class - input class
 *
 *   OUTPUTS:
 *
 *      class_index - output class bit
 *
 *   RETURN VALUE:
 *
 *      None
 *
 *
 *---------------------------------------------------------------------------*/


#ifdef XAFE
void get_class_bit(UWord8 class, UWord8 *class_index)
{


  test();test();test();test();
  switch (class)
  {

    case '0':  *class_index = 0x00; move16();
               break;

    case '1':  *class_index = 0x01; move16();
               break;

    case '2':  *class_index = 0x00; move16();
               break;
  
    case '3':  *class_index = 0x01; move16();
               break;

    default:  fprintf( stderr, "Unrecognized class!\n");
              exit(1);

  }


  return;

}
#endif



/*----------------------------------------------------------------------------
 *
 *   FUNCTION NAME:
 *
 *      Swap16         
 *
 *   PURPOSE:
 *
 *      Little - big endian conversion for 16 bit integer.
 *
 *   INPUTS:
 *
 *      *Short - an 16 bit integer
 *
 *   OUTPUTS:
 *
 *      *Short - endian converted input integer
 *
 *   RETURN VALUE:
 *
 *      None
 *
 *
 *---------------------------------------------------------------------------*/



void Swap16 ( Word16 *Short )
{
  *Short = ((*Short&0x0000ffL)<<8 )|   \
           ((*Short&0x00ff00L)>>8 );
}




/*----------------------------------------------------------------------------
 *
 *   FUNCTION NAME:
 *
 *      Swap32         
 *
 *   PURPOSE:
 *
 *      Little - big endian conversion for 32 bit integer.
 *
 *   INPUTS:
 *
 *      *Long - an 32 bit integer
 *
 *   OUTPUTS:
 *
 *      *Long - endian converted input integer
 *
 *   RETURN VALUE:
 *
 *      None
 *
 *
 *---------------------------------------------------------------------------*/




void Swap32 ( Word32 *Long )
{
  *Long = ((*Long&0x000000ffL)<<24 )|   \
          ((*Long&0x0000ff00L)<<8  )|   \
          ((*Long&0x00ff0000L)>>8  )|   \
          ((*Long&0xff000000L)>>24 );
}


