/*=======================================================================================================*/
/*      Distributed Speech Recognition                                                                   */
/*      ETSI ES 202 050 Advanced Front-End Feature Extraction Algorithm & Compression Algorithm &        */
/*      ETSI ES 202 212 Extended Advanced Front-End Feature Extraction Algorithm & Compression Algorithm */  
/*      Bit exact C-language software implementation using ETSI basic operations library                 */
/*      3GPP TS 26.243 v 6.1.0 November 2004                                                                  */
/*      ANSI C code for the Fixed-Point Distributed Speech Recognition Extended Advanced Front-end       */
/*=======================================================================================================*/
/*-------------------------------------------------------------------------------
 *
 * FILE NAME: decoder.c
 * PURPOSE:   This function is the main program for the decoding of the 
 *	      Motorola, France Telecom, Alcatel Advanced Front End Proposal.  
 *	      It performs decompression and error mitigation of the binary bitstream
 *            format for the DSR standard.
 *
 * INPUT:     decoder reads files containing the binary DSR bitstream format 
 *
 * OUTPUT:    decoder creates an HTK file format files containing the Mel-Cepstrum parameters 
 *	      after decompression containing both C0 and logE, it also creates an aditional
 *	      ascii file containing the VAD flag.
 *
 *-------------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "typedef.h"
#include "common.h"
#include "decoder.h"
#include "golay.h"

/*include the quantiser tables*/
#include "quantiser_16kHz_BEC.tab"
#include "quantiser_8kHz_BEC.tab"


FILE *BitStream_File, *HTK_Quantised_Format_File, *VAD_File;
#ifdef XAFE
FILE *Quantised_Pitch_Period_File, *Classification_File;
FILE *Error_Status_File;
#endif


int dataframes=0; /*Global only for ease of printf statements showing warnings*/


HTK_DataFrame SD, 
   DETECTION_THRESHOLD;


unsigned char sampRate; /* Can be 00 (8), 01 (11) or 11 (16). Extracted from input bitfile. */
#ifdef WMOPS
float FrameRateFactor = VQFRAMERATEFACTOR;
#endif


int main(int argc, char *argv[])
{
   HTK_Header header;

   dframe PreviousDFrame, 
      CurrentDFrame, 
      LastGoodDFrame, 
      *BufferedDFrames;

   mframe *BufferedMultiFrames;

   golaydata *BufferedGolays;

   short int GoodGolay = -1;

   int i=0,
      frame=0,
      result,
      done,
      frames_buffered,
      dataframes_buffered,
      endFrames,
      buffering_data_mode,
      start_of_data_stream;


   process_params(argc, argv);

   init_thresholds();

   header.nSamples = 0; /*Unknown - will be updated at end by dataframes variable*/

   header.sampPeriod = EXPECTED_SAMP_PERIOD;
   header.sampSize = EXPECTED_SAMP_SIZE;
   header.parmKind = EXPECTED_PARM_KIND;

   write_htk_header(&header);
   /*Will be overwritten later*/


   fprintf( stderr, "Dequantising bitstream and creating output files....\n\n");


 
   BufferedMultiFrames = NULL;
   BufferedGolays = NULL;
   buffering_data_mode = TRUE;
   start_of_data_stream = TRUE;
   done = FALSE;
   dataframes_buffered = 0;
   dataframes = 0;
   /************************************************/
   /* Initialization of BufferedDFrames missing!   */
   /* Guenter 23.11.00                             */
   /************************************************/
   BufferedDFrames = NULL;

   while(!done)
   {
      frames_buffered = 1;

      if(get_next_frame(&BufferedMultiFrames, frames_buffered)==EOF)
      {
         frames_buffered--;

         if(start_of_data_stream) /*Couldn't find a good frame to work with.*/
         {
            if(dataframes_buffered > 0) /*incase this file was in fact empty.*/
            {      
#ifdef XAFE
               error_correct_and_write(NULL, 0, NULL, 0, BufferedDFrames, dataframes_buffered); 
#else
               error_correct_and_write(NULL, NULL, BufferedDFrames, dataframes_buffered); 
#endif
               free(BufferedDFrames); BufferedDFrames=NULL;
          
               dataframes_buffered = 0;
            }
         }
         else /*Have finished exactly at the end of a multiframe, and thus still have previous data to put out.*/
         {
            if(buffering_data_mode)
            {
#ifdef XAFE
               error_correct_and_write(&LastGoodDFrame, 1, &LastGoodDFrame, 1, BufferedDFrames, dataframes_buffered); 
#else
               error_correct_and_write(LastGoodDFrame.b, LastGoodDFrame.b, BufferedDFrames, dataframes_buffered); 
#endif
               free(BufferedDFrames); BufferedDFrames=NULL;
          
               dataframes_buffered = 0;
            }
            else
            {
#ifdef XAFE
               write_two_data_frames(PreviousDFrame, TWO_FRAMES, '0');
#else
               write_two_data_frames(PreviousDFrame, TWO_FRAMES);
#endif          
            }


         }
         done = TRUE;
      }
      else /*There are more frames.*/
      {

         get_next_golay(&BufferedGolays, BufferedMultiFrames[frames_buffered - 1], 1); /* the 1 means there is only one golay in buffer now.*/
         /*frames_buffered - 1 will always equal zero here. Included for niceness.*/

         if(GoodGolay == -1) /*i.e. havn't found two good golays that match yet*/
         {

            if(find_good_golay(&BufferedMultiFrames, &BufferedGolays, &GoodGolay, &frames_buffered)) /*Ran out of file before finding a good golay*/
            {
               /******************************************************/
               /* The following loop was moved into the subroutine   */
               /* find_good_golay. Running the loop here over the    */
               /* number of buffered frames is WRONG because there   */
               /* may not exist the same number of buffered Golays   */
               /* than number of buffered frames!                    */
               /*  Guenter 23.11.00                                  */
               /******************************************************/
               /*for(i=0; i<(frames_buffered - 1); i++)
                 BufferedGolays[i].data = GoodGolay;*/ 
            }
            else
            {
               fprintf( stderr, "Warning: Was unable to find a good golay header before end of file!\n" );

               /*Just going to continue and put out all the data as it is.*/
            }



         }
         else /*There is a GoodGolay that has been found, i.e. GoodGolay != -1, only have one buffered golay and MultiFrame here*/
         {

            /*if error, or,  if the data is not the same as the previous golays print a warning but keep the previous header */
            /*The 0x078 mask fixes the bits concerning the header case MULTIFRAMER COUNTER to 1's */
            /*Otherwise the comparation between sucessive frames is always different	*/
            if((golay_errors(&(BufferedGolays[0].data), &(BufferedGolays[0].crc))) || (((0x0078 | BufferedGolays[0].data) ^ (0x0078 | GoodGolay))  != 0x0))
               if((golay_errors(&(BufferedGolays[0].data), &(BufferedGolays[0].crc))) || ((BufferedGolays[0].data ^ GoodGolay)  != 0x0))
               {
                  fprintf( stderr, "Warning: An error found in multiframe header: %3d\n", (dataframes/23));

                  BufferedGolays[0].data = GoodGolay;
               }
         }
     

         frame=0; 


         /******************************************************************************/
         /* Processing of Multiframe headers completed and have either found a         */
         /* good Golay header or we have reached the end of the file.                  */
         /*                                                                            */
         /* The next code blocks process the frame packet stream to unquantise the     */
         /* good data frames which have been received without error and handle the     */
         /* data frames received with errors by applying nearest frame error mitigation */
         /******************************************************************************/


         if(start_of_data_stream) /*Try and get first good dataframe.*/
         {
            buffering_data_mode=TRUE; 

            while((frame < (frames_buffered-1)) && (start_of_data_stream))
            {
               i=0;
               while((i<24) && (start_of_data_stream))
               {
                  result = get_two_data_frames(&CurrentDFrame, BufferedMultiFrames[frame], i);
                  dataframes += 2;

                  if(buffering_data_mode)
                  {
                     if(result & 0x4) /*there is a crc error*/
                     {
                        dataframes_buffered += 2;
                        add_to_buffer(&BufferedDFrames, &CurrentDFrame, dataframes_buffered);
                     }
                     else
                     {
                        memcpy(&PreviousDFrame , &CurrentDFrame, sizeof(struct dframe));
                        buffering_data_mode = FALSE;
                     }
                  }
                  else
                  {
                     if(result & 0x4)/*crc is in error*/
                     {
                        dataframes -= 4; /*Due to the following threshold check.*/
                        if(threshold_error(PreviousDFrame))
                        {
                           dataframes += 4;

                           dataframes_buffered += 2;
                           add_to_buffer(&BufferedDFrames, &PreviousDFrame, dataframes_buffered);

                           dataframes_buffered += 2;
                           add_to_buffer(&BufferedDFrames, &CurrentDFrame, dataframes_buffered);

                           buffering_data_mode = TRUE;
                        }
                        else
                        {
                           dataframes += 4;
                  
                           if(dataframes_buffered>0)
                           {
#ifdef XAFE
                              error_correct_and_write(&PreviousDFrame, 0, &PreviousDFrame, 0, BufferedDFrames, dataframes_buffered); 
#else
                              error_correct_and_write(PreviousDFrame.a, PreviousDFrame.a, BufferedDFrames, dataframes_buffered); 
#endif
                              /*i.e. frame repeat backwards*/
                              free(BufferedDFrames); BufferedDFrames = NULL;
                              dataframes_buffered = 0;
                           }
                           
#ifdef XAFE
                           write_two_data_frames(PreviousDFrame, TWO_FRAMES, '0');
#else
                           write_two_data_frames(PreviousDFrame, TWO_FRAMES);
#endif

                           memcpy(&LastGoodDFrame, &PreviousDFrame, sizeof(struct dframe));

                           dataframes_buffered += 2;
                           add_to_buffer(&BufferedDFrames, &CurrentDFrame, dataframes_buffered);

                           buffering_data_mode = TRUE;
                           start_of_data_stream = FALSE;
                        }
                     }
                     else
                     {
                        if(dataframes_buffered>0)
                        {
#ifdef XAFE
                           error_correct_and_write(&PreviousDFrame, 0, &PreviousDFrame, 0, BufferedDFrames, dataframes_buffered); 
#else
                           error_correct_and_write(PreviousDFrame.a, PreviousDFrame.a, BufferedDFrames, dataframes_buffered); 
#endif
                           /*i.e. frame repeat backwards*/
                           free(BufferedDFrames); BufferedDFrames = NULL;
                           dataframes_buffered = 0;
                        }

#ifdef XAFE
                        write_two_data_frames(PreviousDFrame, TWO_FRAMES, '0');
#else
                        write_two_data_frames(PreviousDFrame, TWO_FRAMES);
#endif

                        memcpy(&LastGoodDFrame, &PreviousDFrame, sizeof(struct dframe));

                        memcpy(&PreviousDFrame, &CurrentDFrame, sizeof(struct dframe));

                        buffering_data_mode = FALSE;

                        start_of_data_stream=FALSE;
                     }
                  }
                  i+=2;
               }
               if(start_of_data_stream) frame++; /*avoids incrementing frame when coming out of the loop.*/
            }
         }
         else
            i=0;











         /* main loop of error mitigation code */
      
         if(!start_of_data_stream)
         {

            while(frame < frames_buffered)
            {
               /*  Note: Do not need to set data frame counter (i) here, since the if statement in the initialisation function above has 
                   i = 0;
               */

               while((!done) && (i<24))
               {

   
                  endFrames = get_two_data_frames(&CurrentDFrame, BufferedMultiFrames[frame], i);


                  switch(endFrames)
                  {
                  case 0:  /*For case 0 and 4, there is no data in the current frame (0 padding detected) */
                  case 4:  /*thus as there is no data, whether the frame had a crc error or not is irrelevant*/
                     if(buffering_data_mode)
                     { 
#ifdef XAFE
                        error_correct_and_write(&LastGoodDFrame, 1, &LastGoodDFrame, 1, BufferedDFrames, dataframes_buffered); /*Just have to frame repeat*/
#else
                        error_correct_and_write(LastGoodDFrame.b, LastGoodDFrame.b, BufferedDFrames, dataframes_buffered); /*Just have to frame repeat*/
#endif
                        free(BufferedDFrames); BufferedDFrames = NULL;
                        dataframes_buffered = 0;

                     }
                     else
                     {
#ifdef XAFE
                        write_two_data_frames(PreviousDFrame, TWO_FRAMES, '0');
#else
                        write_two_data_frames(PreviousDFrame, TWO_FRAMES);
#endif
                     }
                         
                     done = TRUE; /*Think that there were no more frames*/
                     i++;
                     break;

                     /*crc is ok, but only have one HTK frame to write.*/
                  case 1:  
                     dataframes++;

                     if(buffering_data_mode) /*i.e. last one was crc errored.*/
                     {
                        if(threshold_error(CurrentDFrame))
                        {
                           dataframes_buffered++;
                           add_to_buffer(&BufferedDFrames, &CurrentDFrame, dataframes_buffered);
#ifdef XAFE
                           error_correct_and_write(&LastGoodDFrame, 1, &LastGoodDFrame, 1, BufferedDFrames, dataframes_buffered);
#else
                           error_correct_and_write(LastGoodDFrame.b, LastGoodDFrame.b, BufferedDFrames, dataframes_buffered);
#endif
                           /*Frame repeat*/
                           free(BufferedDFrames); BufferedDFrames = NULL;
                           dataframes_buffered = 0;
                        }
                        else
                        {
                           if(dataframes_buffered>0)
                           {
#ifdef XAFE
                              error_correct_and_write(&LastGoodDFrame, 1, &CurrentDFrame, 0, BufferedDFrames, dataframes_buffered);
#else
                              error_correct_and_write(LastGoodDFrame.b, CurrentDFrame.a, BufferedDFrames, dataframes_buffered);
#endif
                              free(BufferedDFrames); BufferedDFrames = NULL;
                              dataframes_buffered = 0;
                           }
#ifdef XAFE
                           write_two_data_frames(PreviousDFrame, TWO_FRAMES, '0');
#else
                           write_two_data_frames(CurrentDFrame, ONE_FRAME); /*Think this is last frame*/
#endif
                        }
                     }
                     else
                     {
                        if(dataframes_buffered>0)
                        {
#ifdef XAFE                           
                           error_correct_and_write(&LastGoodDFrame, 1, &PreviousDFrame, 0, BufferedDFrames, dataframes_buffered);
#else                             
                           error_correct_and_write(LastGoodDFrame.b, PreviousDFrame.a, BufferedDFrames, dataframes_buffered);
#endif
                           free(BufferedDFrames); BufferedDFrames = NULL;
                           dataframes_buffered = 0;
                        }
#ifdef XAFE
                        write_two_data_frames(PreviousDFrame, TWO_FRAMES, '0');
                        write_two_data_frames(CurrentDFrame, ONE_FRAME, '0'); /*Think this is last frame*/
#else
                        write_two_data_frames(PreviousDFrame, TWO_FRAMES);
                        write_two_data_frames(CurrentDFrame, ONE_FRAME); /*Think this is last frame*/
#endif
                     }

                     done = TRUE;
                     i++;
                     break;


                
                  case 5:  /*bad crc and only one HTK Frame to write*/
                     dataframes++;

                     if(buffering_data_mode)
                     {
                        dataframes_buffered++;
                        add_to_buffer(&BufferedDFrames, &CurrentDFrame, dataframes_buffered);
#ifdef XAFE
                        error_correct_and_write(&LastGoodDFrame, 1, &LastGoodDFrame, 1, BufferedDFrames, dataframes_buffered);      
#else
                        error_correct_and_write(LastGoodDFrame.b, LastGoodDFrame.b, BufferedDFrames, dataframes_buffered);
#endif
                        free(BufferedDFrames); BufferedDFrames = NULL;
                        dataframes_buffered = 0;
                     }
                     else
                     {
                        if(threshold_error(PreviousDFrame))
                        {
                           dataframes_buffered+=2;
                           add_to_buffer(&BufferedDFrames, &PreviousDFrame, dataframes_buffered);

                           dataframes_buffered++;
                           add_to_buffer(&BufferedDFrames, &CurrentDFrame, dataframes_buffered);

#ifdef XAFE
                           error_correct_and_write(&LastGoodDFrame, 1, &LastGoodDFrame, 1, BufferedDFrames, dataframes_buffered);
#else
                           error_correct_and_write(LastGoodDFrame.b, LastGoodDFrame.b, BufferedDFrames, dataframes_buffered);
#endif
                           free(BufferedDFrames); BufferedDFrames = NULL;
                           dataframes_buffered = 0;
                        }
                        else
                        {
                           if(dataframes_buffered > 0)
                           {
#ifdef XAFE
                              error_correct_and_write(&LastGoodDFrame, 1, &PreviousDFrame, 0, BufferedDFrames, dataframes_buffered);
#else
                              error_correct_and_write(LastGoodDFrame.b, PreviousDFrame.a, BufferedDFrames, dataframes_buffered);
#endif
                              free(BufferedDFrames); BufferedDFrames = NULL;
                              dataframes_buffered = 0;
                           }
#ifdef XAFE
                           write_two_data_frames(PreviousDFrame, TWO_FRAMES, '0');
#else
                           write_two_data_frames(PreviousDFrame, TWO_FRAMES);
#endif
                           memcpy(&LastGoodDFrame, &PreviousDFrame, sizeof(struct dframe));

                           dataframes_buffered++;
                           add_to_buffer(&BufferedDFrames, &CurrentDFrame, dataframes_buffered);
#ifdef XAFE
                           error_correct_and_write(&LastGoodDFrame, 1, &LastGoodDFrame, 1, BufferedDFrames, dataframes_buffered);
#else
                           error_correct_and_write(LastGoodDFrame.b, LastGoodDFrame.b, BufferedDFrames, dataframes_buffered);
#endif
                           free(BufferedDFrames); BufferedDFrames = NULL;

                           dataframes_buffered = 0;
                        }                
                     }
                     done = TRUE;
                     i++;
                     break;




                  case 2:  /*general case here, two frames of data, bad, or good.*/
                  case 6:  
                     dataframes += 2;
                     buffering_data_mode = process_two_frames(&BufferedDFrames, &CurrentDFrame, &LastGoodDFrame, &PreviousDFrame,  
                                                              &dataframes_buffered, endFrames, buffering_data_mode);

                     i += 2;
                     break;

                  default: 
                     fprintf( stderr, "ERROR: unexpected return value for get_two_data_frames().\n");
                     exit(1);

                  }

               }

               frame++;
               i=0;
          
            }

         }/*end of start_of_data if*/

      }/*end of if(get_next_frame(&BufferedMultiFrames, frames_buffered)==EOF) statement*/

   }/*end of while(!done) loop*/

   /**********************************************/
   /* fclose was missing!  Guenter 23.11.00      */
   /**********************************************/
   fclose(BitStream_File);
   /**********************************************/
   /* free of memory included! Guenter 23.11.00  */
   /**********************************************/
   if (BufferedMultiFrames != NULL)
      free(BufferedMultiFrames);
   if (BufferedGolays != NULL)
      free(BufferedGolays);


   /*Put HTK header back out*/
   header.nSamples = dataframes;

   rewind(HTK_Quantised_Format_File);
   write_htk_header(&header);
   fflush(HTK_Quantised_Format_File);

   printf("%d samples writen.\n", dataframes);
   fflush(HTK_Quantised_Format_File);
   printf("\nDone !\n");
   fclose(HTK_Quantised_Format_File);
#ifdef XAFE
   fclose(Quantised_Pitch_Period_File);
   fclose(Classification_File);
   fclose(Error_Status_File);
#endif

   if(VAD_File != NULL)
   {
      fflush(VAD_File);
      fclose(VAD_File);
   }


   if(BufferedDFrames!=NULL) /*this is just a check that the code is ok*/
   {  
      printf("\nWarning: Data still left in Buffer dframes %d  frame %d  buffer %d\n", dataframes, frame, dataframes_buffered);

      HTK_Quantised_Format_File = fopen("/org/erl/users/alune/DSR/Found_Unemptied buffer", "a");
      fprintf(HTK_Quantised_Format_File, "mare\n"); /*need to do this cause I can't check all the screen fulls of output.*/
      fclose(HTK_Quantised_Format_File);
   }


   return 0;
}









void init_thresholds(void)
{
   SD.c1 =    5.89043;
   SD.c2 =    4.41609;
   SD.c3 =    3.69742;
   SD.c4 =    3.32596;
   SD.c5 =    3.00301;
   SD.c6 =    2.78532;
   SD.c7 =    2.59213;
   SD.c8 =    2.43293;
   SD.c9 =    2.29550;
   SD.c10 =   2.13293;
   SD.c11 =   2.03381;
   SD.c12 =   1.87658;
   SD.c0 =    17.34211;
   SD.logE =  0.81309;




   DETECTION_THRESHOLD.c1   = DETECTION_THRES_COEF * SD.c1;
   DETECTION_THRESHOLD.c2   = DETECTION_THRES_COEF * SD.c2;
   DETECTION_THRESHOLD.c3   = DETECTION_THRES_COEF * SD.c3;
   DETECTION_THRESHOLD.c4   = DETECTION_THRES_COEF * SD.c4;
   DETECTION_THRESHOLD.c5   = DETECTION_THRES_COEF * SD.c5;
   DETECTION_THRESHOLD.c6   = DETECTION_THRES_COEF * SD.c6;
   DETECTION_THRESHOLD.c7   = DETECTION_THRES_COEF * SD.c7;
   DETECTION_THRESHOLD.c8   = DETECTION_THRES_COEF * SD.c8;
   DETECTION_THRESHOLD.c9   = DETECTION_THRES_COEF * SD.c9;
   DETECTION_THRESHOLD.c10  = DETECTION_THRES_COEF * SD.c10;
   DETECTION_THRESHOLD.c11  = DETECTION_THRES_COEF * SD.c11;
   DETECTION_THRESHOLD.c12  = DETECTION_THRES_COEF * SD.c12;
   DETECTION_THRESHOLD.c0   = DETECTION_THRES_COEF * SD.c0;
   DETECTION_THRESHOLD.logE = DETECTION_THRES_COEF * SD.logE;


   return;
}














int process_two_frames(dframe **BufferedDFrames, dframe *CurrentDFrame, dframe *LastGoodDFrame, dframe *PreviousDFrame,  
                       int *dataframes_buffered, int endFrames, int buffering_data_mode)
{

   if(buffering_data_mode)
   {
      if(endFrames & 0x4) /*crc is in error*/
      {
         (*dataframes_buffered) += 2;
         add_to_buffer(BufferedDFrames, CurrentDFrame, (*dataframes_buffered));
      }
      else
      {
         if(threshold_error(*CurrentDFrame))
         {
            (*dataframes_buffered) += 2;
            add_to_buffer(BufferedDFrames, CurrentDFrame, (*dataframes_buffered));
         }
         else
         {

            if((*dataframes_buffered)>0)
            {
#ifdef XAFE
               error_correct_and_write(LastGoodDFrame, 1, CurrentDFrame, 0, *BufferedDFrames, *dataframes_buffered);
#else
               error_correct_and_write(LastGoodDFrame->b, CurrentDFrame->a, *BufferedDFrames, *dataframes_buffered);
#endif
               /*i.e. frame repeat backwards*/
               free(*BufferedDFrames); (*BufferedDFrames) = NULL;
               (*dataframes_buffered) = 0;
            }
            memcpy(PreviousDFrame, CurrentDFrame, sizeof(struct dframe));
            memcpy(LastGoodDFrame, CurrentDFrame, sizeof(struct dframe));

            buffering_data_mode = FALSE;
         }
      }
   }
   else /*Not buffering data mode*/
   {
      if(endFrames & 0x4) /*crc is in error*/
      {
         dataframes -= 4;

         if(threshold_error(*PreviousDFrame))
         {
            dataframes += 4;

            (*dataframes_buffered) += 2;
            add_to_buffer(BufferedDFrames, PreviousDFrame, (*dataframes_buffered));

            (*dataframes_buffered) += 2;
            add_to_buffer(BufferedDFrames, CurrentDFrame, (*dataframes_buffered));

            buffering_data_mode = TRUE;
         }
         else/*threshold of previous is ok*/
         {
            dataframes += 4;
#ifdef XAFE
            write_two_data_frames(*PreviousDFrame, TWO_FRAMES, '0');
#else
            write_two_data_frames(*PreviousDFrame, TWO_FRAMES);
#endif
            memcpy(LastGoodDFrame, PreviousDFrame, sizeof(struct dframe));

            (*dataframes_buffered) += 2;
            add_to_buffer(BufferedDFrames, CurrentDFrame, (*dataframes_buffered));

            buffering_data_mode = TRUE;
         }
      }
      else /*crc was ok*/
      {
#ifdef XAFE
         write_two_data_frames(*PreviousDFrame, TWO_FRAMES, '0');
#else
         write_two_data_frames(*PreviousDFrame, TWO_FRAMES);
#endif
         memcpy(LastGoodDFrame, PreviousDFrame, sizeof(struct dframe));

         memcpy(PreviousDFrame, CurrentDFrame, sizeof(struct dframe));
      }
   }



   return buffering_data_mode;

}










int find_good_golay(mframe **BufferedMultiFrames, golaydata **BufferedGolays, short int *GoodGolay, int *frames_buffered) 
{
   /*here we search for two golays that match each other, and have ok crcs.*/

   int i, j,
      end_of_file= FALSE,
      golays_buffered = 1; /*Must equal one since one was got before.*/ 
 
 
 
   /*note: golay_errors() also performs bit correction on the golay.*/

   /*golays_buffered - 1 will always equal zero here. Included for niceness.*/

   if(golay_errors(&((*BufferedGolays)[golays_buffered - 1].data), &((*BufferedGolays)[golays_buffered - 1].crc)))
   {
      fprintf(stderr,"Warning: Error found in multiframe header: %3d\n", ((dataframes/23)+(*frames_buffered)-1));
      golays_buffered--; /*i.e. frame header just read in is bad, therefore exclude it from comparison*/
   }

   while(((*GoodGolay) == -1) && (!end_of_file))
   {
      (*frames_buffered)++;
      if(get_next_frame(BufferedMultiFrames, (*frames_buffered))==EOF) /*get another frame to compare with previous one(s)*/
      {
         (*frames_buffered)--;
         end_of_file=TRUE;
      }
      else
      {

         golays_buffered++;
         get_next_golay(BufferedGolays, (*BufferedMultiFrames)[(*frames_buffered) - 1], golays_buffered);

         if(golay_errors((&(*BufferedGolays)[golays_buffered - 1].data), &((*BufferedGolays)[golays_buffered - 1].crc)))
         {
            printf("Warning: Error found in multiframe header: %3d\n", ((dataframes/23)+(*frames_buffered)-1));
            golays_buffered--; /*i.e. frame just read in is bad will exclude it from comparison*/
         }


         for(i=0; i<(golays_buffered - 1); i++)
         {
            /*BufferedGolays[golays_buffered - 1].data since already compared other ones*/

            /*i.e. first have [golay a, golay b] -> test whether a==b ? if so, then we have found a good golay. */
            /*if a!=b then buffer c & have [a,b,c]  don't need to do a==b? since done time b4, only a==c? and b==c?*/

            /*Therefore always compare most recent one (c) with all the others (a,b). and exculde itself from the comparison (i.e. c==c).*/

            /*The 0x078 mask fixes the bits concerning the header case MULTIFRAMER COUNTER to 1's */
            /*Otherwise the comparation between sucessive frames is always different */
            if((0x078 | (*BufferedGolays)[i].data) == (0x0078 | (*BufferedGolays)[golays_buffered - 1].data)) /*i.e. found two that match.*/
            {
               *GoodGolay = (*BufferedGolays)[i].data;
               i = golays_buffered; /*To force loop exit*/


               /*Extract common information bits from header.*/

               sampRate = 0x003 & (*GoodGolay); /*selects which quantisers to use.*/
               /**********************************************************/
               /* The following loop was introduced here to set all buffered  */
               /* golays to GoodGolay. (shifted from main routine)            */
               /*   Guenter 23.11.00                                          */
               /***************************************************************/
               for(j=0; j<(golays_buffered - 1); j++)
                  (*BufferedGolays)[j].data = *GoodGolay;      
          
            }
         }
      }
   }


   return !end_of_file; /*i.e. if(eof==true) then we havn't found a good golay.*/

}










void add_to_buffer(dframe **BufferedData, dframe *ToBuffer, int buffered_count)
{

   /*buffered_count+1 since may have half a frame at the end. and will want to alloc space for that too.*/

   (*BufferedData) = (struct dframe *) realloc((struct dframe *)(*BufferedData), sizeof(struct dframe) * ((buffered_count+1)/2));

   if((*BufferedData)==NULL)
   {
      printf("Error: Out of memory allocating space for Data !\n");
      exit(1);
   }

   memcpy(&((*BufferedData)[((buffered_count+1)/2) - 1]), ToBuffer, sizeof(struct dframe));


   return;

}






int threshold_error(dframe DataFrame)
{
   HTK_DataFrame difference, first, second;
   int threshold_errors = 0, result = 0;

   dequantise_dataframe(&first, DataFrame.a);
   dequantise_dataframe(&second, DataFrame.b);


   difference.c1   = fabs(second.c1   - first.c1);
   difference.c2   = fabs(second.c2   - first.c2);
   difference.c3   = fabs(second.c3   - first.c3);
   difference.c4   = fabs(second.c4   - first.c4);
   difference.c5   = fabs(second.c5   - first.c5);
   difference.c6   = fabs(second.c6   - first.c6);
   difference.c7   = fabs(second.c7   - first.c7);
   difference.c8   = fabs(second.c8   - first.c8);
   difference.c9   = fabs(second.c9   - first.c9);
   difference.c10  = fabs(second.c10  - first.c10);
   difference.c11  = fabs(second.c11  - first.c11);
   difference.c12  = fabs(second.c12  - first.c12);
   difference.c0   = fabs(second.c0   - first.c0);
   difference.logE = fabs(second.logE - first.logE);



   threshold_errors = 0;

   if((difference.c1 > DETECTION_THRESHOLD.c1)   || (difference.c2 > DETECTION_THRESHOLD.c2))     threshold_errors++;
   if((difference.c3 > DETECTION_THRESHOLD.c3)   || (difference.c4 > DETECTION_THRESHOLD.c4))     threshold_errors++;
   if((difference.c5 > DETECTION_THRESHOLD.c5)   || (difference.c6 > DETECTION_THRESHOLD.c6))     threshold_errors++;
   if((difference.c7 > DETECTION_THRESHOLD.c7)   || (difference.c8 > DETECTION_THRESHOLD.c8))     threshold_errors++;
   if((difference.c9 > DETECTION_THRESHOLD.c9)   || (difference.c10 > DETECTION_THRESHOLD.c10))   threshold_errors++;
   if((difference.c11 > DETECTION_THRESHOLD.c11) || (difference.c12 > DETECTION_THRESHOLD.c12))   threshold_errors++;
   if((difference.c0 > DETECTION_THRESHOLD.c0)   || (difference.logE > DETECTION_THRESHOLD.logE)) threshold_errors++;

   if(threshold_errors >= ERRORS_THRESHOLD)
   {

      fprintf(stderr, "Warning: %d Threshold error(s) in multiframe: %3d dataframe %2d or %2d\n", threshold_errors,
             (dataframes/24), (dataframes%24), ((dataframes%24)+1));

      result = -1;

   }


   return result;
}










void dequantise_dataframe(HTK_DataFrame *result, unsigned char *given_index)
{
   result->VAD = given_index[5] & 0x01; /*unpack VAD flag*/
   given_index[5]= given_index[5]>>1; /*move c11 and c12 from transmission postion to evaluation postion*/


   /*Select which quantiser tables to use for dequantisation.*/
   switch(sampRate)
   {
   case eight_kHz:
   case eleven_kHz:    
      result->c1   = quantiser8kHz_0_1[given_index[0]][0];
      result->c2   = quantiser8kHz_0_1[given_index[0]][1];
      result->c3   = quantiser8kHz_2_3[given_index[1]][0];
      result->c4   = quantiser8kHz_2_3[given_index[1]][1];
      result->c5   = quantiser8kHz_4_5[given_index[2]][0];
      result->c6   = quantiser8kHz_4_5[given_index[2]][1];
      result->c7   = quantiser8kHz_6_7[given_index[3]][0];
      result->c8   = quantiser8kHz_6_7[given_index[3]][1];
      result->c9   = quantiser8kHz_8_9[given_index[4]][0];
      result->c10  = quantiser8kHz_8_9[given_index[4]][1];
      result->c11  = quantiser8kHz_10_11[given_index[5]][0];
      result->c12  = quantiser8kHz_10_11[given_index[5]][1];
      result->c0   = quantiser8kHz_12_13[given_index[6]][0];
      result->logE = quantiser8kHz_12_13[given_index[6]][1];
      break;


   case sixteen_kHz:
      result->c1   = quantiser16kHz_0_1[given_index[0]][0];
      result->c2   = quantiser16kHz_0_1[given_index[0]][1];
      result->c3   = quantiser16kHz_2_3[given_index[1]][0];
      result->c4   = quantiser16kHz_2_3[given_index[1]][1];
      result->c5   = quantiser16kHz_4_5[given_index[2]][0];
      result->c6   = quantiser16kHz_4_5[given_index[2]][1];
      result->c7   = quantiser16kHz_6_7[given_index[3]][0];
      result->c8   = quantiser16kHz_6_7[given_index[3]][1];
      result->c9   = quantiser16kHz_8_9[given_index[4]][0];
      result->c10  = quantiser16kHz_8_9[given_index[4]][1];
      result->c11  = quantiser16kHz_10_11[given_index[5]][0];
      result->c12  = quantiser16kHz_10_11[given_index[5]][1];
      result->c0   = quantiser16kHz_12_13[given_index[6]][0];
      result->logE = quantiser16kHz_12_13[given_index[6]][1];
      break;

   default: printf("ERROR: Bad sampling rate.\n");
      exit(1);
   }

  
   result -> c1   /= C_FACTOR;
   result -> c2   /= C_FACTOR;
   result -> c3   /= C_FACTOR;
   result -> c4   /= C_FACTOR;
   result -> c5   /= C_FACTOR;
   result -> c6   /= C_FACTOR;
   result -> c7   /= C_FACTOR;
   result -> c8   /= C_FACTOR;
   result -> c9   /= C_FACTOR;
   result -> c10  /= C_FACTOR;
   result -> c11  /= C_FACTOR;
   result -> c12  /= C_FACTOR;
   result -> c0   /= C0_FACTOR;
   result -> logE /= LOGE_FACTOR;

   return;

}




#ifdef XAFE
void error_correct_and_write(dframe *Left, int Left_a_or_b, dframe *Right, int Right_a_or_b, dframe *BufferedData, int count)
#else
void error_correct_and_write(unsigned char Left[7], unsigned char Right[7], dframe *BufferedData, int count)
#endif
{
   int i;
   HTK_DataFrame L, R;
   
#ifdef XAFE
  float quant_pitch_left, quant_pitch_right;
  unsigned char class_left, class_right;
#endif


   if((Left==NULL) && (Right==NULL)) /*No good left or right data given, thus have to just put out bad data.*/
   {
      printf("Writing out bad data !\n");
  
      for(i=0; i<count; i+=2)
      {
#ifdef XAFE
      write_two_data_frames(BufferedData[i/2], TWO_FRAMES, '3');
#else
      write_two_data_frames(BufferedData[i/2], TWO_FRAMES);
#endif
      }
      
      if((i%2)!=0) /*in case of odd one at the end.*/
      {
#ifdef XAFE
         write_two_data_frames(BufferedData[(count+1)/2], ONE_FRAME, '3');
#else
         write_two_data_frames(BufferedData[(count+1)/2], ONE_FRAME);
#endif
      }
   }
   else
   {

      /*Dequantise the given good indicies*/
#ifdef XAFE
     if (Left_a_or_b == 0)
     {
       dequantise_dataframe(&L, Left->a);
     }
     else
     {
       dequantise_dataframe(&L, Left->b);
     }
     if (Right_a_or_b == 0)
     {
       dequantise_dataframe(&R, Right->a);
     }
     else
     {
       dequantise_dataframe(&R, Right->b);
     }

     inv_quant_pitch_abs(Left->pitch1_index,&quant_pitch_left);
     get_class(Left->class1_index,quant_pitch_left,&class_left);

     inv_quant_pitch_abs(Right->pitch1_index,&quant_pitch_right);
     get_class(Right->class1_index,quant_pitch_right,&class_right);

     for(i=1; i<=((count+1)/2); i++)
     {
       write_htk_coefs(&L);
       fwrite(&quant_pitch_left,4,1,Quantised_Pitch_Period_File);
       fprintf(Classification_File,"%c",class_left);
       fprintf(Error_Status_File,"%c",'1');
     }

     for(i=(((count+1)/2)+1); i<=count; i++)
     {
       write_htk_coefs(&R);
       fwrite(&quant_pitch_right,4,1,Quantised_Pitch_Period_File);
       fprintf(Classification_File,"%c",class_right);
       fprintf(Error_Status_File,"%c",'2');
     }
#else
      dequantise_dataframe(&L, Left);
      dequantise_dataframe(&R, Right);

      for(i=1; i<=((count+1)/2); i++)
         write_htk_coefs(&L);

      for(i=(((count+1)/2)+1); i<=count; i++)
         write_htk_coefs(&R);
#endif
   }
   
   return;
}











void get_next_golay(golaydata **BufferedGolays, mframe BufferedMultiFrames, int golays_buffered)
{
   static int last_size=0;
   unsigned char header[4];
   golaydata temp;


   memcpy(header, BufferedMultiFrames.golay, 4 * sizeof(unsigned char));

   temp.data = 0x00ff & header[0];
   temp.data |= 0xff00 & (header[1]<<8);

   temp.crc = 0x00ff & (header[2]); 
   temp.crc |= 0xff00 & (header[3]<<8);
  
   if(last_size!=golays_buffered)
   {
      (*BufferedGolays) = (struct golaydata *) realloc((struct golaydata *)(*BufferedGolays), sizeof(struct golaydata) *
                                                       golays_buffered);

      if((*BufferedGolays)==NULL)
      {
         fprintf( stderr, "Error: Out of memory allocating space for golay !\n");
         exit(1);
      }
      last_size = golays_buffered;
   }


   (*BufferedGolays)[golays_buffered - 1].crc = temp.crc;

   (*BufferedGolays)[golays_buffered - 1].data = temp.data;

   return;
}




int get_next_frame(mframe **BufferedMultiFrames, int frames_buffered)
{
   static int last_size=0;
   int i;
   mframe temp;
   unsigned char sync_byte;

   /*read start sync or inv end sync.*/

   /*get sync, lsb first*/

   fread(&sync_byte, sizeof(unsigned char), 1, BitStream_File);

   temp.sync[1] = sync_byte;

   fread(&sync_byte, sizeof(unsigned char), 1, BitStream_File);

   temp.sync[0] = sync_byte;
  
   /*In a realtime system then would perform sync searching here.*/
   /*Not implemented in this release.*/  


   fread(temp.golay, sizeof(unsigned char), 4, BitStream_File);
#ifdef XAFE
   fread(temp.data, sizeof(unsigned char), 162, BitStream_File);
#else
   fread(temp.data, sizeof(unsigned char), 138, BitStream_File);
#endif

   if(feof(BitStream_File)!=0)
      return EOF;


   if(frames_buffered!=last_size)
   {
      (*BufferedMultiFrames) = (struct mframe *) realloc((struct mframe *)(*BufferedMultiFrames), sizeof(struct mframe) * frames_buffered);


      if((*BufferedMultiFrames)==NULL)
      {
         fprintf( stderr, "Error: Out of memory allocating space for multiframes !\n" );
         exit(1);
      }

      last_size = frames_buffered;
   }


   for(i=0; i<2; i++)
      (*BufferedMultiFrames)[frames_buffered - 1].sync[i] = temp.sync[i];

   for(i=0; i<4; i++)
      (*BufferedMultiFrames)[frames_buffered - 1].golay[i] = temp.golay[i];

#ifdef XAFE
   for(i=0; i<162; i++)
      (*BufferedMultiFrames)[frames_buffered - 1].data[i] = temp.data[i];
#else
   for(i=0; i<138; i++)
      (*BufferedMultiFrames)[frames_buffered - 1].data[i] = temp.data[i];
#endif

   return 0;

}




int get_two_data_frames(dframe *DataFrame, mframe BufferedMultiFrame, int frame)
{
   static int get_two_pairs=TRUE;

#ifdef XAFE
  static unsigned char four_data_frames[27];
#else
  static unsigned char four_data_frames[23];
#endif
   unsigned char two_data_frames[11];

   int index, frames_non_zero=2, zeros_found;

#ifdef XAFE
  unsigned char pc_data[14];
  unsigned char pc_crc[2];
#endif

/*need to get two pairs here, due to the way they fit.*/


   if(get_two_pairs==TRUE)
   {
#ifdef XAFE
      memcpy(four_data_frames, &BufferedMultiFrame.data[frame/4 * 27], sizeof(unsigned char) * 27);
#else
      memcpy(four_data_frames, &BufferedMultiFrame.data[frame/4 * 23], sizeof(unsigned char) * 23);
#endif


      DataFrame->a[0] = (0x3f & four_data_frames[0]);
      DataFrame->a[1] = (0x03 & (four_data_frames[0]>>6));
      DataFrame->a[1] |= (0x3c & (four_data_frames[1]<<2));
      DataFrame->a[2] = (0x0f & (four_data_frames[1]>>4));
      DataFrame->a[2] |= (0x30 & (four_data_frames[2]<<4));
      DataFrame->a[3] = (0x3f & (four_data_frames[2]>>2));
      DataFrame->a[4] = (0x3f & four_data_frames[3]);
      DataFrame->a[5] = (0x03 & (four_data_frames[3]>>6));
      DataFrame->a[5] |= (0x3c & (four_data_frames[4]<<2));
      DataFrame->a[6] = (0x0f & (four_data_frames[4]>>4));
      DataFrame->a[6] |= (0xf0 & (four_data_frames[5]<<4));

      DataFrame->b[0] = (0x0f & (four_data_frames[5])>>4);
      DataFrame->b[0] |= (0x30 & (four_data_frames[6])<<4);
      DataFrame->b[1] = (0x3f & (four_data_frames[6]>>2));
      DataFrame->b[2] = (0x3f & four_data_frames[7]);
      DataFrame->b[3] = (0x03 & (four_data_frames[7]>>6));
      DataFrame->b[3] |= (0x3c & (four_data_frames[8]<<2));
      DataFrame->b[4] = (0x0f & (four_data_frames[8]>>4));
      DataFrame->b[4] |= (0x30 & (four_data_frames[9]<<4));
      DataFrame->b[5] = (0x3f & (four_data_frames[9]>>2));
      DataFrame->b[6] = (0xff & four_data_frames[10]);

      DataFrame->crc = (0x0f & four_data_frames[11]);




#ifdef XAFE
    /*
     * Get pitch and class information
     */

    DataFrame->pitch1_index = (0x0f & (four_data_frames[11] >> 4));
    DataFrame->pitch1_index |= (0x70 & (four_data_frames[12] << 4));
    DataFrame->pitch2_index = (0x1f & (four_data_frames[12] >> 3));
    DataFrame->class1_index = (0x01 & four_data_frames[13]);
    DataFrame->class2_index = (0x01 & (four_data_frames[13] >> 1));

    pc_crc[0] = (0x01 & (four_data_frames[13] >> 2));
    pc_crc[1] = (0x01 & (four_data_frames[13] >> 3));

    for (index = 0; index < 7; index++)
    {
      pc_data[index] = (0x01 & (DataFrame->pitch1_index >> index));
    }
    for (index = 7; index < 12; index++)
    {
      pc_data[index] = (0x01 & (DataFrame->pitch2_index >> (index-7)));
    }
    pc_data[12] = (0x01 & DataFrame->class1_index);
    pc_data[13] = (0x01 & DataFrame->class2_index);
#endif

      for(index=0; index<11; index++) 
         two_data_frames[index] = four_data_frames[index];

      get_two_pairs=FALSE;

   }
   else
   {
#ifdef XAFE
      DataFrame->a[0] = (0x0f & (four_data_frames[13])>>4);
      DataFrame->a[0] |= (0x30 & (four_data_frames[14])<<4);
      DataFrame->a[1] = (0x3f & (four_data_frames[14]>>2));
      DataFrame->a[2] = (0x3f & four_data_frames[15]);
      DataFrame->a[3] = (0x03 & (four_data_frames[15]>>6));
      DataFrame->a[3] |= (0x3c & (four_data_frames[16]<<2));
      DataFrame->a[4] = (0x0f & (four_data_frames[16]>>4));
      DataFrame->a[4] |= (0x30 & (four_data_frames[17]<<4));
      DataFrame->a[5] = (0x3f & (four_data_frames[17]>>2));
      DataFrame->a[6] = (0xff & four_data_frames[18]);
      
      DataFrame->b[0] = (0x3f & four_data_frames[19]);
      DataFrame->b[1] = (0x03 & (four_data_frames[19]>>6));
      DataFrame->b[1] |= (0x3c & (four_data_frames[20]<<2));
      DataFrame->b[2] = (0x0f & (four_data_frames[20]>>4));
      DataFrame->b[2] |= (0x30 & (four_data_frames[21]<<4));
      DataFrame->b[3] = (0x3f & (four_data_frames[21]>>2));
      DataFrame->b[4] = (0x3f & four_data_frames[22]);
      DataFrame->b[5] = (0x03 & (four_data_frames[22]>>6));
      DataFrame->b[5] |= (0x3c & (four_data_frames[23]<<2));
      DataFrame->b[6] = (0x0f & (four_data_frames[23]>>4));
      DataFrame->b[6] |= (0xf0 & (four_data_frames[24]<<4));
            
      DataFrame->crc = (0x0f & (four_data_frames[24]>>4));
#else
      DataFrame->a[0] = (0x0f & (four_data_frames[11])>>4);
      DataFrame->a[0] |= (0x30 & (four_data_frames[12])<<4);
      DataFrame->a[1] = (0x3f & (four_data_frames[12]>>2));
      DataFrame->a[2] = (0x3f & four_data_frames[13]);
      DataFrame->a[3] = (0x03 & (four_data_frames[13]>>6));
      DataFrame->a[3] |= (0x3c & (four_data_frames[14]<<2));
      DataFrame->a[4] = (0x0f & (four_data_frames[14]>>4));
      DataFrame->a[4] |= (0x30 & (four_data_frames[15]<<4));
      DataFrame->a[5] = (0x3f & (four_data_frames[15]>>2));
      DataFrame->a[6] = (0xff & four_data_frames[16]);

      DataFrame->b[0] = (0x3f & four_data_frames[17]);
      DataFrame->b[1] = (0x03 & (four_data_frames[17]>>6));
      DataFrame->b[1] |= (0x3c & (four_data_frames[18]<<2));
      DataFrame->b[2] = (0x0f & (four_data_frames[18]>>4));
      DataFrame->b[2] |= (0x30 & (four_data_frames[19]<<4));
      DataFrame->b[3] = (0x3f & (four_data_frames[19]>>2));
      DataFrame->b[4] = (0x3f & four_data_frames[20]);
      DataFrame->b[5] = (0x03 & (four_data_frames[20]>>6));
      DataFrame->b[5] |= (0x3c & (four_data_frames[21]<<2));
      DataFrame->b[6] = (0x0f & (four_data_frames[21]>>4));
      DataFrame->b[6] |= (0xf0 & (four_data_frames[22]<<4));


      DataFrame->crc = (0x0f & (four_data_frames[22]>>4));
#endif


#ifdef XAFE
    /*
     * Get pitch and class information
     */

    DataFrame->pitch1_index = (0x7f & four_data_frames[25]);
    DataFrame->pitch2_index = (0x01 & (four_data_frames[25] >> 7));
    DataFrame->pitch2_index |= (0x1e & (four_data_frames[26] << 1));
    DataFrame->class1_index = (0x01 & (four_data_frames[26] >> 4));
    DataFrame->class2_index = (0x01 & (four_data_frames[26] >> 5));

    pc_crc[0] = (0x01 & (four_data_frames[26] >> 6));
    pc_crc[1] = (0x01 & (four_data_frames[26] >> 7));

    for (index = 0; index < 7; index++)
    {
      pc_data[index] = (0x01 & (DataFrame->pitch1_index >> index));
    }
    for (index = 7; index < 12; index++)
    {
      pc_data[index] = (0x01 & (DataFrame->pitch2_index >> (index-7)));
    }
    pc_data[12] = (0x01 & DataFrame->class1_index);
    pc_data[13] = (0x01 & DataFrame->class2_index);

#endif

      two_data_frames[0] = (0x3f & DataFrame->a[0]);
      two_data_frames[0] |= (0xc0 & (DataFrame->a[1] << 6));
      two_data_frames[1] = (0x0f & (DataFrame->a[1] >> 2));
      two_data_frames[1] |= (0xf0 & (DataFrame->a[2] << 4));
      two_data_frames[2] = (0x03 & (DataFrame->a[2] >> 4));
      two_data_frames[2] |= (0xfc & (DataFrame->a[3] << 2));
      two_data_frames[3] = (0x3f & DataFrame->a[4]);
      two_data_frames[3] |= (0xc0 & (DataFrame->a[5] << 6));
      two_data_frames[4] =  (0x0f & (DataFrame->a[5] >> 2));
      two_data_frames[4] |= (0xf0 & (DataFrame->a[6] << 4));
      two_data_frames[5] = (0x0f & (DataFrame->a[6] >> 4));

      two_data_frames[5] |= (0xf0 & (DataFrame->b[0] << 4));
      two_data_frames[6] = (0x03 & (DataFrame->b[0] >> 4));
      two_data_frames[6] |= (0xfc & (DataFrame->b[1] << 2));
      two_data_frames[7] = (0x3f & DataFrame->b[2]);
      two_data_frames[7] |= (0xc0 & (DataFrame->b[3] << 6));
      two_data_frames[8] = (0x0f & (DataFrame->b[3] >> 2));
      two_data_frames[8] |= (0xf0 & (DataFrame->b[4] << 4));
      two_data_frames[9] = (0x03 & (DataFrame->b[4] >> 4));
      two_data_frames[9] |= (0xfc & (DataFrame->b[5] << 2));
      two_data_frames[10] = (0xff & DataFrame->b[6]);


  
      get_two_pairs=TRUE;
   }




#ifdef XAFE
  if((mfcc_crc_errors(two_data_frames, DataFrame->crc)==TRUE) ||  (pc_crc_errors(pc_data,pc_crc) == TRUE))
#else
   if(mfcc_crc_errors(two_data_frames, DataFrame->crc)==TRUE)
#endif
   {
      fprintf( stderr, "Warning: A CRC error detected in multiframe: %3d dataframe %2d or %2d\n", (dataframes/24), (dataframes%24), ((dataframes%24)+1));

      zeros_found = 0;

      for(index=0; index<7; index++)
         if(DataFrame->a[index]==0) zeros_found++;

      if(zeros_found>=ZEROS_THRESHOLD)
         frames_non_zero=0;  /*i.e. these are all zero padding frames*/
      else
      {
         zeros_found = 0;

         for(index=0; index<7; index++)
            if(DataFrame->b[index]==0) zeros_found++;

         if(zeros_found>=ZEROS_THRESHOLD)
            frames_non_zero=1;
         else
            frames_non_zero=2;
      }



      frames_non_zero |= 0x4;
   }
   else
   {

      zeros_found = 0;

      for(index=0; index<7; index++)
         if(DataFrame->a[index]==0) zeros_found++;

      if(zeros_found==7)  /*CRC is ok therefore all indicies have to be zero for padding*/
         frames_non_zero=0;
      else
      {
         zeros_found = 0;

         for(index=0; index<7; index++)
            if(DataFrame->b[index]==0) zeros_found++;

         if(zeros_found==7)
            frames_non_zero=1;
         else
            frames_non_zero=2;
      }
  
   }



   return frames_non_zero;

}






#ifdef XAFE
void write_two_data_frames(dframe given, int numFrames, unsigned char error_status)
#else
void write_two_data_frames(dframe given, int numFrames)
#endif
{
   HTK_DataFrame quantised_frame;
   
#ifdef XAFE
   float quant_pitch1, quant_pitch2;
   unsigned char class1, class2;
   static float pfQPHistory[3] = {0.0, 0.0, 0.0};
   static int iReliableFlag = FALSE;
#endif
   
   dequantise_dataframe(&quantised_frame, given.a);
   write_htk_coefs(&quantised_frame);
   
#ifdef XAFE
   inv_quant_pitch_abs(given.pitch1_index,&quant_pitch1);
   get_class(given.class1_index,quant_pitch1,&class1);
   pfQPHistory[2] = pfQPHistory[1];
   pfQPHistory[1] = pfQPHistory[0];
   pfQPHistory[0] = quant_pitch1;
   
   fwrite(&quant_pitch1,4,1,Quantised_Pitch_Period_File);
   fprintf(Classification_File,"%c",class1);
   fprintf(Error_Status_File,"%c",error_status);
#endif
   
   if(numFrames==TWO_FRAMES)
   {
      dequantise_dataframe(&quantised_frame, given.b);
      write_htk_coefs(&quantised_frame);
      
#ifdef XAFE
      inv_quant_pitch_diff(given.pitch2_index,pfQPHistory,
                           &iReliableFlag,&quant_pitch2);
      get_class(given.class2_index,quant_pitch2,&class2);
      pfQPHistory[2] = pfQPHistory[1];
      pfQPHistory[1] = pfQPHistory[0];
      pfQPHistory[0] = quant_pitch2;
      
      fwrite(&quant_pitch2,4,1,Quantised_Pitch_Period_File);
      fprintf(Classification_File,"%c",class2);
      fprintf(Error_Status_File,"%c",error_status);   
#endif
   }
     
   return;
}












void process_params(int argc, char *argv[])
{
   int errors=0;
   int argCounter=0;
   int htk_file_arg;
   char *ptP;
   char VADFilename[199];

#ifdef XAFE
   if(argc!=7)
   {
      print_usage();
      exit(1);
   }
#else
   if(argc!=4)
   {
      print_usage();
      exit(1);
   }
#endif

   if((BitStream_File = fopen(argv[++argCounter], "rb"))==NULL)
   {
      fprintf( stderr, "\nError: cannot open file: %s\n", argv[argCounter] );
      errors++;
   }


   if((HTK_Quantised_Format_File = fopen(argv[++argCounter], "wb"))==NULL)
   {
      fprintf( stderr, "\nError: cannot create file: %s\n", argv[argCounter] );
      errors++;
   }
   htk_file_arg = argCounter;

#ifdef XAFE
  if((Quantised_Pitch_Period_File = fopen(argv[++argCounter], "wb"))==NULL)
  {
    fprintf( stderr, "\nError: cannot create file: %s\n", argv[argCounter]);
    errors++;
  }


  if((Classification_File = fopen(argv[++argCounter], "w"))==NULL)
  {
    fprintf( stderr, "\nError: cannot create file: %s\n", argv[argCounter]);
    errors++;
  }


  if((Error_Status_File = fopen(argv[++argCounter], "w"))==NULL)
  {
    fprintf( stderr, "\nError: cannot create file: %s\n", argv[argCounter]);
    errors++;
  }
#endif 

   if (strcmp("-VAD", argv[++argCounter])==0)
   {
      fprintf( stderr, "Running with VAD\n" );
      strcpy(VADFilename, argv[htk_file_arg]);
      if( (ptP= strrchr(VADFilename,'.')) != NULL) 
      {
          *ptP = '\0' ;
      } 
      strcat(VADFilename,".vad_o");
     
      if((VAD_File = fopen(VADFilename, "w"))==NULL)
      {
         fprintf(stderr, "Error: Cannot open %s\n", VADFilename);
         errors++;  
      }
   }	

   else
   {
      if (strcmp("-No_VAD", argv[argCounter])!=0)
         errors++;
      fprintf( stderr, "Running without VAD\n");
      VAD_File = NULL;    
   }

   if(errors!=0)
   {
      fprintf( stderr, "%d errors in arguments found, exiting...\n", errors);
      exit(1);
   }

   return;
}


void print_usage( void )
{

#ifdef XAFE
    fprintf( stderr, 
             "Decoder for Extended Advanced DSR parameter bit stream\n"
             "A program to take a binary BitStream File as input, and produce\n"
             "(1) HTK file containing the decoded Mel-Cepstrum parameters\n"
             "(2) Pitch Period file containing the decoded pitch values\n"
             "(3) Classification file containing the class information, and\n"
             "(4) Error status file indicating the status of each frame with\n"
             "    respect to channel errors\n\n"
             "Usage: ExtDecoder bitstream_file htk_file pitch_file class_file\n"
             "       error_file -VAD/No_VAD\n\n"
             "               bitstream_file: input binary bitstream\n"
             "               htk_file:       output quantised parameter file\n"
             "               pitch_file:     output quantised pitch file\n"
             "               class_file:     output class file\n"
             "               error_file:     output error status file\n"
             "               -VAD:           write integer VAD data to VAD file (.vad)\n"
             "               -No_VAD:        Do not use VAD info\n" );
#else
     fprintf( stderr,
              "Decoder for DSR split VQ quanitser\n"
              "A program to take a binary BitStream File as an input, and produce an \n"
              "HTK file containining the decoded Mel-Cepstrum parameters\n"     
              "Usage: decoder bitstream_file htk_file\n\n"
              "               bitstream_file: input binary bitstream\n"
              "               htk_file:       output quantised parameter file.\n"
              "             -VAD :       write integer VAD data to VAD file (.vad)\n"
              "             -No_VAD :    Do not use VAD info\n");
#endif

    return;
}




void write_htk_coefs(HTK_DataFrame *given)
{

   fwrite(&given->c1, 4, 1, HTK_Quantised_Format_File);
   fwrite(&given->c2, 4, 1, HTK_Quantised_Format_File);
   fwrite(&given->c3, 4, 1, HTK_Quantised_Format_File);
   fwrite(&given->c4, 4, 1, HTK_Quantised_Format_File);
   fwrite(&given->c5, 4, 1, HTK_Quantised_Format_File);
   fwrite(&given->c6, 4, 1, HTK_Quantised_Format_File);
   fwrite(&given->c7, 4, 1, HTK_Quantised_Format_File);
   fwrite(&given->c8, 4, 1, HTK_Quantised_Format_File);
   fwrite(&given->c9, 4, 1, HTK_Quantised_Format_File);
   fwrite(&given->c10, 4, 1, HTK_Quantised_Format_File);
   fwrite(&given->c11, 4, 1, HTK_Quantised_Format_File);
   fwrite(&given->c12, 4, 1, HTK_Quantised_Format_File);
   fwrite(&given->c0, 4, 1, HTK_Quantised_Format_File);
   fwrite(&given->logE, 4, 1, HTK_Quantised_Format_File);

   if(VAD_File != NULL)
   {
      fprintf(VAD_File,"%d ",given->VAD);/* if output VAD file is required then write VAD bit */
      if(ferror(VAD_File)!=0)
      {
         printf("An error occured while writing VAD values.\n");
         exit(1);
      }
      fflush(VAD_File);
   }
  

   if(ferror(HTK_Quantised_Format_File)!=0)
   {
      printf("An error occured while writing HTK coefs.\n");
      exit(1);
   }


   fflush(HTK_Quantised_Format_File);

   return;
}




void write_htk_header(HTK_Header *header)
{

   /*This function writes the header from the struct*/

   fwrite(&header->nSamples, 4, 1, HTK_Quantised_Format_File);
   fwrite(&header->sampPeriod, 4, 1, HTK_Quantised_Format_File);
   fwrite(&header->sampSize, 2, 1, HTK_Quantised_Format_File);
   fwrite(&header->parmKind, 2, 1, HTK_Quantised_Format_File);

   if(ferror(HTK_Quantised_Format_File)!=0)
   {
      printf("An error occured while writing Header.\n");
      exit(1);
   }

   fflush(HTK_Quantised_Format_File);

   return;

}




void print_htk_coefs(HTK_DataFrame given)
{
   printf("c1   %e\n", given.c1);
   printf("c2   %e\n", given.c2);
   printf("c3   %e\n", given.c3);
   printf("c4   %e\n", given.c4);
   printf("c5   %e\n", given.c5);
   printf("c6   %e\n", given.c6);
   printf("c7   %e\n", given.c7);
   printf("c8   %e\n", given.c8);
   printf("c9   %e\n", given.c9);
   printf("c10  %e\n", given.c10);
   printf("c11  %e\n", given.c11);
   printf("c12  %e\n", given.c12);
   printf("c0   %e\n", given.c0);
   printf("logE %e\n\n", given.logE);

   return;
}

#ifdef XAFE
/*
***********************************************************************
*
*   FUNCTION NAME:
*
*      inv_quant_pitch_abs
*
*   PURPOSE:
*
*      To inverse quantize the pitch period absolutely
*
*   INPUTS:
*
*      cQPIndex - Index of the quantized pitch period
*
*   OUTPUTS:
*
*      *pfQuantPeriod - Quantized pitch period
*
*
*   RETURN VALUE:
*
*      None
*
*   IMPLEMENTATION/COMMENTS:
*
*   AUTHOR/MODIFICATIONS:
*
*   KEYWORDS:
*
*      Inverse quantization, Pitch quantization, Absolute quantization
*
*
**********************************************************************
*/



/*
**********************************************************************
*                                                                    *
*                            The function                            *
*                                                                    *
**********************************************************************
*/

void inv_quant_pitch_abs(UWord8 cQPIndex, float *pfQuantPeriod)
{


  /*
  ********************************************************************
  *                                                                  *
  *                         Static variables                         *
  *                                                                  *
  ********************************************************************
  */

static const int piQuantLevels[127] = {
1255013,
1274906,
1295114,
1315641,
1336495,
1357679,
1379198,
1401059,
1423266,
1445825,
1468742,
1492022,
1515671,
1539695,
1564100,
1588891,
1614076,
1639659,
1665648,
1692050,
1718869,
1746114,
1773790,
1801905,
1830466,
1859479,
1888953,
1918893,
1949308,
1980206,
2011592,
2043477,
2075867,
2108770,
2142194,
2176149,
2210641,
2245681,
2281276,
2317435,
2354167,
2391481,
2429387,
2467893,
2507010,
2546747,
2587114,
2628120,
2669777,
2712094,
2755081,
2798750,
2843111,
2888175,
2933954,
2980458,
3027699,
3075689,
3124440,
3173963,
3224271,
3275377,
3327293,
3380032,
3433606,
3488030,
3543316,
3599479,
3656532,
3714489,
3773365,
3833174,
3893931,
3955651,
4018349,
4082041,
4146743,
4212470,
4279239,
4347066,
4415969,
4485963,
4557067,
4629298,
4702674,
4777213,
4852933,
4929854,
5007993,
5087372,
5168008,
5249923,
5333136,
5417668,
5503539,
5590772,
5679388,
5769408,
5860855,
5953751,
6048120,
6143984,
6241369,
6340296,
6440792,
6542880,
6646587,
6751938,
6858958,
6967675,
7078115,
7190305,
7304274,
7420049,
7537659,
7657133,
7778501,
7901793,
8027039,
8154270,
8283517,
8414814,
8548191,
8683683,
8821322,
8961143,
9103179 };



  /*
  ********************************************************************
  *                                                                  *
  *                        Automatic variables                       *
  *                                                                  *
  ********************************************************************
  */


  /*
  ********************************************************************
  *                                                                  *
  *                         Executable Code                          *
  *                                                                  *
  ********************************************************************
  */

  if (cQPIndex == 0)
  {
    *pfQuantPeriod = 0.0;
  }
  else
  {
    *pfQuantPeriod = (float)piQuantLevels[cQPIndex-1]/65536.0;
  }

  return;

}




/*
***********************************************************************
*
*   FUNCTION NAME:
*
*      inv_quant_pitch_diff
*
*   PURPOSE:
*
*      To inverse quantize the pitch period differentially
*
*   INPUTS:
*
*      cQPIndex - Index of the quantized pitch period
*
*      pfQPHistory[0:2] - History of three previous quantized
*           pitch periods: pfQPHistory[0] corresponds to the
*           previous frame, pfQPHistory[1] corresponds to one
*           before the previous frame, and pfQPHistory[2]
*           corresponds to two frames before the previous frame.
*
*   INPUT/OUTPUT:
*
*      *piReliableFlag - This flag indicates if the pitch
*           period of the second frame of a frame-pair has
*           been quantized differentially using a reference
*           value from the previous frame pair.  In this
*           case, the quantized pitch period is considered
*           unreliable and the flag is set to FALSE.  If the
*           pitch period has been absolutely quantized or it
*           has been quantized differentially using the pitch
*           period value of the first frame of the same
*           frame-pair, then the flag is set to TRUE to
*           indicate that this value is reliable.  This
*           flag helps to avoid propagating errors through
*           differential quantization from frame-pair to
*           frame-pair.  On input, this flag indicates if
*           pfQPHistory[1] is reliable.  On output, this
*           flag indicates if the output *pfQuantPeriod is
*           reliable.  pfQPHistory[0] and pfQPHistory[2]
*           are always considered reliable since they are
*           absolutely quantized.
*
*   OUTPUTS:
*
*      *pfQuantPeriod - Quantized pitch period
*
*
*   RETURN VALUE:
*
*      None
*
*   IMPLEMENTATION/COMMENTS:
*
*   AUTHOR/MODIFICATIONS:
*
*   KEYWORDS:
*
*      Pitch quantization, Differential quantization
*
**********************************************************************
*/




/*
**********************************************************************
*                                                                    *
*                            The function                            *
*                                                                    *
**********************************************************************
*/

void inv_quant_pitch_diff(UWord8 cQPIndex, float *pfQPHistory,
                          int *piReliableFlag, float *pfQuantPeriod)
{


  /*
  ********************************************************************
  *                                                                  *
  *                         Static variables                         *
  *                                                                  *
  ********************************************************************
  */

  static const int piQuantLevels[31] = {
    1285948,
    1371524,
    1462794,
    1560138,
    1663959,
    1774690,
    1892790,
    2018748,
    2153089,
    2296370,
    2449185,
    2612170,
    2786001,
    2971400,
    3169136,
    3380032,
    3604961,
    3844859,
    4100721,
    4373610,
    4664658,
    4975075,
    5306150,
    5659256,
    6035860,
    6437525,
    6865921,
    7322824,
    7810133,
    8329870,
    8884194 };


  /*
   * Ratio threshold table and multi-level threshold index
   * table for Case 2
   */

  static const int piRatioThld_1[4][6] = {

    {3350, 5013, 8192, 12288, 16384, 20480},
    {3350, 5013, 6144,  8192, 10240, 12288},
    {2048, 2731, 3350,  5013,  6144,  8192},
    {1024, 1365, 2048,  2731,  3350,  5013}

  };


  static const int piMultiLevelIndex_1[4] = {

    0,
    0,
    2,
    4

  };
    

  static const int iAlpha1 = 4160;


  /*
   * Ratio threshold table and multi-level threshold index
   * table for Cases 3 and 4
   */

  static const int piRatioThld_2[4][8] = {

    {3182, 5269, 6144, 8192, 10240, 12288, 16384, 20480},
    {2731, 3182, 5269, 6144,  8192, 10240, 12288, 16384},
    {1365, 2048, 2731, 3182,  5269,  6144,  7168,  8192},
    { 819, 1024, 1365, 2048,  2731,  3182,  5269,  6144}

  };


  static const int piMultiLevelIndex_2[4] = {

    0,
    1,
    3,
    5

  };


  static const int iAlpha2 = 4183;
    

  /*
  ********************************************************************
  *                                                                  *
  *                        Automatic variables                       *
  *                                                                  *
  ********************************************************************
  */

  int iMultiLevelIndex;
  int iRowIndex;
  int iColIndex;

  float fRefPeriod;

  int iRefPeriod;
  int iLowPeriod;
  int iHighPeriod;
  int iRefHigh;
  int iRefLow;
  int iTemp;
  int iTempPeriod;
  int iTempLow;
  int iTempHigh;
  int iNumInterval;
  int iQuantPeriod;


  /*
  ********************************************************************
  *                                                                  *
  *                         Executable Code                          *
  *                                                                  *
  ********************************************************************
  */


  /*
   * First, handle the unvoiced condition
   */

  if (cQPIndex == 0)
  {
    *pfQuantPeriod = 0.0;
    *piReliableFlag = FALSE;
  }


  /*
   * Next, handle case 1 - Absolute quantization
   */

  else if ((pfQPHistory[0] <= 0.0) &&
           ((pfQPHistory[1] <= 0.0) || (*piReliableFlag == FALSE)) &&
           (pfQPHistory[2] <= 0.0))
  {
    *pfQuantPeriod = (float)piQuantLevels[cQPIndex-1]/65536.0;
    *piReliableFlag = TRUE;
  }


  /*
   * Next, handle case 2 - Differential quantization with
   * pfQPHistory[0] as reference
   */

  else if (pfQPHistory[0] > 0.0)
  {

    fRefPeriod = pfQPHistory[0];
    iRefPeriod = fRefPeriod * 65536.0;
    iTemp = (iRefPeriod << 7);
    iRefHigh = (iTemp >> 16);
    iRefLow = (iTemp >> 1) & 0x7FFF;


    /*
     * Establish the appropriate row of ratio thresholds to use
     * based on the reference pitch period
     */

    if (iRefPeriod <= 1966080)
    {
      iRowIndex = 0;
    }
    else if (iRefPeriod <= 3932160)
    {
      iRowIndex = 1;
    }
    else if (iRefPeriod <= 6225920)
    {
      iRowIndex = 2;
    }
    else
    {
      iRowIndex = 3;
    }

    iMultiLevelIndex = piMultiLevelIndex_1[iRowIndex];


    /*
     * Inverse quantize the pitch period
     */

    /*
     * First, check the thresholds below the multi-level threshold
     * to inverse quantize the pitch period
     */

    if (cQPIndex <= iMultiLevelIndex+1)
    {
      iColIndex = cQPIndex-1;
      iTemp = piRatioThld_1[iRowIndex][iColIndex];
      *pfQuantPeriod = (float)((((iRefHigh*iTemp) << 3) + 
                                ((iRefLow*iTemp) >> 13)) >> 6) / 65536.0;    
      *piReliableFlag = TRUE;
      return;
    }


    /*
     * Next, check the multi-level threshold to inverse quantize
     * the pitch period
     */

    else if ((cQPIndex > iMultiLevelIndex+1) &&
             (cQPIndex < iMultiLevelIndex + 1 + NUM_MULTI_LEVELS_1))
    {

      iColIndex = iMultiLevelIndex;
      iTemp = piRatioThld_1[iRowIndex][iColIndex];
      iLowPeriod = (((iRefHigh*iTemp) << 3) + ((iRefLow*iTemp) >> 13)) >> 6;    
      if (iLowPeriod < MIN_PERIOD)
      {
        iLowPeriod = MIN_PERIOD;
      }

      iTemp = piRatioThld_1[iRowIndex][iColIndex+1];
      iHighPeriod = (((iRefHigh*iTemp) << 3) + ((iRefLow*iTemp) >> 13)) >> 6;
      if (iHighPeriod > MAX_PERIOD)
      {
        iHighPeriod = MAX_PERIOD;
      }

      iTempPeriod = iLowPeriod;
      iNumInterval = cQPIndex - 1 - iMultiLevelIndex;
      while (iNumInterval > 0)
      {
        iTemp = (iTempPeriod << 7);
        iTempHigh = (iTemp >> 16);
        iTempLow = (iTemp >> 1) & 0x7FFF;
        iTempPeriod = (((iTempHigh*iAlpha1)<<3)+((iTempLow*iAlpha1)>>13)) >> 6;
        iNumInterval--;
      }
      *pfQuantPeriod = (float)iTempPeriod / 65536.0;
      *piReliableFlag = TRUE;

      return;

    }


    /*
     * Finally, check the thresholds above the multi-level
     * threshold to inverse quantize the pitch period
     */

    else
    {

      iColIndex = cQPIndex - NUM_MULTI_LEVELS_1;
      iTemp = piRatioThld_1[iRowIndex][iColIndex];
      iQuantPeriod = (((iRefHigh*iTemp) << 3) + ((iRefLow*iTemp) >> 13)) >> 6; 
      if (iQuantPeriod > MAX_PERIOD)
      {
        iQuantPeriod = MAX_PERIOD;
      }
      *pfQuantPeriod = (float)iQuantPeriod / 65536.0;
      *piReliableFlag = TRUE;
      return;

    }

  }


  /*
   * Next, handle cases 3 and 4 - Differential quantization
   * with pfQPHistory[1] and pfQPHistory[2] as references
   */

  else
  {

    if ((pfQPHistory[1] > 0.0) && (*piReliableFlag == TRUE))
    {
      fRefPeriod = pfQPHistory[1];
    }
    else
    {
      fRefPeriod = pfQPHistory[2];
    }
    iRefPeriod = fRefPeriod * 65536.0;
    iTemp = (iRefPeriod << 7);
    iRefHigh = (iTemp >> 16);
    iRefLow = (iTemp >> 1) & 0x7FFF;


    /*
     * Establish the appropriate row of ratio thresholds to use
     * based on the reference pitch period
     */

    if (fRefPeriod <= 1966080)
    {
      iRowIndex = 0;
    }
    else if (fRefPeriod <= 3932160)
    {
      iRowIndex = 1;
    }
    else if (fRefPeriod <= 6225920)
    {
      iRowIndex = 2;
    }
    else
    {
      iRowIndex = 3;
    }
    
    iMultiLevelIndex = piMultiLevelIndex_2[iRowIndex];


    /*
     * Inverse quantize the pitch period
     */

    /*
     * First, check the thresholds below the multi-level threshold
     * to inverse quantize the pitch period
     */

    if (cQPIndex <= iMultiLevelIndex+1)
    {
      iColIndex = cQPIndex-1;
      iTemp = piRatioThld_2[iRowIndex][iColIndex];
      *pfQuantPeriod = (float)((((iRefHigh*iTemp) << 3) +
                                ((iRefLow*iTemp) >> 13)) >> 6) / 65536.0;
      *piReliableFlag = FALSE;
      return;
    }


    /*
     * Next, check the multi-level threshold to inverse quantize
     * the pitch period
     */

    else if ((cQPIndex > iMultiLevelIndex+1) &&
             (cQPIndex < iMultiLevelIndex + 1 + NUM_MULTI_LEVELS_2))
    {

      iColIndex = iMultiLevelIndex;
      iTemp = piRatioThld_2[iRowIndex][iColIndex];
      iLowPeriod = (((iRefHigh*iTemp) << 3) + ((iRefLow*iTemp) >> 13)) >> 6;
      if (iLowPeriod < MIN_PERIOD)
      {
        iLowPeriod = MIN_PERIOD;
      }

      iTemp = piRatioThld_2[iRowIndex][iColIndex+1];
      iHighPeriod = (((iRefHigh*iTemp) << 3) + ((iRefLow*iTemp) >> 13)) >> 6;
      if (iHighPeriod > MAX_PERIOD)
      {
        iHighPeriod = MAX_PERIOD;
      }

      iTempPeriod = iLowPeriod;
      iNumInterval = cQPIndex - 1 - iMultiLevelIndex;
      while (iNumInterval > 0)
      {
        iTemp = (iTempPeriod << 7);
        iTempHigh = (iTemp >> 16);
        iTempLow = (iTemp >> 1) & 0x7FFF;
        iTempPeriod = (((iTempHigh*iAlpha2)<<3)+((iTempLow*iAlpha2)>>13)) >> 6;
        iNumInterval--;
      }

      *pfQuantPeriod = (float)iTempPeriod / 65536.0;
      *piReliableFlag = FALSE;

      return;

    }


    /*
     * Finally, check the thresholds above the multi-level
     * threshold to inverse quantize the pitch period
     */

    else
    {

      iColIndex = cQPIndex - NUM_MULTI_LEVELS_2;
      iTemp = piRatioThld_2[iRowIndex][iColIndex];
      iQuantPeriod = (((iRefHigh*iTemp) << 3) + ((iRefLow*iTemp) >> 13)) >> 6;
      if (iQuantPeriod > MAX_PERIOD)
      {
        iQuantPeriod = MAX_PERIOD;
      }
      *pfQuantPeriod = (float)iQuantPeriod / 65536.0;
      *piReliableFlag = FALSE;
      return;

    }

  }

}





void get_class(UWord8 class_index, float pitch_period, UWord8 *class)
{


  if (class_index == 0x00)
  {
    if (pitch_period <= 0.0)
    {
      *class = '0';
    }
    else
    {
      *class = '2';
    }
  }
  else if (class_index == 0x01)
  {
    if (pitch_period <= 0.0)
    {
      *class = '1';
    }
    else
    {
      *class = '3';
    }
  }
  else
  {
    printf("ERROR found in class index\n\n");
  }

  return;

}


#endif
